// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Urls;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Client {
  public partial class ImapMailbox : IImapUrl {
    public ImapClient Client {
      get { return client; }
    }

    public Uri Url {
      get { return mailbox.Url; }
    }

    public string FullName {
      get { return mailbox.Name; }
    }

    public string Name {
      get { return mailbox.LeafName; }
    }

    public string ParentMailboxName {
      get { return mailbox.SuperiorName; }
    }

    public string MailboxSeparator {
      get { return mailbox.HierarchyDelimiter; }
    }

    public long UidValidity {
      get { return mailbox.UidValidity; }
    }

    public long UnseenMessageCount {
      get { return mailbox.UnseenMessage; }
    }

    [Obsolete("use ExistingMessageCount instead")]
    public long ExistMessageCount => ExistingMessageCount;

    public long ExistingMessageCount {
      get { return mailbox.ExistsMessage; }
    }

    public long RecentMessageCount {
      get { return mailbox.RecentMessage; }
    }

    public long NextUid {
      get { return mailbox.UidNext; }
    }

    public IReadOnlyStringSet<ImapMailboxFlag> Flags {
      get { return mailbox.Flags; }
    }

    [CLSCompliant(false)]
    public ulong HighestModSeq {
      get { return mailbox.HighestModSeq; }
    }

    public bool IsModSequencesAvailable {
      get { return mailbox.ModificationSequences || 0 < mailbox.HighestModSeq; }
    }

    public bool Exists {
      get { return !deleted && !mailbox.Flags.Contains(ImapMailboxFlag.NonExistent); }
    }

    public bool IsUnselectable {
      get { return mailbox.IsUnselectable; }
    }

    public virtual bool IsOpen {
      get {
        if (client.OpeningMailbox == null)
          return false;
        else
          return mailbox.NameEquals(client.OpeningMailbox.mailbox);
      }
    }

    public bool IsInbox {
      get { return mailbox.IsInbox; }
    }

    public bool CanHaveChild {
      get { return !CanNotHaveChild; }
    }

    protected bool CanNotHaveChild {
      get { return mailbox.Flags.Contains(ImapMailboxFlag.NoInferiors) || string.IsNullOrEmpty(mailbox.HierarchyDelimiter); }
    }

    internal protected ImapMailboxStatus Mailbox {
      get { return mailbox; }
    }

    internal ImapMailbox(ImapClient client, ImapMailboxStatus mailbox)
    {
      this.client = client;
      this.mailbox = mailbox;
    }

    private ImapClient client;
    private ImapMailboxStatus mailbox;
    private bool deleted = false;

    /*
     * operations
     */
    public void Refresh(bool reretrieveMailboxFlags = false)
    {
      client.RunExclusively(() => RefreshCore(reretrieveMailboxFlags));
    }

    public Task RefreshAsync(bool reretrieveMailboxFlags = false)
    {
      return client.StartTaskExclusively(() => RefreshCore(reretrieveMailboxFlags),
                                         CancellationToken.None);
    }

    protected internal virtual void RefreshCore(bool reretrieveMailboxFlags)
    {
      if (reretrieveMailboxFlags)
        ReRetrieveMailboxFlags();

      if (!Exists || mailbox.IsUnselectable)
        return;

      ProcessResult(client.Session.Status(mailbox,
                                          client.GetAvailableStatusDataItems()));
    }

    protected void ReRetrieveMailboxFlags()
    {
#if DEBUG
      client.ThrowIfNotInExclusiveOperation();
#endif

      ProcessResult(client.Session.List(string.Empty, mailbox.Name, out var listedMailbox));

      var newMailbox = listedMailbox.FirstOrDefault(mailbox.NameEquals);

      if (newMailbox == null) {
        deleted = true;
      }
      else {
        deleted = false;
        mailbox = newMailbox; // instance will be replaced if mailbox is re-created
      }
    }

    /*
     * SELECT/EXAMINE
     */
    public virtual ImapOpenMailbox Open(bool asReadOnly = ImapClient.DefaultSelectReadOnly)
    {
      ThrowIfNonExistentOrUnselectable();

      return client.OpenMailbox(this, asReadOnly);
    }

    public virtual Task<ImapOpenMailbox> OpenAsync(bool asReadOnly = ImapClient.DefaultSelectReadOnly)
    {
      ThrowIfNonExistentOrUnselectable();

      return client.OpenMailboxAsync(this, asReadOnly);
    }

    /*
     * DELETE
     */
    public void Delete(bool unsubscribe = ImapClient.DefaultSubscription)
    {
      client.RunExclusively(() => DeleteCore(unsubscribe));
    }

    public Task DeleteAsync(bool unsubscribe = ImapClient.DefaultSubscription)
    {
      return client.StartTaskExclusively(() => DeleteCore(unsubscribe),
                                         CancellationToken.None);
    }

    private void DeleteCore(bool unsubscribe)
    {
      client.CloseMailboxIfSelected(this);

      if (Exists)
        ProcessResult(client.Session.Delete(mailbox),
                      delegate(ImapTaggedStatusResponse status) {
          if (status.ResponseText.Code == ImapResponseCode.NonExistent)
            return false; // throw no exception
          else
            return true;
        });

      deleted = true;

      if (unsubscribe)
        ProcessResult(client.Session.Unsubscribe(mailbox.Name)); // mailbox is detached already
    }

    /*
     * CREATE
     */
    public void Create(bool subscribe = ImapClient.DefaultSubscription)
    {
      client.RunExclusively(() => CreateCore(subscribe));
    }

    public Task CreateAsync(bool subscribe = ImapClient.DefaultSubscription)
    {
      return client.StartTaskExclusively(() => CreateCore(subscribe),
                                         CancellationToken.None);
    }

    private void CreateCore(bool subscribe)
    {
      ProcessResult(client.Session.Create(mailbox.Name),
                    delegate(ImapTaggedStatusResponse status) {
        if (status.ResponseText.Code == ImapResponseCode.AlreadyExists)
          return false; // throw no exception
        else
          return true;
      });

      deleted = false;

      // remove \NonExistent flag
      if (mailbox.Flags.Contains(ImapMailboxFlag.NonExistent))
        mailbox.Flags = new ImapMailboxFlagSet(mailbox.Flags.Where(f => f != ImapMailboxFlag.NonExistent));

      if (subscribe)
        ProcessResult(client.Session.Subscribe(mailbox.Name));
    }

    public static ImapMailbox Create(ImapClient client,
                                     string mailboxName,
                                     bool subscribe = ImapClient.DefaultSubscription)
    {
      if (client == null)
        throw new ArgumentNullException(nameof(client));

      return client.CreateMailbox(mailboxName, subscribe);
    }

    public static Task<ImapMailbox> CreateAsync(ImapClient client,
                                                string mailboxName,
                                                bool subscribe = ImapClient.DefaultSubscription)
    {
      if (client == null)
        throw new ArgumentNullException(nameof(client));

      return client.CreateMailboxAsync(mailboxName, subscribe);
    }

    /*
     * RENAME
     */
    public void MoveTo(ImapMailbox destinationMailbox,
                       bool subscribe = ImapClient.DefaultSubscription)
    {
      ValidateMoveDestinationMailbox(destinationMailbox);

      MoveTo(destinationMailbox.GetFullNameOf(mailbox.LeafName),
             subscribe);
    }

    public Task MoveToAsync(ImapMailbox destinationMailbox,
                            bool subscribe = ImapClient.DefaultSubscription)
    {
      ValidateMoveDestinationMailbox(destinationMailbox);

      return MoveToAsync(destinationMailbox.GetFullNameOf(mailbox.LeafName),
                         subscribe);
    }

    private void ValidateMoveDestinationMailbox(ImapMailbox destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));
      if (destinationMailbox.client != this.client)
        throw new NotSupportedException("moving to mailbox of different session is not supported. use ImapClient.MoveMailbox().");

      destinationMailbox.ThrowIfCanNotHaveChild();
    }

    public void MoveTo(string newMailboxName,
                       bool subscribe = ImapClient.DefaultSubscription)
    {
      ImapMailboxNames.GetValidated(newMailboxName, nameof(newMailboxName));

      ValidateMoveDestinationMailboxName(newMailboxName);

      client.RunExclusively(() => MoveToCore(newMailboxName, subscribe));
    }

    public Task MoveToAsync(string newMailboxName,
                            bool subscribe = ImapClient.DefaultSubscription)
    {
      ImapMailboxNames.GetValidated(newMailboxName, nameof(newMailboxName));

      ValidateMoveDestinationMailboxName(newMailboxName);

      return client.StartTaskExclusively(() => MoveToCore(newMailboxName, subscribe),
                                         CancellationToken.None);
    }

    private void ValidateMoveDestinationMailboxName(string destinationMailboxName)
    {
      if (mailbox.NameEquals(destinationMailboxName))
        throw new ImapProtocolViolationException("new mailbox name must be different from current mailbox name");
      if (ImapMailboxNames.IsInbox(destinationMailboxName))
        throw new ImapProtocolViolationException("can't move to INBOX");
    }

    private void MoveToCore(string newMailboxName,
                            bool subscribe)
    {
      var isInbox = mailbox.IsInbox;
      var oldMailboxName = mailbox.Name;

      client.CloseMailboxIfSelected(this);

      Dictionary<string, ImapMailbox> subscribingChildren = null;

      if (subscribe && !isInbox) {
        subscribingChildren = new Dictionary<string, ImapMailbox>();

        foreach (var subscribingChild in client.GetMailboxesCore(mailbox, ImapMailboxListOptions.SubscribedOnly)) {
          subscribingChildren.Add(subscribingChild.FullName, subscribingChild);
        }
      }

      ProcessResult(client.Session.Rename(mailbox, newMailboxName));

      if (subscribe) {
        if (!isInbox)
          // don't UNSUBSCRIBE old INBOX
          ProcessResult(client.Session.Unsubscribe(oldMailboxName));

        if (Exists)
          ProcessResult(client.Session.Subscribe(newMailboxName));

        if (!isInbox) {
          foreach (var pair in subscribingChildren) {
            ProcessResult(client.Session.Unsubscribe(pair.Key));

            if (pair.Value.Exists)
              pair.Value.SubscribeCore(false); // mailbox name is renamed
          }
        }
      }
    }

    /*
     * SUBSCRIBE/UNSUBSCRIBE
     */
    private const bool DefaultSubscribeRecursive = false;

    public void Subscribe(bool recursive = DefaultSubscribeRecursive)
    {
      client.RunExclusively(() => SubscribeCore(recursive));
    }

    public Task SubscribeAsync(bool recursive = DefaultSubscribeRecursive)
    {
      return client.StartTaskExclusively(() => SubscribeCore(recursive),
                                         CancellationToken.None);
    }

    private void SubscribeCore(bool recursive)
    {
      ProcessResult(client.Session.Subscribe(mailbox));

      if (!recursive)
        return;

      foreach (var childMailbox in GetChildrenCore(ImapMailboxListOptions.Default)) {
        ProcessResult(client.Session.Subscribe(childMailbox.mailbox));
      }
    }

    public void Unsubscribe(bool recursive = DefaultSubscribeRecursive)
    {
      client.RunExclusively(() => UnsubscribeCore(recursive));
    }

    public Task UnsubscribeAsync(bool recursive = DefaultSubscribeRecursive)
    {
      return client.StartTaskExclusively(() => UnsubscribeCore(recursive),
                                         CancellationToken.None);
    }

    private void UnsubscribeCore(bool recursive)
    {
      ProcessResult(client.Session.Unsubscribe(mailbox));

      if (!recursive)
        return;

      foreach (var childMailbox in GetChildrenCore(ImapMailboxListOptions.SubscribedOnly)) {
        ProcessResult(client.Session.Unsubscribe(childMailbox.mailbox));
      }
    }

    /*
     * LIST/LSUB/CREATE
     */
    public IReadOnlyList<ImapMailbox> GetChildren(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return client.RunExclusively(() => GetChildrenCore(options));
    }

    public Task<IReadOnlyList<ImapMailbox>> GetChildrenAsync(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return client.StartTaskExclusively(() => GetChildrenCore(options),
                                         CancellationToken.None);
    }

    private IReadOnlyList<ImapMailbox> GetChildrenCore(ImapMailboxListOptions options)
    {
      return client.GetMailboxesCore(mailbox, options);
    }

    public ImapMailbox GetParent(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      if (mailbox.SuperiorName.Length == 0)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is top-level mailbox", FullName));

      return client.GetMailbox(mailbox.SuperiorName, options);
    }

    public Task<ImapMailbox> GetParentAsync(ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      if (mailbox.SuperiorName.Length == 0)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is top-level mailbox", FullName));

      return client.GetMailboxAsync(mailbox.SuperiorName, options);
    }

    public ImapMailbox GetOrCreateParent(ImapMailboxListOptions options)
    {
      return GetOrCreateParent(ImapClient.DefaultSubscription,
                               options);
    }

    public ImapMailbox GetOrCreateParent(bool subscribeIfCreated = ImapClient.DefaultSubscription,
                                         ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      if (mailbox.SuperiorName.Length == 0)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is top-level mailbox", FullName));

      return client.GetOrCreateMailbox(mailbox.SuperiorName,
                                       subscribeIfCreated,
                                       options);
    }

    public Task<ImapMailbox> GetOrCreateParentAsync(ImapMailboxListOptions options)
    {
      return GetOrCreateParentAsync(ImapClient.DefaultSubscription,
                                    options);
    }

    public Task<ImapMailbox> GetOrCreateParentAsync(bool subscribeIfCreated = ImapClient.DefaultSubscription,
                                                    ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      if (mailbox.SuperiorName.Length == 0)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is top-level mailbox", FullName));

      return client.GetOrCreateMailboxAsync(mailbox.SuperiorName,
                                            subscribeIfCreated,
                                            options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(string inferiorName)
    {
      return GetFullNameOf(Singleton.CreateList(inferiorName));
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(string inferiorName1, string inferiorName2)
    {
      return GetFullNameOf((IEnumerable<string>)new string[] {inferiorName1, inferiorName2});
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(string inferiorName1, string inferiorName2, string inferiorName3)
    {
      return GetFullNameOf((IEnumerable<string>)new string[] {inferiorName1, inferiorName2, inferiorName3});
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(params string[] inferiorNames)
    {
      return GetFullNameOf((IEnumerable<string>)inferiorNames);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string GetFullNameOf(IEnumerable<string> inferiorNames)
    {
      if (inferiorNames == null)
        throw new ArgumentNullException(nameof(inferiorNames));

      ThrowIfCanNotHaveChild();

      var builder = new StringBuilder(128);

      builder.Append(mailbox.Name);

      foreach (var inferiorName in inferiorNames) {
        builder.Append(mailbox.HierarchyDelimiter);
        builder.Append(ImapMailboxNames.GetValidated(inferiorName, nameof(inferiorNames)));
      }

      return builder.ToString();
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public string[] SplitMailboxHierarchy(string mailboxName)
    {
      ImapMailboxNames.GetValidated(mailboxName, nameof(mailboxName));

      ThrowIfCanNotHaveChild();

      return mailboxName.Split(new[] {mailbox.HierarchyDelimiter}, StringSplitOptions.None);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailbox CreateChild(string name,
                                   bool subscribe = ImapClient.DefaultSubscription)
    {
      return client.CreateMailbox(GetFullNameOf(name), subscribe);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public Task<ImapMailbox> CreateChildAsync(string name,
                                              bool subscribe = ImapClient.DefaultSubscription)
    {
      return client.CreateMailboxAsync(GetFullNameOf(name), subscribe);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailbox GetChild(string name,
                                ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return client.GetMailbox(GetFullNameOf(name), options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public Task<ImapMailbox> GetChildAsync(string name,
                                           ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return client.GetMailboxAsync(GetFullNameOf(name), options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public ImapMailbox GetOrCreateChild(string name,
                                        ImapMailboxListOptions options)
    {
      return GetOrCreateChild(name,
                              ImapClient.DefaultSubscription,
                              options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public ImapMailbox GetOrCreateChild(string name,
                                        bool subscribeIfCreated = ImapClient.DefaultSubscription,
                                        ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return client.GetOrCreateMailbox(GetFullNameOf(name),
                                       subscribeIfCreated,
                                       options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<ImapMailbox> GetOrCreateChildAsync(string name,
                                                   ImapMailboxListOptions options)
    {
      return GetOrCreateChildAsync(name,
                                   ImapClient.DefaultSubscription,
                                   options);
    }

    /// <exception cref="ImapProtocolViolationException">Throws if <see cref="CanHaveChild"/> is false.</exception>
    public Task<ImapMailbox> GetOrCreateChildAsync(string name,
                                                   bool subscribeIfCreated = ImapClient.DefaultSubscription,
                                                   ImapMailboxListOptions options = ImapMailboxListOptions.Default)
    {
      return client.GetOrCreateMailboxAsync(GetFullNameOf(name),
                                            subscribeIfCreated,
                                            options);
    }

    /*
     * GETQUOTAROOT
     */
    public IReadOnlyList<ImapQuota> GetQuota()
    {
      if (!client.IsCapabilityAvailable(ImapCapability.Quota))
#if NET45
        return ArrayExtensions.Empty<ImapQuota>();
#else
        return Array.Empty<ImapQuota>();
#endif

      return client.RunExclusively(() => GetQuotaCore());
    }

    public Task<IReadOnlyList<ImapQuota>> GetQuotaAsync()
    {
      if (!client.IsCapabilityAvailable(ImapCapability.Quota)) {
        var tcs = new TaskCompletionSource<IReadOnlyList<ImapQuota>>();

#if NET45
        tcs.SetResult(ArrayExtensions.Empty<ImapQuota>());
#else
        tcs.SetResult(Array.Empty<ImapQuota>());
#endif

        return tcs.Task;
      }

      return client.StartTaskExclusively(() => GetQuotaCore(),
                                         CancellationToken.None);
    }

    private IReadOnlyList<ImapQuota> GetQuotaCore()
    {
      ProcessResult(client.Session.GetQuotaRoot(Mailbox, out var roots));

      var ret = new List<ImapQuota>();

      foreach (var pair in roots) {
        ret.AddRange(pair.Value);
      }

      return ret;
    }

    internal protected virtual ImapCommandResult ProcessResult(ImapCommandResult result,
                                                               Func<ImapTaggedStatusResponse, bool> throwIfErrorResponse = default)
      => client.ProcessResult(result, throwIfErrorResponse);

    internal void ThrowIfCanNotHaveChild()
    {
      if (CanNotHaveChild)
        throw new ImapProtocolViolationException(string.Format("It is not possible for any child levels of hierarchy to exist under this mailbox. (mailbox name: '{0}')", FullName));
    }

    internal void ThrowIfNonExistentOrUnselectable()
    {
      if (!Exists)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is not existent", FullName));
      if (mailbox.IsUnselectable)
        throw new ImapProtocolViolationException(string.Format("mailbox '{0}' is unselectable", FullName));
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      throw new NotSupportedException();
    }

    public override string ToString()
    {
      return string.Format("{{ImapMailbox: Authority='{0}', FullName='{1}', UidValidity={2}, NextUid={3}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           FullName,
                           UidValidity,
                           NextUid);
    }
  }
}
