// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Collections;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessage {
    public void Refresh()
    {
      Client.RunExclusively(() => FetchDynamicAttributes());
    }

    public Task RefreshAsync()
    {
      return Client.StartTaskExclusively(() => FetchDynamicAttributes(),
                                         CancellationToken.None);
    }

    private bool FindBySequence(ImapMessageAttributesBase attributes)
    {
      return attributes.Sequence == Sequence;
    }

    internal static readonly ImapFetchDataItems FetchStaticDataItems =
      ImapFetchDataItems.BodyStructure +
      ImapFetchDataItems.Envelope + 
      ImapFetchDataItems.InternalDate +
      ImapFetchDataItems.Rfc822Size;

    private void EnsureStaticAttributesFetched()
    {
      if (staticAttributes != null)
        return;

      Client.RunExclusively(() => {
        ThrowIfSequenceOrUidSetIsNotAvailable(out var sequenceOrUidSet);

        Mailbox.ProcessResult(Client.Session.Fetch(sequenceOrUidSet,
                                                   FetchStaticDataItems,
                                                   out ImapMessageStaticAttributes[] staticAttrs),
                              null);

        var staticAttr = staticAttrs.FirstOrDefault(FindBySequence);

        if (staticAttr == null) {
          SetDeleted();
          throw new ImapMessageVanishedException(this);
        }
        else {
          staticAttributes = staticAttr;
        }
      });
    }

    private static readonly ImapFetchDataItems fetchFlagsAndModSeq = ImapFetchDataItems.Flags + ImapFetchDataItems.ModSeq;

    internal static ImapFetchDataItems GetFetchDynamicDataItems(ImapOpenMailbox mailbox)
    {
      if (mailbox.IsModSequencesAvailable)
        return fetchFlagsAndModSeq;
      else
        return ImapFetchDataItems.Flags;
    }

    internal static ImapFetchDataItems TranslateFetchOption(ImapOpenMailbox mailbox,
                                                            ImapMessageFetchAttributeOptions options,
                                                            out bool fetchStaticAttr,
                                                            out bool fetchDynamicAttr)
    {
      fetchStaticAttr = (int)(options & ImapMessageFetchAttributeOptions.StaticAttributes) != 0;
      fetchDynamicAttr = (int)(options & ImapMessageFetchAttributeOptions.DynamicAttributes) != 0;

      var fetchDataItems = ImapFetchDataItems.Uid;

      if (fetchStaticAttr)
        fetchDataItems += ImapMessage.FetchStaticDataItems;
      if (fetchDynamicAttr)
        fetchDataItems += ImapMessage.GetFetchDynamicDataItems(mailbox);

      return fetchDataItems;
    }

    private void EnsureDynamicAttributesFetched()
    {
      if (dynamicAttributes != null)
        return;

      Client.RunExclusively(() => FetchDynamicAttributes());
    }

    private void FetchDynamicAttributes()
    {
#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      ThrowIfSequenceOrUidSetIsNotAvailable(out var sequenceOrUidSet);

      Mailbox.ProcessResult(Client.Session.Fetch(sequenceOrUidSet,
                                                 GetFetchDynamicDataItems(Mailbox),
                                                 out ImapMessageDynamicAttributes[] dynamicAttrs),
                            null);

      var dynamicAttr = dynamicAttrs.FirstOrDefault(FindBySequence);

      if (dynamicAttr == null) {
        SetDeleted();
        throw new ImapMessageVanishedException(this);
      }
      else {
        dynamicAttributes = dynamicAttr;

        // ImapOpenMailbox.ProcessResult does not raise event if dynamic attribute created newly, so raise here
        Mailbox.QueueMessageStatusChangedEvent(Singleton.CreateList(this));
      }
    }

    public void ToggleFlags(ImapMessageFlag flag,
                            params ImapMessageFlag[] flags)
    {
      ToggleFlags(new ImapMessageFlagSet(ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags)));
    }

    public void ToggleKeywords(string keyword,
                               params string[] keywords)
    {
      ToggleFlags(new ImapMessageFlagSet(Enumerable.Empty<ImapMessageFlag>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords)));
    }

    public void ToggleFlags(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      if (flagsAndKeywords == null)
        throw new ArgumentNullException(nameof(flagsAndKeywords));

      Client.RunExclusively(() => ToggleFlagsCore(flagsAndKeywords));
    }

    public Task ToggleFlagsAsync(ImapMessageFlag flag,
                                 params ImapMessageFlag[] flags)
    {
      return ToggleFlagsAsync(new ImapMessageFlagSet(ParamArrayUtils.ToEnumerableNonNullable(nameof(flags), flag, flags)));
    }

    public Task ToggleKeywordsAsync(string keyword,
                                    params string[] keywords)
    {
      return ToggleFlagsAsync(new ImapMessageFlagSet(Enumerable.Empty<ImapMessageFlag>(), ParamArrayUtils.ToEnumerableNonNullable(nameof(keywords), keyword, keywords)));
    }

    public Task ToggleFlagsAsync(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      if (flagsAndKeywords == null)
        throw new ArgumentNullException(nameof(flagsAndKeywords));

      return Client.StartTaskExclusively(() => ToggleFlagsCore(flagsAndKeywords),
                                         CancellationToken.None);
    }

    private void ToggleFlagsCore(IEnumerable<ImapMessageFlag> flagsAndKeywords)
    {
      var addFlags = new List<ImapMessageFlag>();
      var removeFlags = new List<ImapMessageFlag>();

      if (dynamicAttributes == null)
        FetchDynamicAttributes();

      foreach (var flag in flagsAndKeywords) {
        if (dynamicAttributes.Flags.Contains(flag))
          removeFlags.Add(flag);
        else
          addFlags.Add(flag);
      }

      if (0 < removeFlags.Count)
        StoreCore(ImapStoreDataItems.RemoveFlags(removeFlags));

      if (0 < addFlags.Count)
        StoreCore(ImapStoreDataItems.AddFlags(addFlags));
    }
  }
}
