// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Urls;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Client {
  public sealed partial class ImapMessage :
    ImapMessageReference,
    IImapMessageDynamicAttributes,
    IImapMessageStaticAttributes,
    IImapUrl,
    IImapMessageSequenceNumber
  {
    public static readonly long ExpungedMessageSequenceNumber = 0L;

    public long Sequence {
      get; internal set;
    }

    public long Uid {
      get; private set;
    }

    public Uri Url {
      get { return uriBuilder.Uri; }
    }

    public ImapMessageBody Body { get; }
    public ImapMessageHeader Header { get; }

    /*
     * IImapMessageDynamicAttribute
     */
    private ImapMessageDynamicAttributes dynamicAttributes;

    internal ImapMessageDynamicAttributes DynamicAttributes {
      get { EnsureDynamicAttributesFetched(); return dynamicAttributes; }
      set { dynamicAttributes = value; }
    }

    internal ImapMessageDynamicAttributes GetDynamicAttributes()
    {
      return dynamicAttributes;
    }

    public IReadOnlyStringSet<ImapMessageFlag> Flags => DynamicAttributes.Flags;

    [CLSCompliant(false)]
    public ulong ModSeq {
      get { return DynamicAttributes.ModSeq; }
    }

    /*
     * IImapMessageStaticAttribute
     */
    private ImapMessageStaticAttributes staticAttributes;

    internal ImapMessageStaticAttributes StaticAttributes {
      get { EnsureStaticAttributesFetched(); return staticAttributes; }
      set { staticAttributes = value; }
    }

    private ImapMessagePart messagePart = null;

    public ImapMessagePart MessagePart {
      get {
        return LazyInitializer.EnsureInitialized(ref messagePart, delegate {
          return new ImapMessagePart(this);
        });
      }
    }

    public ImapEnvelope Envelope {
      get { return StaticAttributes.Envelope; }
    }

    [Obsolete("use DeliveredDate instead")]
    public DateTimeOffset InternalDate {
      get { return DeliveredDate; }
    }

    public DateTimeOffset DeliveredDate {
      get { return StaticAttributes.InternalDate.Value; }
    }

    DateTimeOffset? IImapMessageStaticAttributes.InternalDate {
      get { return StaticAttributes.InternalDate; }
    }

    public long Length {
      get { return StaticAttributes.Rfc822Size; }
    }

    long IImapMessageStaticAttributes.Rfc822Size {
      get { return StaticAttributes.Rfc822Size; }
    }

    IReadOnlyCollection<string> IImapMessageStaticAttributes.BinarySizeSpecifiers => StaticAttributes.BinarySizeSpecifiers;

    long IImapMessageStaticAttributes.GetBinarySizeOf(string section)
    {
      return StaticAttributes.GetBinarySizeOf(section);
    }

    IImapBodyStructure IImapMessageStaticAttributes.BodyStructure {
      get { return StaticAttributes.BodyStructure; }
    }

    /*
     * miscellaneous getter properties
     */
    public bool IsMarkedAsDeleted {
      get { return Flags.Contains(ImapMessageFlag.Deleted); }
    }

    public bool IsSeen {
      get { return Flags.Contains(ImapMessageFlag.Seen); }
    }

    public bool IsRecent {
      get { return Flags.Contains(ImapMessageFlag.Recent); }
    }

    public bool IsAnswered {
      get { return Flags.Contains(ImapMessageFlag.Answered); }
    }

    public bool IsDraft {
      get { return Flags.Contains(ImapMessageFlag.Draft); }
    }

    public bool IsFlagged {
      get { return Flags.Contains(ImapMessageFlag.Flagged); }
    }

    public bool IsDeleted {
      get { return Sequence == ExpungedMessageSequenceNumber; }
    }

    public string EnvelopeSubject {
      get { return StaticAttributes.Envelope.GetDecodedSubject(null, false); }
    }

    public DateTimeOffset? EnvelopeDate {
      get { return StaticAttributes.Envelope.GetDecodedDate(false); }
    }

    public MimeType MediaType {
      get { return StaticAttributes.BodyStructure.MediaType; }
    }

    public bool IsMultiPart {
      get { return StaticAttributes.BodyStructure.IsMultiPart; }
    }

    public bool HasEncapsulatedMessage {
      get { return StaticAttributes.BodyStructure.HasEncapsulatedMessage; }
    }

    internal ImapMessage(ImapOpenMailbox mailbox, long uid, long sequence)
      : base(mailbox)
    {
      this.Uid = uid;
      this.Sequence = sequence;
      this.Body = new ImapMessageBody(this);
      this.Header = new ImapMessageHeader(this);

      (this as IImapUrl).SetBaseUrl(mailbox.Mailbox.UrlBuilder);
    }

    private ImapUriBuilder uriBuilder;

    protected override IImapNumberSet GetSequenceOrUidSet() => ImapUidSet.Of(Uid);

    internal void SetDeleted()
    {
      Sequence = ExpungedMessageSequenceNumber;

      if (IsAvailable)
        Mailbox.Detach(this);
    }

    /*
     * message part
     */
    public ImapMessagePart GetMessagePart(string partSpecifier)
    {
      partSpecifier = ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(partSpecifier);

      if (string.IsNullOrEmpty(partSpecifier)) {
        return MessagePart;
      }
      else {
        return MessagePart.Find(delegate(ImapMessagePart part) {
          return string.Equals(part.Specifier, partSpecifier, StringComparison.Ordinal);
        });
      }
    }

    public ImapMessagePart GetMessagePart(params int[] partSpecifiers)
    {
      return GetMessagePart((IEnumerable<int>)partSpecifiers);
    }

    public ImapMessagePart GetMessagePart(/*params*/ IEnumerable<int> partSpecifiers)
    {
      var builder = new StringBuilder(0x10);

      foreach (var partSpecifier in partSpecifiers ?? Enumerable.Empty<int>()) {
        if (0 < builder.Length)
          builder.Append('.');

        builder.Append(partSpecifier); // not check whether each value is valid or not here
      }

      return GetMessagePart(builder.ToString());
    }

    public ImapMessagePart FindMessagePart(Predicate<ImapMessagePart> match)
      => MessagePart.Find(match);

    public IReadOnlyList<ImapMessagePart> FindAllMessageParts(Predicate<ImapMessagePart> match)
      => MessagePart.FindAll(match);

    // FindMessagePartByMimeType
    public ImapMessagePart FindMessagePart(MimeType mediaType)
    {
      if (mediaType == null)
        throw new ArgumentNullException(nameof(mediaType));

      return MessagePart.Find(delegate(ImapMessagePart part) {
        return mediaType.EqualsIgnoreCase(part.ContentType);
      });
    }

    public IReadOnlyList<ImapMessagePart> FindAllMessageParts(MimeType mediaType)
    {
      if (mediaType == null)
        throw new ArgumentNullException(nameof(mediaType));

      return MessagePart.FindAll(delegate(ImapMessagePart part) {
        return mediaType.EqualsIgnoreCase(part.ContentType);
      });
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      uriBuilder = baseUrl.Clone();

      uriBuilder.Uid = Uid;
      uriBuilder.UidValidity = UidValidity;
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessage: Authority='{0}', Mailbox='{1}', Sequence={2}, Uid={3}, UidValidity={4}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           Mailbox.FullName,
                           Sequence,
                           Uid,
                           UidValidity);
    }
  }
}
