// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;

namespace Smdn.Net.Imap4.Client {
  internal sealed class ImapMessageCollection :
    IEnumerable<ImapMessage>,
    IImapMessageAttributeCollection<ImapMessageDynamicAttributes>
  {
    IEnumerator<ImapMessage> IEnumerable<ImapMessage>.GetEnumerator()
    {
      return uidMap.Values.GetEnumerator();
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return (this as IEnumerable<ImapMessage>).GetEnumerator();
    }

    internal static int CompareMessageAttributeSequence(IImapMessageSequenceNumber x, IImapMessageSequenceNumber y)
    {
      if (x.Sequence < y.Sequence)
        return -1;
      else if (x.Sequence > y.Sequence)
        return 1;
      else
        return 0;
    }

    public void Add(ImapMessage item)
    {
      uidMap.Add(item.Uid, item);
      items.Add(item);

      sorted = false;
    }

    public void Remove(ImapMessage item)
    {
      uidMap.Remove(item.Uid);
      items.Remove(item);
    }

    public void Clear()
    {
      uidMap.Clear();
      items.Clear();
    }

    public void Expunge(long sequenceNumber, out ImapMessage expungedItem)
    {
      expungedItem = null;

      foreach (var item in items) {
        if (item.Sequence == sequenceNumber)
          expungedItem = item;
        else if (sequenceNumber < item.Sequence)
          item.Sequence--;
      }

      // order of items is not changing
      //sorted = false;

      expungedItem?.SetDeleted();
    }

    public void ChangeSequenceNumber(ImapMessage item, long newSequenceNumber)
    {
      item.Sequence = newSequenceNumber;

      sorted = false;
    }

    public ImapMessage FindByUid(long uid)
    {
      if (uidMap.TryGetValue(uid, out var ret))
        return ret;
      else
        return null;
    }

    public bool TryGetByUid(long uid, out ImapMessage item) => uidMap.TryGetValue(uid, out item);

    public bool TryGetBySequence(long sequenceNumber, out ImapMessage item)
    {
      if (!sorted) {
        items.Sort(CompareMessageAttributeSequence);
        sorted = true;
      }

      item = FindBySequence(items, sequenceNumber);

      return item != null;
    }

    /// <param name="list">list of IImapMessageAttribute. must be sorted.</param>
    internal static TImapMessage FindBySequence<TImapMessage>(IReadOnlyList<TImapMessage> list, long sequenceNumber)
      where TImapMessage : class, IImapMessageSequenceNumber
    {
      // binary search
      var min = 0;
      var max = list.Count - 1;

      while (min <= max) {
        var mid = min + (max - min) / 2;
        var item = list[mid];

        if (item.Sequence == sequenceNumber)
          return item;
        else if (item.Sequence < sequenceNumber)
          min = mid + 1;
        else if (item.Sequence > sequenceNumber)
          max = mid - 1;
      }

      return null;
    }

    ImapMessageDynamicAttributes IImapMessageAttributeCollection<ImapMessageDynamicAttributes>.Find(long sequence)
    {
      if (TryGetBySequence(sequence, out var item))
        return item.GetDynamicAttributes();
      else
        return null;
    }

    void IImapMessageAttributeCollection<ImapMessageDynamicAttributes>.Add(ImapMessageDynamicAttributes @value)
    {
      if (TryGetBySequence(@value.Sequence, out var item))
        item.DynamicAttributes = @value;
    }

    private bool sorted = false;
    private readonly Dictionary<long, ImapMessage> uidMap = new Dictionary<long, ImapMessage>();
    private readonly List<ImapMessage> items = new List<ImapMessage>();
  }
}
