// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  partial class ImapOpenMailbox {
    /*
     * GetMessage by FETCHing with uid or sequence
     */

    /*
     * 6.4.8. UID Command
     *       A non-existent unique identifier is ignored without any error
     *       message generated.  Thus, it is possible for a UID FETCH command
     *       to return an OK without any data or a UID COPY or UID STORE to
     *       return an OK without performing any operations.
     */
    public ImapMessage GetMessageByUid(long uid,
                                       ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (uid <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(uid), uid);

      return (new ConcreteMessageQuery(this,
                                       options,
                                       ImapUidSet.Of(uid))).ToSingleMessage();
    }

    public Task<ImapMessage> GetMessageByUidAsync(long uid,
                                                  ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (uid <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(uid), uid);

      return (new ConcreteMessageQuery(this,
                                       options,
                                       ImapUidSet.Of(uid))).ToSingleMessageAsync();
    }

    private long GetValidatedSequenceNumber(long sequence)
    {
      if (sequence <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(sequence), sequence);

      if (ExistingMessageCount < sequence)
        throw new ArgumentOutOfRangeException(nameof(sequence),
                                              sequence,
                                              string.Format("specified sequence number is greater than exist message count. (exist message count = {0})",
                                                            ExistingMessageCount));

      return sequence;
    }

    public ImapMessage GetMessageBySequence(long sequence,
                                            ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return (new ConcreteMessageQuery(this,
                                       options,
                                       ImapSequenceSet.Of(GetValidatedSequenceNumber(sequence)))).ToSingleMessage();
    }

    public Task<ImapMessage> GetMessageBySequenceAsync(long sequence,
                                                       ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return (new ConcreteMessageQuery(this,
                                       options,
                                       ImapSequenceSet.Of(GetValidatedSequenceNumber(sequence)))).ToSingleMessageAsync();
    }

    public ImapMessage GetMessageFirstUnseen(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (FirstUnseenMessageNumber == 0L)
        return null;
      else
        return GetMessageBySequence(FirstUnseenMessageNumber, options);
    }

    public Task<ImapMessage> GetMessageFirstUnseenAsync(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (FirstUnseenMessageNumber == 0L) {
        var tcs = new TaskCompletionSource<ImapMessage>();

        tcs.SetResult(null);

        return tcs.Task;
      }
      else {
        return GetMessageBySequenceAsync(FirstUnseenMessageNumber, options);
      }
    }

    /*
     * GetMessages by FETCHing
     */
    public IReadOnlyList<ImapMessage> GetMessages(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessages(options).ToList();

    public ImapMessageQuery QueryMessages(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return new AllMessageQuery(this, options);
    }

    private class AllMessageQuery : ImapMessageQuery {
      public AllMessageQuery(ImapOpenMailbox mailbox,
                             ImapMessageFetchAttributeOptions fetchOptions)
        : base(mailbox, fetchOptions, false)
      {
      }

      protected override IImapNumberSet GetSequenceOrUidSet()
      {
        Mailbox.RefreshCore(false);

        if (Mailbox.ExistingMessageCount <= 0L)
          return ImapSequenceSet.Empty();
        else
          return ImapSequenceSet.Range(1L, Mailbox.ExistingMessageCount);
      }
    }

    public IReadOnlyList<ImapMessage> GetMessages(long uid, params long[] uids)
    {
      return GetMessages(ImapMessageFetchAttributeOptions.Default,
                         uid,
                         uids);
    }

    public IReadOnlyList<ImapMessage> GetMessages(ImapMessageFetchAttributeOptions options,
                                                  long uid,
                                                  params long[] uids)
      => QueryMessages(options, uid, uids).ToList();

    public IReadOnlyList<ImapMessage> GetMessages(IEnumerable<long> uids,
                                                  ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessages(uids, options).ToList();

    public ImapMessageQuery QueryMessages(long uid, params long[] uids)
    {
      return QueryMessages(ImapMessageFetchAttributeOptions.Default,
                           uid,
                           uids);
    }

    public ImapMessageQuery QueryMessages(ImapMessageFetchAttributeOptions options,
                                          long uid,
                                          params long[] uids)
    {
      return new ConcreteMessageQuery(this,
                                      options,
                                      ImapUidSet.Of(uid, uids));
    }

    public ImapMessageQuery QueryMessages(IEnumerable<long> uids,
                                          ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (uids == null)
        throw new ArgumentNullException(nameof(uids));

      return new ConcreteMessageQuery(this,
                                      options,
                                      ImapUidSet.Of(uids));
    }
    /*
     * GetMessages by SEARCHing
     */
    public IReadOnlyList<ImapMessage> GetMessagesRecent(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessagesRecent(options).ToList();

    public IReadOnlyList<ImapMessage> GetMessagesUnseen(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessagesUnseen(options).ToList();

    public IReadOnlyList<ImapMessage> GetMessagesNew(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessagesNew(options).ToList();

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public IReadOnlyList<ImapMessage> GetMessages(ImapSearchCriteria searchCriteria,
                                                  ImapMessageFetchAttributeOptions options)
      => QueryMessages(searchCriteria, options: options).ToList();

    public IReadOnlyList<ImapMessage> GetMessages(ImapSearchCriteria searchCriteria,
                                                  Encoding encoding = null,
                                                  ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessages(searchCriteria, encoding, options).ToList();

    public ImapMessageQuery QueryMessagesRecent(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return QueryMessages(ImapSearchCriteria.Recent, encoding: null, options: options);
    }

    public ImapMessageQuery QueryMessagesUnseen(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return QueryMessages(ImapSearchCriteria.Unseen, encoding: null, options: options);
    }

    public ImapMessageQuery QueryMessagesNew(ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return QueryMessages(ImapSearchCriteria.New, encoding: null, options: options);
    }

    public ImapMessageQuery QueryMessages(ImapSearchCriteria searchCriteria,
                                          Encoding encoding = null,
                                          ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (searchCriteria == null)
        throw new ArgumentNullException(nameof(searchCriteria));

      return new SearchMessageQuery(this,
                                    options,
                                    searchCriteria,
                                    encoding);
    }

    private abstract class SearchMessageQueryBase : ImapMessageQuery {
      public ImapSearchCriteria SearchCriteria {
        get; private set;
      }

      public Encoding Encoding {
        get; private set;
      }

      protected SearchMessageQueryBase(ImapOpenMailbox mailbox,
                                       ImapMessageFetchAttributeOptions fetchOptions,
                                       ImapSearchCriteria searchCriteria,
                                       Encoding encoding,
                                       bool orderBySequenceNumber)
        : base(mailbox, fetchOptions, orderBySequenceNumber)
      {
        this.SearchCriteria = searchCriteria;
        this.Encoding = encoding;
      }

      protected bool ThrowIfBadCharset(ImapTaggedStatusResponse status)
      {
        // 6.4.4. SEARCH Command
        //       If the server does not support the specified [CHARSET], it MUST
        //       return a tagged NO response (not a BAD).  This response SHOULD
        //       contain the BADCHARSET response code, which MAY list the
        //       [CHARSET]s supported by the server.
        if (status.ResponseText.Code == ImapResponseCode.BadCharset) {
          var message = string.Format("charset '{0}' is not supported by server", Encoding.WebName);
          var supportedCharsets = ImapResponseTextParser.FromBadCharset(status.ResponseText);

          if (0 < supportedCharsets.Count) {
            message += "; following charsets are supported: '" +
              string.Join("', '", supportedCharsets) +
                "'";
          }

          throw new EncodingNotSupportedException(Encoding.WebName, message);
        }

        return true;
      }
    }

    private class SearchMessageQuery : SearchMessageQueryBase {
      public SearchMessageQuery(ImapOpenMailbox mailbox,
                                ImapMessageFetchAttributeOptions fetchOptions,
                                ImapSearchCriteria searchCriteria,
                                Encoding encoding)
        : base(mailbox, fetchOptions, searchCriteria, encoding, false)
      {
      }

      protected override IImapNumberSet GetSequenceOrUidSet()
      {
        ImapMatchedSequenceSet matchedSequenceNumbers = null;

        if (Mailbox.Client.IsCapabilityAvailable(ImapCapability.Searchres)) {
          Mailbox.ProcessResult(Mailbox.Client.Session.ESearch(out matchedSequenceNumbers,
                                                               SearchCriteria,
                                                               ImapSearchResultOptions.Save,
                                                               Encoding),
                                delegate(ImapTaggedStatusResponse status) {
            if (status.ResponseText.Code == ImapResponseCode.NotSaved)
              return false; // throw no exception
            else
              return ThrowIfBadCharset(status);
          });

          if (matchedSequenceNumbers != null)
            return matchedSequenceNumbers;
        }

        Mailbox.ProcessResult(Mailbox.Client.Session.Search(out matchedSequenceNumbers,
                                                            SearchCriteria,
                                                            Encoding),
                              ThrowIfBadCharset);

        return matchedSequenceNumbers;
      }
    }

    /*
     * GetMessages by SORTing
     */
    public IReadOnlyList<ImapMessage> GetSortedMessages(ImapSortCriteria sortOrder,
                                                        ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QuerySortedMessages(sortOrder, options).ToList();

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public IReadOnlyList<ImapMessage> GetSortedMessages(ImapSortCriteria sortOrder,
                                                        ImapSearchCriteria searchCriteria,
                                                        ImapMessageFetchAttributeOptions options)
      => QuerySortedMessages(sortOrder, searchCriteria, options: options).ToList();

    public IReadOnlyList<ImapMessage> GetSortedMessages(ImapSortCriteria sortOrder,
                                                        ImapSearchCriteria searchCriteria,
                                                        Encoding encoding = null,
                                                        ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QuerySortedMessages(sortOrder, searchCriteria, encoding, options).ToList();

    public ImapMessageQuery QuerySortedMessages(ImapSortCriteria sortOrder,
                                                ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return QuerySortedMessages(sortOrder,
                                 ImapSearchCriteria.All,
                                 null,
                                 options);
    }

    public ImapMessageQuery QuerySortedMessages(ImapSortCriteria sortOrder,
                                                ImapSearchCriteria searchCriteria,
                                                Encoding encoding = null,
                                                ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (sortOrder == null)
        throw new ArgumentNullException(nameof(sortOrder));
      if (searchCriteria == null)
        throw new ArgumentNullException(nameof(searchCriteria));

      Client.ThrowIfCapabilityUnavailable(sortOrder);

      return new SortedSearchMessageQuery(this,
                                          options,
                                          sortOrder,
                                          searchCriteria,
                                          encoding);
    }

    private class SortedSearchMessageQuery : SearchMessageQueryBase {
      public SortedSearchMessageQuery(ImapOpenMailbox mailbox,
                                   ImapMessageFetchAttributeOptions fetchOptions,
                                   ImapSortCriteria sortOrder,
                                   ImapSearchCriteria searchCriteria,
                                   Encoding encoding)
        : base(mailbox, fetchOptions, searchCriteria, encoding, true)
      {
        this.sortOrder = sortOrder;
      }

      protected override IImapNumberSet GetSequenceOrUidSet()
      {
        Mailbox.ProcessResult(Mailbox.Client.Session.Sort(out var matchedSequenceNumbers,
                                                          sortOrder,
                                                          SearchCriteria,
                                                          Encoding),
                              ThrowIfBadCharset);

        return matchedSequenceNumbers;
      }

      private readonly ImapSortCriteria sortOrder;
    }

    /*
     * GetMessages by SEARCHing with X-GM-RAW
     * https://developers.google.com/gmail/imap_extensions
     * https://support.google.com/mail/answer/7190
     */
    public IReadOnlyList<ImapMessage> GetMessagesGmailSearch(string gmailSearchQuery,
                                                             ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
      => QueryMessagesGmailSearch(gmailSearchQuery, options).ToList();

    public ImapMessageQuery QueryMessagesGmailSearch(string gmailSearchQuery,
                                                     ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (gmailSearchQuery == null)
        throw new ArgumentNullException(nameof(gmailSearchQuery));
      if (gmailSearchQuery.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(gmailSearchQuery));

      Client.ThrowIfCapabilityUnavailable(ImapCapability.GimapGmExt1);

      return QueryMessages(ImapSearchCriteria.XGMRaw(gmailSearchQuery), Encoding.UTF8, options);
    }

    /*
     * GetMessages by THREADing
     */
    public ImapMessageThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                 ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 ImapSearchCriteria.All,
                                 null,
                                 options);
    }

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public ImapMessageThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                 ImapSearchCriteria searchCriteria,
                                                 ImapMessageFetchAttributeOptions options)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 searchCriteria,
                                 null,
                                 options);
    }

    public ImapMessageThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                 ImapSearchCriteria searchCriteria,
                                                 Encoding encoding = null,
                                                 ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (threadingAlgorithm == null)
        throw new ArgumentNullException(nameof(threadingAlgorithm));
      if (searchCriteria == null)
        throw new ArgumentNullException(nameof(searchCriteria));

      Client.ThrowIfCapabilityUnavailable(threadingAlgorithm);

      return (new ThreadedSearchMessageQuery(this, options, threadingAlgorithm, searchCriteria, encoding)).ToThread();
    }

    public Task<ImapMessageThread> GetThreadedMessagesAsync(ImapThreadingAlgorithm threadingAlgorithm,
                                                            ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      return GetThreadedMessagesAsync(threadingAlgorithm,
                                      ImapSearchCriteria.All,
                                      null,
                                      options);
    }

    [Obsolete(ImapClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<ImapMessageThread> GetThreadedMessagesAsync(ImapThreadingAlgorithm threadingAlgorithm,
                                                            ImapSearchCriteria searchCriteria,
                                                            ImapMessageFetchAttributeOptions options)
    {
      return GetThreadedMessagesAsync(threadingAlgorithm,
                                      searchCriteria,
                                      null,
                                      options);
    }

    public Task<ImapMessageThread> GetThreadedMessagesAsync(ImapThreadingAlgorithm threadingAlgorithm,
                                                            ImapSearchCriteria searchCriteria,
                                                            Encoding encoding = null,
                                                            ImapMessageFetchAttributeOptions options = ImapMessageFetchAttributeOptions.Default)
    {
      if (threadingAlgorithm == null)
        throw new ArgumentNullException(nameof(threadingAlgorithm));
      if (searchCriteria == null)
        throw new ArgumentNullException(nameof(searchCriteria));

      Client.ThrowIfCapabilityUnavailable(threadingAlgorithm);

      return (new ThreadedSearchMessageQuery(this, options, threadingAlgorithm, searchCriteria, encoding)).ToThreadAsync();
    }

    private class ThreadedSearchMessageQuery : SearchMessageQueryBase {
      public ThreadedSearchMessageQuery(ImapOpenMailbox mailbox,
                                        ImapMessageFetchAttributeOptions fetchOptions,
                                        ImapThreadingAlgorithm threadingAlgorithm,
                                        ImapSearchCriteria searchCriteria,
                                        Encoding encoding)
        : base(mailbox, fetchOptions, searchCriteria, encoding, false)
      {
        this.threadingAlgorithm = threadingAlgorithm;
      }

      protected override IImapNumberSet GetSequenceOrUidSet()
      {
        Mailbox.ProcessResult(Mailbox.Client.Session.Thread(out threadList,
                                                            threadingAlgorithm,
                                                            SearchCriteria,
                                                            Encoding),
                              ThrowIfBadCharset);

        return threadList.ToSequenceSet();
      }

      public ImapMessageThread ToThread()
      {
        return Yield(ToThread);
      }

      public Task<ImapMessageThread> ToThreadAsync()
      {
        return YieldAsync(ToThread, CancellationToken.None);
      }

      private ImapMessageThread ToThread(IEnumerable<ImapMessage> messages)
      {
        // yield messages first
        var messageSequenceMap = messages.ToDictionary(m => m.Sequence, m => m);

        return new ImapMessageThread(threadList, messageSequenceMap);
      }

      private readonly ImapThreadingAlgorithm threadingAlgorithm;
      private ImapThreadList threadList = null;
    }
  }
}

