// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.Net.Imap4.Urls;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures {
  public class ImapSinglePartBodyStructure : IImapBodyStructure {
    /*
     * IImapUrl impl
     */
    public Uri Url => ImapUriBuilder.BuildUri(uriBuilder);

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      uriBuilder = baseUrl.Clone();
      uriBuilder.Section = Section;

      (this as ImapMessageRfc822BodyStructure)?.EncapsulatedMessageStructure?.SetBaseUrl(baseUrl);
    }

    private ImapUriBuilder uriBuilder = null;

    /*
     * IImapBodyStructure impl
     */
    public IImapBodyStructure ParentStructure { get; internal set; }
    public string Section { get; }
    public MimeType MediaType { get; }
    public bool IsMultiPart => false;
    public virtual bool HasEncapsulatedMessage => false;

    /*
     * basic fields
     */
    public IReadOnlyDictionary<string, string> Parameters { get; }
    public string Id { get; }
    public string Description { get; }
    public string Encoding { get; }
    public long Size { get; }

    /*
     * type TEXT, type MESSAGE
     */
    // the size of the body in text lines(type TEXT)
    // size in text lines of the encapsulated message(type MESSAGE)
    public long LineCount { get; }

    internal ImapSinglePartBodyStructure(string section,
                                         MimeType mediaType,
                                         IReadOnlyDictionary<string, string> parameters,
                                         string id,
                                         string description,
                                         string encoding,
                                         long size,
                                         long lineCount)
    {
      Section         = section ?? throw new ArgumentNullException(nameof(section));
      MediaType       = mediaType ?? throw new ArgumentNullException(nameof(mediaType));
      Parameters      = parameters ?? ReadOnlyDictionary<string, string>.Empty;
      Id              = id;
      Description     = description;
      Encoding        = encoding;
      Size            = size < 0 ? throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(size), size) : size;
      LineCount       = lineCount < 0 ? throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(lineCount), lineCount) : lineCount;
    }

    private protected ImapSinglePartBodyStructure(ImapSinglePartBodyStructure baseStructure)
    {
      if (baseStructure == null)
        throw new ArgumentNullException(nameof(baseStructure));

      Section         = baseStructure.Section;
      MediaType       = baseStructure.MediaType;
      Parameters      = baseStructure.Parameters;
      Id              = baseStructure.Id;
      Description     = baseStructure.Description;
      Encoding        = baseStructure.Encoding;
      Size            = baseStructure.Size;
      LineCount       = baseStructure.LineCount;
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Section), Section),
        (nameof(MediaType), MediaType),
        (nameof(Parameters), Parameters),
        (nameof(Id), Id),
        (nameof(Description), Description),
        (nameof(Encoding), Encoding),
        (nameof(Size), Size),
        (nameof(LineCount), LineCount),
      });
    }
  }
}
