// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes {
  public class ImapMessageBodyAttributes :
    ImapMessageAttributesBase
  {
    public IReadOnlyCollection<string> BodySpecifiers
#if NET45
      // KeyCollection does not implement IReadOnlyCollection<TKey>
      => new ReadOnlyBodySpecifierCollection(bodyData?.Keys);

      private class ReadOnlyBodySpecifierCollection : IReadOnlyCollection<string>, ICollection<string> {
        private readonly ICollection<string> specifiers;

        public ReadOnlyBodySpecifierCollection(ICollection<string> specifiers)
        {
          this.specifiers = specifiers ?? ArrayExtensions.Empty<string>();
        }

        public int Count => specifiers.Count;

        public IEnumerator<string> GetEnumerator() => specifiers.GetEnumerator();
        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() => GetEnumerator();

        /* ICollection<T> */
        public bool IsReadOnly => true;
        public bool Contains(string item) => specifiers.Contains(item); // for Enumerable.Contains()
        public void CopyTo(string[] array, int arrayIndex) => specifiers.CopyTo(array, arrayIndex);

        public void Add(string item) => throw new NotSupportedException("read only");
        public void Clear() => throw new NotSupportedException("read only");
        public bool Remove(string item) => throw new NotSupportedException("read only");
      }
#else
      => (IReadOnlyCollection<string>)bodyData?.Keys ?? Array.Empty<string>();
#endif



    internal ImapMessageBodyAttributes(long sequence)
      : base(sequence)
    {
    }

    internal override ImapMessageDynamicAttributes GetDynamicAttributes() => null;
    internal override ImapMessageStaticAttributes GetStaticAttributes() => null;
    internal override ImapMessageBodyAttributes GetBodyAttributes() => this;

    internal void SetBodyData(string specifier, ImapData bodyText)
    {
      if (bodyData == null)
        bodyData = new Dictionary<string, ImapData>(StringComparer.OrdinalIgnoreCase);

      bodyData[specifier] = bodyText;
    }

    private ImapData GetFirstBodyData() => (bodyData ?? throw new InvalidOperationException("has no body data")).Values.First();
    public void ReadFirstBody(Action<Stream> readBody) => GetFirstBodyData().ReadText(readBody);
    public long GetFirstBodyLength() => GetFirstBodyData().GetTextLength();

    internal Stream GetFirstBody() => GetFirstBodyData().GetTextAsStream();
    internal int GetFirstBody(ref byte[] buffer) => GetFirstBodyData().GetText(ref buffer);

    private ImapData GetBodyData(string specifier)
    {
      if (specifier == null)
        throw new ArgumentNullException(nameof(specifier));

      if (bodyData != null && bodyData.TryGetValue(specifier, out var body))
        return body;
      else
        throw new ArgumentException("no such content", nameof(specifier));
    }

    public void ReadBodyOf(string specifier, Action<Stream> readBody) => GetBodyData(specifier).ReadText(readBody);
    public long GetBodyLengthOf(string specifier) => GetBodyData(specifier).GetTextLength();

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Sequence), Sequence),
        (nameof(BodySpecifiers), BodySpecifiers),
      });
    }

    private Dictionary<string, ImapData> bodyData;
  }
}
