// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.Formats;
using Smdn.Formats.Mime;
using Smdn.Text;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  // 7.4.2. FETCH Response
  //       ENVELOPE
  //         The fields of the envelope structure are in the following
  //         order: date, subject, from, sender, reply-to, to, cc, bcc,
  //         in-reply-to, and message-id.  The date, subject, in-reply-to,
  //         and message-id fields are strings.  The from, sender, reply-to,
  //         to, cc, and bcc fields are parenthesized lists of address
  //         structures.
  public sealed class ImapEnvelope {
    /*
    envelope        = "(" env-date SP env-subject SP env-from SP
                      env-sender SP env-reply-to SP env-to SP env-cc SP
                      env-bcc SP env-in-reply-to SP env-message-id ")"
    env-bcc         = "(" 1*address ")" / nil
    env-cc          = "(" 1*address ")" / nil
    env-date        = nstring
    env-from        = "(" 1*address ")" / nil
    env-in-reply-to = nstring
    env-message-id  = nstring
    env-reply-to    = "(" 1*address ")" / nil
    env-sender      = "(" 1*address ")" / nil
    env-subject     = nstring
    env-to          = "(" 1*address ")" / nil
    */
    public string Date { get; }
    public string Subject { get; }
    public IReadOnlyList<ImapAddress> From { get; }
    public IReadOnlyList<ImapAddress> Sender { get; }
    public IReadOnlyList<ImapAddress> ReplyTo { get; }
    public IReadOnlyList<ImapAddress> To { get; }
    public IReadOnlyList<ImapAddress> Cc { get; }
    public IReadOnlyList<ImapAddress> Bcc { get; }
    public string InReplyTo { get; }
    public string MessageId { get; }

    internal ImapEnvelope(string date,
                          string subject,
                          IReadOnlyList<ImapAddress> @from,
                          IReadOnlyList<ImapAddress> sender,
                          IReadOnlyList<ImapAddress> replyTo,
                          IReadOnlyList<ImapAddress> @to,
                          IReadOnlyList<ImapAddress> cc,
                          IReadOnlyList<ImapAddress> bcc,
                          string inReplyTo,
                          string messageId)
    {
      Date        = date;
      Subject     = subject;
      From        = ToNonNull(@from);
      Sender      = ToNonNull(sender);
      ReplyTo     = ToNonNull(replyTo);
      To          = ToNonNull(@to);
      Cc          = ToNonNull(cc);
      Bcc         = ToNonNull(bcc);
      InReplyTo   = inReplyTo;
      MessageId   = messageId;

      IReadOnlyList<ImapAddress> ToNonNull(IReadOnlyList<ImapAddress> list)
      {
        return list ??
#if NET45
          ArrayExtensions.Empty<ImapAddress>();
#else
          Array.Empty<ImapAddress>();
#endif
      }
    }

    public DateTimeOffset? GetDecodedDate(bool throwIfMalformed = false)
    {
      /*
       * 7.4.2. FETCH Response
       *       ENVELOPE
       *             Note: some servers may return a NIL envelope member in the
       *             "present but empty" case.  Clients SHOULD treat NIL and
       *             empty string as identical.
       */
      try {
        if (string.IsNullOrEmpty(Date))
          return null;

        var dateString = MimeUtils.RemoveHeaderWhiteSpaceAndComment(Date);

        return DateTimeFormat.FromRFC822DateTimeOffsetString(dateString);
      }
      catch (FormatException) {
        if (throwIfMalformed)
          throw;
        else
          return null;
      }
    }

    public string GetDecodedSubject(EncodingSelectionCallback selectFallbackEncoding = null,
                                    bool throwIfMalformedOrUnsupported = false)
    {
      if (string.IsNullOrEmpty(Subject))
        return null;

      return MimeHeaderUtils.GetDecodedValue(Subject,
                                             selectFallbackEncoding,
                                             throwIfMalformedOrUnsupported);
    }

    public override string ToString()
    {
      return StringConversion.ToString(GetType(), new (string, object)[] {
        (nameof(Date), Date),
        (nameof(Subject), Subject),
        (nameof(From), From),
        (nameof(Sender), Sender),
        (nameof(ReplyTo), ReplyTo),
        (nameof(To), To),
        (nameof(Cc), Cc),
        (nameof(Bcc), Bcc),
        (nameof(InReplyTo), InReplyTo),
        (nameof(MessageId), MessageId),
      });
    }
  }
}

