// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  public readonly struct ImapFetchRange : IEquatable<ImapFetchRange>, IFormattable {
    public static readonly ImapFetchRange All = new ImapFetchRange(0L);

    private const long MaxValue = (long)UInt32.MaxValue;

    public long Start { get; }
    public long? Length { get; }

    public bool IsLengthSpecified => Length.HasValue;
    public bool IsPartial => Start != 0L || Length.HasValue;

    public ImapFetchRange(long start)
      : this()
    {
      if (start < 0L || MaxValue < start)
        throw ExceptionUtils.CreateArgumentMustBeInRange(0L, MaxValue, nameof(start), start);

      Start = start;
      Length = null;
    }

    public ImapFetchRange(long start, long length)
      : this()
    {
      if (start < 0L || MaxValue < start)
        throw ExceptionUtils.CreateArgumentMustBeInRange(0L, MaxValue, nameof(start), start);
      if (length <= 0L || MaxValue < length)
        throw ExceptionUtils.CreateArgumentMustBeInRange(1L, MaxValue, nameof(length), length);
      if (MaxValue < start + length)
        throw ExceptionUtils.CreateArgumentMustBeInRange(1L, MaxValue, $"{nameof(start)} + {nameof(length)}", start + length);

      Start = start;
      Length = length;
    }

    public override bool Equals(object obj)
    {
      if (obj is ImapFetchRange range)
        return Equals(range);
      else
        return false;
    }

    public bool Equals(ImapFetchRange other)
    {
      return (this.Start == other.Start && this.Length == other.Length);
    }

    public override int GetHashCode()
    {
      return ToString().GetHashCode();
    }

    public override string ToString()
    {
      return ToString(null, null);
    }

    public string ToString(string format)
    {
      return ToString(format, null);
    }

    public string ToString(string format, IFormatProvider formatProvider)
    {
      string open = "<";
      string close = ">";

      switch (format) {
        case "u": // IMAP-URL style
          open = "/;PARTIAL=";
          close = null;
          break;

        case "f": // fetch-att style
          if (!IsLengthSpecified)
            throw new FormatException("length must be specified");
          break;

        default:
          break;
      }

      var sb = new StringBuilder(64);

      sb.Append(open);
      sb.Append(Start);

      if (IsLengthSpecified) {
        sb.Append('.');
        sb.Append(Length.Value);
      }

      sb.Append(close);

      return sb.ToString();
    }
  }
}
