// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Connection;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Urls;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Protocol.Session {
  partial class ImapSession {
    public static Task<ImapSession> CreateAsync(IImapSessionProfile profile,
                                                SaslClientMechanism authMechanism = null,
                                                ResolveHostAddressCallback resolveHostAddress = null,
                                                UpgradeConnectionStreamCallback createSslStream = null,
                                                CancellationToken cancellationToken = default)
    {
      if (profile == null)
        throw new ArgumentNullException(nameof(profile));

      var authority = profile.Authority;
      var securePort = string.Equals(authority.Scheme, ImapUri.UriSchemeImaps, StringComparison.OrdinalIgnoreCase);

      if (securePort && createSslStream == null)
        throw new ArgumentNullException(nameof(createSslStream));

      if (cancellationToken.IsCancellationRequested) {
#if NET45
        return new Task<ImapSession>(() => null, cancellationToken);
#else
        return Task.FromCanceled<ImapSession>(cancellationToken);
#endif
      }

      return CreateSessionAsyncImpl();

      async Task<ImapSession> CreateSessionAsyncImpl()
      {
        ImapSession s = null;

        try {
          s = new ImapSession();

          s.TransactionTimeout = profile.Timeout;
          s.SendTimeout = profile.SendTimeout;
          s.ReceiveTimeout = profile.ReceiveTimeout;
          s.HandlesReferralAsException = true;

          await s.ConnectAsync(
            hostname: authority.Host,
            port: authority.Port,
            resolveHostAddress: resolveHostAddress,
            createAuthenticatedStream: securePort ? createSslStream : null,
            cancellationToken: cancellationToken
          ).ConfigureAwait(false);

          s.HandlesIncapableAsException = false;
          s.HandlesReferralAsException = false;

          if (s.ServerCapabilities.Count == 0) {
            // cancel before querying server capability
            cancellationToken.ThrowIfCancellationRequested();

            // try querying server capability (ignore error)
            s.Capability();
          }

          if (!s.ServerCapabilities.Contains(ImapCapability.Imap4Rev1))
            throw new ImapIncapableException(ImapCapability.Imap4Rev1);

          if (profile.UseTlsIfAvailable && s.ServerCapabilities.Contains(ImapCapability.StartTls) && !s.IsSecureConnection) {
            // cancel before upgrading connection
            cancellationToken.ThrowIfCancellationRequested();

            var r = s.StartTls(createSslStream, true);

            if (!r.IsSuccess)
              throw new ImapSecureConnectionException(string.Concat("STARTTLS failed: ", r.Description));
            if (!s.ServerCapabilities.Contains(ImapCapability.Imap4Rev1))
              throw new ImapIncapableException(ImapCapability.Imap4Rev1);
          }

          ImapCommandResult authenticationResult = null;

          if (authMechanism == null) {
            authenticationResult = Authenticate(s, profile, cancellationToken);
          }
          else {
            // cancel before authentication
            cancellationToken.ThrowIfCancellationRequested();

            authenticationResult = s.Authenticate(authMechanism,
                                                  true,
                                                  cancellationToken);
          }

          if (authenticationResult == null)
            throw new ImapNoAppropriateAuthMechanismException();
          if (!authenticationResult.IsSuccess)
            throw new ImapAuthenticationException(authenticationResult);
        }
        catch {
          s?.Dispose();
          s = null;
          throw;
        }

        try {
          if (profile.UseDeflateIfAvailable && s.ServerCapabilities.Contains(ImapCapability.CompressDeflate))
            // try compress (ignore error)
            s.Compress(ImapCompressionMechanism.Deflate);
        }
        catch {
          s?.Dispose();
          s = null;
          throw;
        }

        return s;
      }
    }

    private static ImapCommandResult Authenticate(ImapSession session,
                                                  IImapSessionProfile profile,
                                                  CancellationToken cancellationToken)
    {
      var authority = profile.Authority;
      var username = ImapStyleUriParser.GetUser(authority);
      var authMechanism = ImapStyleUriParser.GetAuthType(authority);

      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       *
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */

      // TODO: URLAUTH
      /*
      if (!session.ServerCapabilities.Has(ImapCapability.UrlAuth)
        throw new ImapWebException("URLAUTH incapable", WebExceptionStatus.ProtocolError);
      */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = ImapAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = ImapAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == ImapAuthenticationMechanism.SelectAppropriate) {
        var allowUsingPlainTextMechanism = session.IsSecureConnection || profile.AllowInsecureLogin;

        return AuthenticateWithAppropriateMechanism(session,
                                                    allowUsingPlainTextMechanism,
                                                    profile.Credentials,
                                                    username,
                                                    SaslClientMechanism.GetAvailableMechanisms(profile.UsingSaslMechanisms),
                                                    cancellationToken);
      }
      else if (authMechanism == ImapAuthenticationMechanism.Anonymous) {
        return AuthenticateAsAnonymous(session,
                                       username,
                                       canFallback,
                                       cancellationToken);
      }
      else {
        return AuthenticateWithSuppliedMechanism(session,
                                                 profile.Credentials,
                                                 username,
                                                 authMechanism,
                                                 cancellationToken);
      }
    }

    private static ImapCommandResult AuthenticateWithAppropriateMechanism(ImapSession session,
                                                                          bool allowUsingPlainTextMechanism,
                                                                          ICredentialsByHost credentials,
                                                                          string username,
                                                                          IEnumerable<string> usingSaslMechanisms,
                                                                          CancellationToken cancellationToken)
    {
      ImapCommandResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (!allowUsingPlainTextMechanism && SaslClientMechanism.IsMechanismPlainText(mechanism))
          // disallow plain text mechanism
          continue;

        if (string.Equals(mechanism, SaslMechanisms.Anonymous, StringComparison.OrdinalIgnoreCase))
          // disallow 'ANONYMOUS' mechanism
          continue;

        var authMechanism = ImapAuthenticationMechanism.GetKnownOrCreate(mechanism);

        if (session.IsCapable(authMechanism, throwIfIncapable: false)) {
          cancellationToken.ThrowIfCancellationRequested();

          result = session.Authenticate(authMechanism,
                                        credentials,
                                        username,
                                        reissueCapability: true,
                                        cancellationToken);

          if (result.IsSuccess)
            break;
        }
      }

      if ((result == null || !result.IsSuccess) &&
          allowUsingPlainTextMechanism &&
          !session.ServerCapabilities.Contains(ImapCapability.LoginDisabled)) {
        cancellationToken.ThrowIfCancellationRequested();

        result = session.Login(credentials, username, true);
      }

      return result;
    }

    private static ImapCommandResult AuthenticateWithSuppliedMechanism(ImapSession session,
                                                                       ICredentialsByHost credentials,
                                                                       string username,
                                                                       ImapAuthenticationMechanism authMechanism,
                                                                       CancellationToken cancellationToken)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       */
      cancellationToken.ThrowIfCancellationRequested();

      return session.Authenticate(authMechanism,
                                  credentials,
                                  username,
                                  reissueCapability: true,
                                  cancellationToken);
    }

    private static ImapCommandResult AuthenticateAsAnonymous(ImapSession session,
                                                             string username,
                                                             bool canFallback, 
                                                             CancellationToken cancellationToken)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    If no user name and no authentication mechanism are supplied, the
       *    client MUST authenticate as anonymous to the server.  If the server
       *    advertises AUTH=ANONYMOUS IMAP capability, the client MUST use the
       *    AUTHENTICATE command with ANONYMOUS [ANONYMOUS] SASL mechanism.  If
       *    SASL ANONYMOUS is not available, the (case-insensitive) user name
       *    "anonymous" is used with the "LOGIN" command and the Internet email
       *    address of the end user accessing the resource is supplied as the
       *    password.  The latter option is given in order to provide for
       *    interoperability with deployed servers.
       * 
       *    Note that, as described in RFC 3501, the "LOGIN" command MUST NOT be
       *    used when the IMAP server advertises the LOGINDISABLED capability.
       */
      ImapCommandResult result = null;

      if (string.IsNullOrEmpty(username))
        username = "anonymous@";

      if (session.IsCapable(ImapAuthenticationMechanism.Anonymous, throwIfIncapable: false)) {
        cancellationToken.ThrowIfCancellationRequested();

        // try AUTHENTICATE ANONYUMOUS
        result = session.Authenticate(ImapAuthenticationMechanism.Anonymous,
                                      new NetworkCredential(username, string.Empty),
                                      reissueCapability: true,
                                      cancellationToken: cancellationToken);
      }

      if ((result == null || (!result.IsSuccess && canFallback)) &&
          !session.ServerCapabilities.Contains(ImapCapability.LoginDisabled)) {
        cancellationToken.ThrowIfCancellationRequested();

        // try anonymous LOGIN
        result = session.Login(new NetworkCredential("anonymous", username),
                               null,
                               true);
      }

      return result;
    }
  }
}
