// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Connection;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Protocol.Transmission.Transaction;
using Smdn.Net.Imap4.Urls;
using Smdn.Net.MessageAccessProtocols.Strings;

using Trace = Smdn.Net.Imap4.Client.Trace;

namespace Smdn.Net.Imap4.Protocol.Session {
  // Layers:
  // * Session.ImapSession
  //   Transaction.IImapTransaction
  //   Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public sealed partial class ImapSession : IConnectionInfo, IDisposable {
    string IConnectionInfo.Host { get { CheckDisposed(); return connection.Host; } }
    int IConnectionInfo.Port { get { CheckDisposed(); return connection.Port; } }

    /*
     * properties
     */
    internal int? Id {
      get { return (connection == null) ? (int?)null : connection.Id; }
    }

    public bool HandlesReferralAsException {
      get { return handlesReferralAsException; }
      set { handlesReferralAsException = value; }
    }

    public bool HandlesIncapableAsException {
      get { return handlesIncapableAsException; }
      set { handlesIncapableAsException = value; }
    }

    public int SendTimeout {
      get { CheckDisposed(); return connection.SendTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.SendTimeout = value; }
    }

    public int ReceiveTimeout {
      get { CheckDisposed(); return connection.ReceiveTimeout; }
      set { CheckDisposed(); RejectTransactionProceeding(); connection.ReceiveTimeout = value; }
    }

    public bool IsSecureConnection {
      get { CheckDisposed(); return connection.IsSecurePortConnection || connection.IsSecureConnection; }
    }

    public int TransactionTimeout {
      get { CheckDisposed(); return transactionTimeout; }
      set
      {
        CheckDisposed();
        RejectTransactionProceeding();

        if (value < -1)
          throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(TransactionTimeout), value);
        transactionTimeout = value;
      }
    }

    public bool IsTransactionProceeding {
      get
      {
        if (Monitor.TryEnter(transactionLockObject, 0)) {
          Monitor.Exit(transactionLockObject);
          return false;
        }
        else {
          return true;
        }
      }
    }

    /*
     * construction / destruction
     */
    public ImapSession()
    {
    }

    public ImapSession(string hostname,
                       int port = -1,
                       int transactionTimeout = DefaultTransactionTimeout,
                       int sendTimeout = DefaultSendTimeout,
                       int receiveTimeout = DefaultReceiveTimeout,
                       bool handlesReferralAsException = false,
                       UpgradeConnectionStreamCallback createAuthenticatedStream = null)
    {
      if (transactionTimeout < -1)
        throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(transactionTimeout), transactionTimeout);

      this.handlesReferralAsException = handlesReferralAsException;
      this.transactionTimeout = transactionTimeout;
      this.SendTimeout = sendTimeout;
      this.ReceiveTimeout = receiveTimeout;

      using (var cts = new CancellationTokenSource(transactionTimeout)) {
        ConnectAsync(
          hostname: hostname,
          port: port,
          createAuthenticatedStream: createAuthenticatedStream,
          cancellationToken: cts.Token
        ).GetAwaiter().GetResult();
      }
    }

    public override string ToString()
    {
      CheckDisposed();

      return string.Format("{{Connection={0}, State={1}}}",
                           connection,
                           State);
    }

    private const int DefaultTransactionTimeout = Timeout.Infinite;
    private const int DefaultSendTimeout = Timeout.Infinite;
    private const int DefaultReceiveTimeout = Timeout.Infinite;

    private bool handlesReferralAsException;
    private bool handlesIncapableAsException = false;
    private int transactionTimeout = DefaultTransactionTimeout;
    private readonly object transactionLockObject = new object();

    /*
     * session state properties
     */
    public ImapSessionState State { get; private set; } = ImapSessionState.NotConnected;

    public bool IsIdling {
      get { CheckDisposed(); return connection.IsIdling; }
    }

    public ImapMailboxStatus SelectedMailbox {
      get { CheckDisposed(); return selectedMailbox; }
    }

    internal bool UpdateSelectedMailboxSizeAndStatus {
      get { return updateSelectedMailboxSizeAndStatus; }
      set { updateSelectedMailboxSizeAndStatus = value; }
    }

    public Uri Authority {
      get { CheckDisposed(); return authority == null ? null : authority.Uri; }
    }

    internal ImapUriBuilder AuthorityBuilder {
      get { return authority; }
    }

    public IReadOnlyStringSet<ImapCapability> ServerCapabilities {
      get { CheckDisposed(); return serverCapabilities; }
    }

    public IReadOnlyDictionary<string, string> HierarchyDelimiters {
      get { CheckDisposed(); return hierarchyDelimiters; }
    }

    public ImapNamespace Namespaces {
      get { CheckDisposed(); return namespaces; }
    }

    public IReadOnlyDictionary<string, string> ServerID {
      get { CheckDisposed(); return serverID; }
    }

    public string SelectedLanguage {
      get { CheckDisposed(); return selectedLanguage; }
    }

    public ImapCollationAlgorithm ActiveComparator {
      get { CheckDisposed(); return activeComparator; }
    }

    private ImapConnection connection = new ImapConnection();
    private ImapMailboxManager mailboxManager = null;
    private ImapMailboxStatus selectedMailbox = null;
    private bool updateSelectedMailboxSizeAndStatus = true;
    private ImapUriBuilder authority = new ImapUriBuilder();
    private IReadOnlyStringSet<ImapCapability> serverCapabilities = initialServerCapability;
    private Dictionary<string, string> hierarchyDelimiters = new Dictionary<string, string>(StringComparer.Ordinal);
    private ImapNamespace namespaces = new ImapNamespace();
    private string selectedLanguage = null;
    private ImapCollationAlgorithm activeComparator = ImapCollationAlgorithm.Default;

    private static readonly IReadOnlyStringSet<ImapCapability> initialServerCapability = new ImapCapabilitySet(new[] {
      ImapCapability.Imap4Rev1 /* required by GreetingTransaction */
    }, asReadOnly: true);

    private IReadOnlyDictionary<string, string> serverID = Smdn.Collections.ReadOnlyDictionary<string, string>.Empty;

    /*
     * transaction methods : connect/disconnect
     */
    public Task ConnectAsync(string hostname,
                             int port = -1,
                             ResolveHostAddressCallback resolveHostAddress = null,
                             UpgradeConnectionStreamCallback createAuthenticatedStream = null,
                             CancellationToken cancellationToken = default)
    {
      CheckDisposed();

      if (connection.IsConnected)
        throw new InvalidOperationException("already connected");

      return ConnectAsyncImpl();

      async Task ConnectAsyncImpl()
      {
        TraceInfo("connecting");

        await connection.ConnectAsync(
          hostname: hostname,
          port: port,
          resolveHostAddress: resolveHostAddress,
          createAuthenticatedStream: createAuthenticatedStream,
          cancellationToken: cancellationToken
        ).ConfigureAwait(false);

        TraceInfo("connected");

        commandTag = 0;

        var result = ProcessGreeting();

        if (result.IsFinalResponseCondition(ImapResponseCondition.PreAuth)) {
          UpdateAuthority(null, null);
          TransitStateTo(ImapSessionState.Authenticated);
        }
        else {
          TransitStateTo(ImapSessionState.NotAuthenticated);
        }

        // 7.1. Server Responses - Status Responses
        //       CAPABILITY
        //          Followed by a list of capabilities.  This can appear in the
        //          initial OK or PREAUTH response to transmit an initial
        //          capabilities list.  This makes it unnecessary for a client to
        //          send a separate CAPABILITY command if it recognizes this
        //          response.
        var capabilityResponseCode = result.GetStatusResponseOfCode(ImapResponseCode.Capability);

        if (capabilityResponseCode == null)
          // clear server capabilities
          SetServerCapabilities(null);
        else
          SetServerCapabilities(ImapResponseTextParser.FromCapability(capabilityResponseCode.ResponseText));
      }

      ImapCommandResult ProcessGreeting()
      {
        using (var t = new GreetingTransaction(connection, this)) {
          try {
            ProcessTransaction(t);
          }
          catch (ImapMalformedResponseException ex) {
            throw new ImapConnectionException("session establishment failed (malformed or invalid response)", ex);
          }
          catch (ImapUnexpectedResponseException ex) {
            throw new ImapConnectionException("session establishment failed (unexpected response)", ex);
          }
          catch (TimeoutException ex) {
            throw new ImapConnectionException("connection timed out", ex);
          }

          if (!t.Result.IsFinalResponseCondition(ImapResponseCondition.Bye))
            return t.Result;

          /*
           * BYE response
           */
          var refferalResponseCode = t.Result.GetStatusResponseOfCode(ImapResponseCode.Referral);

          if (refferalResponseCode != null) {
            // RFC 2221 IMAP4 Login Referrals
            // http://tools.ietf.org/html/rfc2221
            // 4.2. BYE at connection startup referral
            //    An IMAP4 server MAY respond with an untagged BYE and a REFERRAL
            //    response code that contains an IMAP URL to a home server if it is not
            //    willing to accept connections and wishes to direct the client to
            //    another IMAP4 server.
            var referToUri = ImapResponseTextParser.FromReferral(refferalResponseCode.ResponseText)[0];

            if (handlesReferralAsException)
              throw new ImapLoginReferralException(string.Format("try another server: '{0}'", refferalResponseCode.ResponseText.Text),
                                                   referToUri);

            TraceInfo("login referral: '{0}'", refferalResponseCode.ResponseText.Text);
            TraceInfo(string.Concat("  try to connect to ", referToUri));
          }

          throw new ImapConnectionException(string.Concat("session establishment failed: ", t.Result.Description));
        }
      }
    }

    /// <summary>disconnects session</summary>
    /// <remarks>this method sends CLOSE and LOGOUT before disconnect the session.</remarks>
    public void Disconnect(bool logout = true)
    {
      CheckDisposed();

      if (logout) {
        if (State == ImapSessionState.Selected)
          Close();
        if (State == ImapSessionState.Authenticated)
          Logout();
      }

      CloseConnection();
    }

    public void Dispose() => CloseConnection();

    private void CloseConnection()
    {
      if (connection == null)
        return;

      try {
        TraceInfo("disconnecting");

        connection.Close();

        TraceInfo("disconnected");
      }
      finally {
        TransitStateTo(ImapSessionState.Disconnected);

        connection = null;
      }
    }

    /*
     * transaction methods : generic use, exntension, etc.
     */

    /// <summary>sends generic/undefined command</summary>
    public ImapCommandResult GenericCommand(string command, params ImapString[] arguments)
    {
      return GenericCommand(command, out _, arguments);
    }

    /// <summary>sends generic/undefined command</summary>
    public ImapCommandResult GenericCommand(string command, out IReadOnlyList<ImapServerDataResponse> dataResponses, params ImapString[] arguments)
    {
      RejectNonConnectedState();

      if (command == null)
        throw new ArgumentNullException(nameof(command));
      if (command.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(command));

      dataResponses = null;

      using (var t = new GenericCommandTransaction(connection, this, command)) {
        if (arguments != null && 0 < arguments.Length)
          t.RequestArguments["arguments"] = ImapStringList.Create(arguments);

        if (ProcessTransaction(t).IsSuccess)
          dataResponses = t.Result.ReceivedResponses.OfType<ImapServerDataResponse>().ToList();

        return t.Result;
      }
    }

    /*
     * methods for internal state management
     */
    internal void SetServerCapabilities(IReadOnlyStringSet<ImapCapability> newCapabilities)
    {
      if (newCapabilities == null)
        serverCapabilities = ImapCapabilitySet.ReadOnlyEmpty;
      else
        serverCapabilities = newCapabilities;
    }

    private void TransitStateTo(ImapSessionState newState)
    {
      if (newState == State)
        return;

      switch (newState) {
        case ImapSessionState.NotConnected:
          throw new InvalidOperationException("it is not allowed to transit to non-connected state");

        case ImapSessionState.NotAuthenticated:
          UpdateAuthority(null, null);
          if (State == ImapSessionState.Selected || State == ImapSessionState.Authenticated) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
            TraceInfo("logged out");
          }
          TraceInfo("now in non-authenticated state");
          break;

        case ImapSessionState.Authenticated:
          TraceInfo("now in authenticated state");
          if (State == ImapSessionState.NotAuthenticated || State == ImapSessionState.NotConnected) {
            mailboxManager = new ImapMailboxManager(this);
            TraceInfo("logged in");
          }
          break;

        case ImapSessionState.Selected:
          TraceInfo("now in selected state (selected '{0}')", selectedMailbox.Name);
          break;

        case ImapSessionState.Disconnected:
          authority = null;
          selectedMailbox = null;
          if (mailboxManager != null) {
            mailboxManager.DetachFromSession();
            mailboxManager = null;
          }
          TraceInfo("disconnected");
          break;
      }

      State = newState;
    }

    /*
     * methods for checking and validating session status
     */
    private void CheckDisposed()
    {
      if (State == ImapSessionState.Disconnected)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private void RejectNonConnectedState()
    {
      CheckDisposed();

      if (State == ImapSessionState.NotConnected)
        throw new InvalidOperationException("session is not connected yet");
    }

    private void RejectNonAuthenticatedState()
    {
      RejectNonConnectedState();

      if (State == ImapSessionState.NotAuthenticated)
        throw new ImapProtocolViolationException("session is not in authenticated state");
    }

    private void RejectNonSelectedState()
    {
      RejectNonConnectedState();

      if (State != ImapSessionState.Selected)
        throw new ImapProtocolViolationException("session is not in selected state");
    }

    private ImapMailboxStatus ValidateMailboxRelationship(ImapMailboxStatus mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException(nameof(mailbox));

      if (mailbox.Session != this)
        throw new ArgumentException("mailbox is not attached to current session", nameof(mailbox));

      return mailbox;
    }

    private void RejectTransactionProceeding()
    {
      if (IsTransactionProceeding)
        throw new InvalidOperationException("another transaction is proceeding");
    }

    private bool IsCapable(ImapCapability capability, bool throwIfIncapable)
    {
      var isCapable = serverCapabilities.Contains(capability);

      if (throwIfIncapable && !isCapable)
        throw new ImapIncapableException(capability);

      return isCapable;
    }

    internal bool IsCapable(IImapExtension extension, bool throwIfIncapable)
    {
      var isCapable = serverCapabilities.IsCapable(extension);

      if (throwIfIncapable && !isCapable)
        throw new ImapIncapableException(extension.RequiredCapabilities);

      return isCapable;
    }

    /*
     * tracing
     */
    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string message)
    {
      TraceInfo(true, message);
    }

    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(bool condition, string message)
    {
      if (condition)
        Trace.Log(this, message);
    }

    [System.Diagnostics.Conditional("TRACE")]
    private void TraceInfo(string format, params object[] arguments)
    {
      Trace.Log(this, format, arguments);
    }
  }
}
