// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Formats.ModifiedBase64;

namespace Smdn.Net.Imap4.Protocol.Strings {
  // string types:
  //   ImapString
  //     => handles 'string'
  //     ImapQuotedString
  //       => handles 'quoted'
  //       ImapMailboxNameString (internal)
  //         => handles 'mailbox name'
  //     ImapLiteralString
  //       => handles 'literal'
  //     ImapNilString
  //       => handles 'NIL'
  //     ImapStringList
  //       => list of ImapString
  //       ImapParenthesizedString
  //         => handles 'parenthesized list'
  //     ImapStringEnum
  //       => string enumeration type
  //     ImapCombinableDataItem
  //       => combinable data item type

  internal sealed class ImapMailboxNameString : ImapString, IImapStringValue {
    public static ImapString CreateListMailboxName(string mailboxName) => new ImapMailboxNameString(mailboxName, false, ImapOctets.IndexOfNonListChar);
    public static ImapString CreateReferenceName(string mailboxName) => new ImapMailboxNameString(mailboxName, false, ImapOctets.IndexOfNonAstringChar);
    public static ImapString CreateMailboxName(string mailboxName) => new ImapMailboxNameString(mailboxName, true, ImapOctets.IndexOfNonAstringChar);

    private readonly string originalMailboxName;
    private readonly IImapStringValue mailboxName;

    string IImapStringValue.Value => originalMailboxName;
    byte[] IImapStringValue.GetBytes() => mailboxName.GetBytes();

    private ImapMailboxNameString(string mailboxName, bool rejectEmpty, Func<string, int> indexOfEscapeChar)
    {
      this.originalMailboxName = mailboxName ?? throw new ArgumentNullException(nameof(mailboxName));

      if (mailboxName.Length == 0) {
        if (rejectEmpty)
          throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(mailboxName));

        this.mailboxName = (IImapStringValue)ImapQuotedString.Empty;
      }
      else {
        var encoded = ModifiedUTF7.Encode(mailboxName);

        if (0 <= indexOfEscapeChar(encoded))
          this.mailboxName = (IImapStringValue)ImapQuotedString.Create(encoded);
        else
          this.mailboxName = new PlainString(encoded);
      }
    }
  }
}
