// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.Strings {
  // string types:
  //   ImapString
  //     => handles 'string'
  //     ImapPreformattedString
  //       => handles pre-encoded/pre-formatted octets
  //     ImapQuotedString
  //       => handles 'quoted'
  //       ImapMailboxNameString (internal)
  //         => handles 'mailbox name'
  //     IImapLiteralString
  //       => handles 'literal'
  //     ImapNilString
  //       => handles 'NIL'
  //     ImapStringList
  //       => list of ImapString
  //       ImapParenthesizedString
  //         => handles 'parenthesized list'
  //     ImapStringEnum
  //       => string enumeration type
  //     ImapCombinableDataItem
  //       => combinable data item type

#if SERIALIZATION
  [Serializable]
#endif
  public abstract class ImapString {
    internal static ImapString CreateFromRawBytes(byte[] bytes)
    {
      return new PlainString(bytes ?? throw new ArgumentNullException(nameof(bytes)));
    }

    internal static IReadOnlyCollection<ImapString> ToList(string val)
    {
      return Singleton.CreateList(new PlainString(val));
    }

    internal static IReadOnlyCollection<ImapString> ToList(string first, params string[] subsequence)
    {
      return subsequence.Convert(s => new PlainString(s)).Prepend(new PlainString(first));
    }

    internal static void Traverse(IEnumerable<ImapString> strings, Action<ImapString> action)
    {
      foreach (var str in strings) {
        Traverse(str, action);
      }
    }

    internal static void Traverse(ImapString str, Action<ImapString> action)
    {
      action(str);

      if (str is IImapStringList list) {
        foreach (var item in list.Items) {
          Traverse(item, action);
        }
      }
    }

    /*
     * concrete classes
     */
    internal sealed class PlainString : ImapString, IImapStringValue {
      private readonly byte[] bytes;

      string IImapStringValue.Value => OctetEncoding.SevenBits.GetString(bytes);
      byte[] IImapStringValue.GetBytes() => bytes;

      public PlainString(string value)
        : this(value == null ? throw new ArgumentNullException(nameof(value)) : OctetEncoding.SevenBits.GetBytes(value))
      {
      }

      public PlainString(byte[] value)
      {
        bytes = value;
      }
    }

    /*
     * instance members
     */
    public static implicit operator ImapString(string str)
    {
      return str == null ? null : new PlainString(str);
    }

    public override string ToString()
    {
      if (this is IImapStringList list)
        return ImapStringList.ToString(list);
      if (this is IImapStringValue value)
        return value.Value;

      return GetType().FullName;
    }
  }
}
