// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections;
using System.Collections.Generic;

using Smdn.Collections;

namespace Smdn.Net.Imap4.Protocol.Strings {
  // string types:
  //   ImapString
  //     => handles 'string'
  //     ImapQuotedString
  //       => handles 'quoted'
  //       ImapMailboxNameString (internal)
  //         => handles 'mailbox name'
  //     IImapLiteralString
  //       => handles 'literal'
  //     ImapNilString
  //       => handles 'NIL'
  //     ImapStringList
  //       => list of ImapString
  //       ImapParenthesizedString
  //         => handles 'parenthesized list'
  //     ImapStringEnum
  //       => string enumeration type
  //     ImapCombinableDataItem
  //       => combinable data item type

  public class ImapStringList :
    ImapString,
    IImapStringList,
    IReadOnlyCollection<ImapString>
  {
    public static ImapStringList Create(string value)
    {
      return new ImapStringList(value, parenthesize: false);
    }

    public static ImapStringList Create(IReadOnlyCollection<string> values)
    {
      return new ImapStringList(values, parenthesize: false);
    }

    public static ImapStringList Create(IReadOnlyCollection<ImapString> values)
    {
      return new ImapStringList(values, parenthesize: false);
    }

    public static ImapStringList CreateParenthesized(string value)
    {
      return new ImapStringList(value, parenthesize: true);
    }

    public static ImapStringList CreateParenthesized(IReadOnlyCollection<string> values)
    {
      return new ImapStringList(values, parenthesize: true);
    }

    public static ImapStringList CreateParenthesized(IReadOnlyCollection<ImapString> values)
    {
      return new ImapStringList(values, parenthesize: true);
    }

    /*
     * instance members
     */
    bool IImapStringList.IsParenthesized => parenthesize;
    IEnumerable<ImapString> IImapStringList.Items => values;
    int IReadOnlyCollection<ImapString>.Count => values.Count;

    private ImapStringList(string value, bool parenthesize)
      : this(Singleton.CreateList((ImapString)value), parenthesize)
    {
    }

    private ImapStringList(IReadOnlyCollection<string> values, bool parenthesize)
      : this(values?.ConvertAll(s => (ImapString)s), parenthesize)
    {
    }

    private protected ImapStringList(IReadOnlyCollection<ImapString> values, bool parenthesize)
    {
      this.values = values ?? throw new ArgumentNullException(nameof(values));
      this.parenthesize = parenthesize;
    }

    public IEnumerator<ImapString> GetEnumerator()
    {
      return (values as IEnumerable<ImapString>).GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator()
    {
      return values.GetEnumerator();
    }

    internal static string ToString(IImapStringList list)
    {
      return string.Concat(
        list.IsParenthesized ? "(" : string.Empty,
        string.Join(" ", list.Items),
        list.IsParenthesized ? ")" : string.Empty
      );
    }

    public override string ToString() => ToString(this);

    private readonly IReadOnlyCollection<ImapString> values;
    private readonly bool parenthesize;
  }
}
