// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Threading;

using Smdn.Formats;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal sealed class AuthenticateTransaction : ImapTransactionBase {
    public AuthenticateTransaction(ImapConnection connection,
                                   IImapCommandTagGenerator commandTagGenerator,
                                   SaslClientMechanism authMechanism,
                                   bool sendInitialResponse,
                                   CancellationToken cancellationToken)
      : base(connection, commandTagGenerator)
    {
      this.authMechanism = authMechanism ?? throw new ArgumentNullException(nameof(authMechanism));
      this.sendInitialResponse = sendInitialResponse;
      this.cancellationToken = cancellationToken;
    }

    // 6.2.2. AUTHENTICATE Command
    //    Arguments:  authentication mechanism name
    //    Responses:  continuation data can be requested
    //    Result:     OK - authenticate completed, now in authenticated state
    //                NO - authenticate failure: unsupported authentication
    //                     mechanism, credentials rejected
    //                BAD - command unknown or arguments invalid,
    //                     authentication exchange cancelled
    internal class SaslInitialExchangeFailedException : InvalidOperationException {
      public SaslInitialExchangeFailedException(SaslClientMechanism causedSaslMechanism)
        : base($"SASL client did not generate initial-response (name: {causedSaslMechanism.Name}, status: {causedSaslMechanism.ExchangeStatus})")
      {
      }
    }

    protected override string GetCommand() => "AUTHENTICATE";

    protected override ImapString[] GetCommandArguments()
    {
      authMechanism.Initialize();

      /*
       * http://tools.ietf.org/html/rfc4959
       * RFC 4959 - IMAP Extension for Simple Authentication and Security Layer (SASL) Initial Client Response
       * 3. IMAP Changes to the IMAP AUTHENTICATE Command
       *    Note: support and use of the initial client response is optional for
       *    both clients and servers.  Servers that implement this extension MUST
       *    support clients that omit the initial client response, and clients
       *    that implement this extension MUST NOT send an initial client
       *    response to servers that do not advertise the SASL-IR capability.
       */
      byte[] initialClientResponse = null;

      if (sendInitialResponse && authMechanism.ClientFirst) {
        if (authMechanism.GetInitialResponse(out initialClientResponse) == SaslExchangeStatus.Failed ||
            initialClientResponse == null)
          throw new SaslInitialExchangeFailedException(authMechanism);
      }

      if (initialClientResponse == null)
        return new[] { RequestArguments["authentication mechanism name"] };
      else
        /*
         * http://tools.ietf.org/html/rfc4959
         * RFC 4959 - IMAP Extension for Simple Authentication and Security Layer (SASL) Initial Client Response
         * 3. IMAP Changes to the IMAP AUTHENTICATE Command
         *    This extension adds an optional second argument to the AUTHENTICATE
         *    command that is defined in Section 6.2.2 of [RFC3501].  If this
         *    second argument is present, it represents the contents of the
         *    "initial client response" defined in Section 5.1 of [RFC4422].
         */
        return new[] { RequestArguments["authentication mechanism name"], ImapString.CreateFromRawBytes(Base64.Encode(initialClientResponse)) };
    }

    private static readonly ImapString cancelRequest = ImapString.CreateFromRawBytes(new[] { ImapOctets.Asterisk });

    protected override void OnCommandContinuationRequestReceived(ImapCommandContinuationRequest continuationRequest)
    {
      // The authentication protocol exchange consists of a series of
      // server challenges and client responses that are specific to the
      // authentication mechanism.  A server challenge consists of a
      // command continuation request response with the "+" token followed
      // by a BASE64 encoded string.  The client response consists of a
      // single line consisting of a BASE64 encoded string.  If the client
      // wishes to cancel an authentication exchange, it issues a line
      // consisting of a single "*".  If the server receives such a
      // response, it MUST reject the AUTHENTICATE command by sending a
      // tagged BAD response.
      if (cancellationToken.IsCancellationRequested) {
        SendContinuation(cancelRequest);
        return;
      }

      var serverChallenge = Base64.Decode(continuationRequest.Text.Segment.Array,
                                          continuationRequest.Text.Segment.Offset,
                                          continuationRequest.Text.Segment.Count);
      var status = authMechanism.Exchange(serverChallenge, out var clientResponse);

      if (status == SaslExchangeStatus.Failed || clientResponse == null)
        SendContinuation(cancelRequest);
      else
        SendContinuation(ImapString.CreateFromRawBytes(Base64.Encode(clientResponse)));
    }

    private readonly SaslClientMechanism authMechanism;
    private readonly bool sendInitialResponse;
    private readonly CancellationToken cancellationToken;
  }
}
