// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal class CopyMoveTransaction : ImapTransactionBase<ImapCopiedUidSet>, IImapExtension {
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get {
        if (!commandTrueForCopy)
          yield return ImapCapability.Move;
      }
    }

    internal CopyMoveTransaction(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator, bool commandTrueForCopy, bool uid)
      : base(connection, commandTagGenerator)
    {
      this.commandTrueForCopy = commandTrueForCopy;
      this.command = commandTrueForCopy ? (uid ? "UID COPY" : "COPY") : (uid ? "UID MOVE" : "MOVE");
    }

    // 6.4.7. COPY Command
    //    Arguments:  sequence set
    //                mailbox name
    //    Responses:  no specific responses for this command
    //    Result:     OK - copy completed
    //                NO - copy error: can't copy those messages or to that
    //                     name
    //                BAD - command unknown or arguments invalid

    // http://tools.ietf.org/html/rfc6851
    // 3.1. MOVE Command
    //    Arguments: sequence set
    //               mailbox name
    //    Responses: no specific responses for this command
    //    Result: OK - move completed
    //            NO - move error: can't move those messages or to that name
    //            BAD - command unknown or arguments invalid
    private readonly bool commandTrueForCopy;
    private readonly string command;

    protected override string GetCommand() => command;

    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("mailbox name") ||
          !RequestArguments.ContainsKey("sequence set"))
        throw CreateRequestArgumentsException("arguments 'mailbox name' and 'sequence set' must be setted");
#endif

      return new[] {
        RequestArguments["sequence set"],
        RequestArguments["mailbox name"],
      };
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok &&
          tagged.ResponseText.Code == ImapResponseCode.CopyUid)
        Finish(tagged, ImapResponseTextParser.FromCopyUid(tagged.ResponseText));
      else
        base.OnTaggedStatusResponseReceived(tagged);
    }
  }
}
