// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal sealed class FetchTransaction<TMessageAttributes> :
    FetchStoreTransactionBase<TMessageAttributes>,
    IImapExtension
    where TMessageAttributes : ImapMessageAttributesBase
  {
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get
      {
        if (fetchModifiersCapabilityRequirement != null)
          yield return fetchModifiersCapabilityRequirement;
      }
    }

    public FetchTransaction(ImapConnection connection,
                            IImapCommandTagGenerator commandTagGenerator,
                            IImapMessageAttributeCollection<TMessageAttributes> messageAttributeCollection,
                            bool uid,
                            ImapCapability fetchModifiersCapabilityRequirement)
      : base(connection, commandTagGenerator, messageAttributeCollection)
    {
      this.command = uid ? "UID FETCH" : "FETCH";
      this.fetchModifiersCapabilityRequirement = fetchModifiersCapabilityRequirement;
    }

    // RFC 4466 - Collected Extensions to IMAP4 ABNF
    // http://tools.ietf.org/html/rfc4466
    // 2.4. Extensions to FETCH and UID FETCH Commands
    //    Arguments:  sequence set
    //                message data item names or macro
    //                OPTIONAL fetch modifiers
    //    Responses:  untagged responses: FETCH
    //    Result:     OK - fetch completed
    //                NO - fetch error: cannot fetch that data
    //                BAD - command unknown or arguments invalid
    private readonly string command;

    protected override string GetCommand() => command;

    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("sequence set") ||
          !RequestArguments.ContainsKey("message data item names or macro"))
        throw CreateRequestArgumentsException("arguments 'sequence set' and 'message data item names or macro' must be setted");
#endif

      if (RequestArguments.TryGetValue("fetch modifiers", out var fetchModifiers))
        return new[] {
          RequestArguments["sequence set"],
          RequestArguments["message data item names or macro"],
          fetchModifiers,
        };
      else
        return new[] {
          RequestArguments["sequence set"],
          RequestArguments["message data item names or macro"],
        };
    }

    private /*readonly*/ ImapCapability fetchModifiersCapabilityRequirement;
  }
}
