// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  /*
   * RFC 2087 - IMAP4 QUOTA extension
   * http://tools.ietf.org/html/rfc2087
   */
  internal sealed class GetQuotaRootTransaction : ImapTransactionBase<IReadOnlyDictionary<string, IReadOnlyList<ImapQuota>>>, IImapExtension {
    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get { yield return ImapCapability.Quota; }
    }

    public GetQuotaRootTransaction(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator)
      : base(connection, commandTagGenerator)
    {
    }

    /*
     * 4.3. GETQUOTAROOT Command
     *    Arguments:  mailbox name
     * 
     *    Data:       untagged responses: QUOTAROOT, QUOTA
     * 
     *    Result:     OK - getquota completed
     *                NO - getquota error: no such mailbox, permission denied
     *                BAD - command unknown or arguments invalid
     */
    protected override string GetCommand() => "GETQUOTAROOT";

    protected override ImapString[] GetCommandArguments()
    {
#if DEBUG
      if (!RequestArguments.ContainsKey("mailbox name"))
        throw CreateRequestArgumentsException("arguments 'mailbox name' must be setted");
#endif

      return new[] { RequestArguments["mailbox name"] };
    }

    protected override void OnDataResponseReceived(ImapServerDataResponse data)
    {
      if (data.Type == ImapServerDataResponseType.QuotaRoot) {
        quotaRoots.Add(ImapServerDataResponseParser.FromQuotaRoot(data, out var quotaRootNames), quotaRootNames);
      }
      else if (data.Type == ImapServerDataResponseType.Quota) {
        quotas.Add(ImapServerDataResponseParser.FromQuota(data));
      }

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok) {
        var result = new Dictionary<string, IReadOnlyList<ImapQuota>>(StringComparer.OrdinalIgnoreCase);

        foreach (var quotaRoot in quotaRoots) {
          var rootedQuotas = new List<ImapQuota>();

          foreach (var quota in quotas.ToArray()) {
            foreach (var root in quotaRoot.Value) {
              if (string.Equals(quota.Root, root, StringComparison.OrdinalIgnoreCase)) {
                rootedQuotas.Add(quota);
                quotas.Remove(quota);
              }
            }
          }

          result.Add(quotaRoot.Key, rootedQuotas);
        }

        Finish(tagged, result);
      }
      else {
        base.OnTaggedStatusResponseReceived(tagged);
      }
    }

    private Dictionary<string, IReadOnlyList<string>> quotaRoots = new Dictionary<string, IReadOnlyList<string>>(StringComparer.OrdinalIgnoreCase);
    private List<ImapQuota> quotas = new List<ImapQuota>();
  }
}

