// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission.Transaction {
  internal abstract class SearchTransactionBase : ImapTransactionBase<ImapMatchedSequenceSet>, IImapExtension {
    protected virtual ImapCapability RequiredCapability {
      get { return resultSpecifierCapabilityRequirement; }
    }

    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get
      {
        if (RequiredCapability != null)
          yield return RequiredCapability;
      }
    }

    internal bool Uid { get; }

    protected SearchTransactionBase(ImapConnection connection, IImapCommandTagGenerator commandTagGenerator, bool uid, ImapCapability resultSpecifierCapabilityRequirement)
      : base(connection, commandTagGenerator)
    {
      Uid = uid;
      this.resultSpecifierCapabilityRequirement = resultSpecifierCapabilityRequirement;
    }

    // RFC 4466 - Collected Extensions to IMAP4 ABNF
    // http://tools.ietf.org/html/rfc4466
    // 2.6. Extensions to SEARCH Command
    // 2.6.1. Extended SEARCH Command
    //    Arguments:  OPTIONAL result specifier
    //                OPTIONAL [CHARSET] specification
    //                searching criteria (one or more)
    //    Responses:  REQUIRED untagged response: SEARCH (*)
    //    Result:     OK - search completed
    //                NO - search error: cannot search that [CHARSET] or
    //                     criteria
    //                BAD - command unknown or arguments invalid

    // http://tools.ietf.org/html/rfc5256
    // BASE.6.4.SORT. SORT Command
    //    Arguments:  sort program
    //                charset specification
    //                searching criteria (one or more)
    //    Data:       untagged responses: SORT
    //    Result:     OK - sort completed
    //                NO - sort error: can't sort that charset or
    //                     criteria
    //                BAD - command unknown or arguments invalid
    protected override void OnDataResponseReceived(ImapServerDataResponse data)
    {
      // 7.2.5 SEARCH Response
      //   Contents:   zero or more numbers
      //   Example:    S: * SEARCH 2 3 6
      if (this is SearchTransaction && data.Type == ImapServerDataResponseType.Search)
        messages = ImapServerDataResponseParser.FromSearch(data, Uid);
      // http://tools.ietf.org/html/rfc5256
      // BASE.7.2.SORT. SORT Response
      //   Data:       zero or more numbers
      //   Example:    S: * SORT 2 3 6
      else if (this is SortTransaction && data.Type == ImapServerDataResponseType.Sort)
        messages = ImapServerDataResponseParser.FromSort(data, Uid);
      // http://tools.ietf.org/html/rfc4466
      // 2.6.2. ESEARCH untagged response
      //    Contents:   one or more search-return-data pairs
      else if (data.Type == ImapServerDataResponseType.ESearch)
        messages = ImapServerDataResponseParser.FromESearch(data);

      base.OnDataResponseReceived(data);
    }

    protected override void OnTaggedStatusResponseReceived(ImapTaggedStatusResponse tagged)
    {
      if (tagged.Condition == ImapResponseCondition.Ok)
        Finish(tagged, messages);
      else
        base.OnTaggedStatusResponseReceived(tagged);
    }

    private /*readonly*/ ImapCapability resultSpecifierCapabilityRequirement;
    private ImapMatchedSequenceSet messages = null;
  }
}
