// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net.Sockets;
using System.Threading;

using Smdn.IO.Streams.LineOriented;
using Smdn.Net.Connection;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  // Layers:
  //   Session.ImapSession
  //   Transaction.IImapTransaction
  // * Protocol.ImapConnection/Protocol.ImapCommand/Protocol.ImapResponse
  //   (IMAP4)

  public sealed class ImapConnection : ConnectionBase {
    /*
     * class members
     */
    public static TraceSource ConnectionTraceSource { get; } = new TraceSource("IMAP");

    /*
     * properties
     */
    protected override TraceSource TraceSource => ConnectionTraceSource;

    public bool IsIdling {
      get { ThrowIfNotConnected(); return isIdling; }
      internal set { isIdling = value; }
    }

    /*
     * RFC 5255 - Internet Message Access Protocol Internationalization
     * http://tools.ietf.org/html/rfc5255
     */
    internal bool ReceiveResponseAsUTF8 {
      get { return receiver.ReceiveResponseAsUTF8; }
      set { receiver.ReceiveResponseAsUTF8 = value; }
    }

    protected override int GetDefaultPort(bool asSecurePortConnection)
      => asSecurePortConnection ? ImapDefaultPorts.Imaps : ImapDefaultPorts.Imap;

    protected override void OnConnectionStreamChanged(Stream stream)
    {
      var bufferedStream = new LineOrientedBufferedStream(stream);

      receiver = new ImapResponseReceiver(bufferedStream);
      sender = new ImapCommandSender(bufferedStream);
    }

    protected override Exception CreateConnectException(string message, Exception baseException)
      => new ImapConnectionException(message, baseException);

    protected override Exception CreateUpgradeToSecureConnectionException(string message, Exception baseException)
      => new ImapSecureConnectionException(message, baseException);

    protected override Exception CreateUpgradeConnectionException(string message, Exception baseException)
      => new ImapUpgradeConnectionException(message, baseException);

#region "seinding / receiving"
    public class CommandContinuationContext {
      internal ImapCommand Command { get; }
      internal bool Sending { get; set; }

      internal CommandContinuationContext(ImapCommand command)
      {
        Command = command;
      }
    }

    public CommandContinuationContext SendCommand(ImapCommand command)
    {
      return SendCommand(new CommandContinuationContext(command));
    }

    public CommandContinuationContext SendCommand(CommandContinuationContext context)
    {
      ThrowIfNotConnected();

      try {
        if (context.Sending) {
          sender.SendRemainder();
        }
        else {
          sender.Send(context.Command);
          context.Sending = true;
        }

        if (sender.IsCommandContinuing)
          return context;
        else
          return null;
      }
      catch (SocketException ex) {
        throw CreateSocketErrorException(ex, false);
      }
      catch (IOException ex) {
        var socketException = ex.InnerException as SocketException;

        if (socketException == null)
          // unexpected IO error
          throw;
        else
          throw CreateSocketErrorException(socketException, false);
      }
    }

    public ImapResponse TryReceiveResponse()
    {
      ThrowIfNotConnected();

      // return queued response
      if (0 < responseQueue.Count)
        return responseQueue.Dequeue();

      // receive and parse response
      for (;;) {
        try {
          var response = receiver.ReceiveResponse();

          if (response == null)
            break;

          responseQueue.Enqueue(response);

          if (!receiver.IsResponseContinuing)
            break;
        }
        catch (SocketException ex) {
          throw CreateSocketErrorException(ex, true);
        }
        catch (IOException ex) {
          var socketException = ex.InnerException as SocketException;

          if (socketException == null)
            // unexpected IO error
            throw;

          var exp = CreateSocketErrorException(socketException, true);

          if (IsIdling && exp is TimeoutException) {
            // expected timeout
            Smdn.Net.Imap4.Client.Trace.Verbose("idling");
            continue;
          }
          else {
            throw exp;
          }
        }
      }

      // return received response
      if (0 < responseQueue.Count)
        return responseQueue.Dequeue();

      return null;
    }

    private Exception CreateSocketErrorException(SocketException ex, bool receiving)
    {
      switch (ex.SocketErrorCode) {
        case SocketError.WouldBlock:
        case SocketError.TimedOut:
          // timeout error
          return new TimeoutException(receiving ? "receive timeout" : "send timeout", ex);

        default:
          // unexpected socket exception
          return new ImapConnectionException(string.Format("unexpected socket error ({0})", ex.SocketErrorCode),
                                             ex);
      }
    }
#endregion

    private bool isIdling = false;
    private ImapResponseReceiver receiver = null;
    private ImapCommandSender sender = null;
    private readonly Queue<ImapResponse> responseQueue = new Queue<ImapResponse>();
  }
}
