// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.IO.Streams;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  // handles '4. Data Formats'
#if SERIALIZATION
  [Serializable]
#endif
  public abstract class ImapData
#if SERIALIZATION
    : ISerializable
#endif
  {
    // 4.1. Atom, 4.2. Number, 4.3. String
    internal static ImapData CreateTextData(ByteString text) => new ImapTextData(text);
    internal static ImapData CreateTextData(ChunkedMemoryStream textStream) => new ImapTextStreamData(textStream);

    // 4.4. Parenthesized List
    internal static ImapData CreateListData(IReadOnlyList<ImapData> list) => new ImapListData(list);

    // 4.5. NIL
    internal static ImapData Nil { get; } = new ImapNilData();

    /*
     * instance members
     */
    public abstract ImapDataFormat Format { get; }
    public virtual IReadOnlyList<ImapData> List => throw CreateNotListException();

    private ImapData()
    {
    }

#if SERIALIZATION
    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
    }
#endif

    public virtual string GetTextAsString() => throw CreateNotTextException();
    public virtual ByteString GetTextAsByteString() => throw CreateNotTextException();
    public virtual void ReadText(Action<Stream> read) => throw CreateNotTextException();
    public virtual long GetTextLength() => throw CreateNotTextException();

    internal virtual Stream GetTextAsStream() => throw CreateNotTextException();
    internal virtual int GetText(ref byte[] buffer) => throw CreateNotTextException();

    private Exception CreateNotListException() => new InvalidOperationException("data is not a list");
    private Exception CreateNotTextException() => new InvalidOperationException("data is not a text");

    /*
     * data classes
     */
#if SERIALIZATION
    [Serializable]
#endif
    private class ImapNilData : ImapData {
      public override ImapDataFormat Format => ImapDataFormat.Nil;

      public ImapNilData()
      {
      }

#if SERIALIZATION
      private ImapNilData(SerializationInfo info, StreamingContext context)
      {
      }
#endif

      public override string ToString() => "{NIL}";
    }

#if SERIALIZATION
    [Serializable]
#endif
    private sealed class ImapListData : ImapData {
      public override ImapDataFormat Format => ImapDataFormat.List;
      public override IReadOnlyList<ImapData> List { get; }

      public ImapListData(IReadOnlyList<ImapData> list)
      {
        List = list;
      }

#if SERIALIZATION
      private ImapListData(SerializationInfo info, StreamingContext context)
      {
        List = (ImapData[])info.GetValue("list", typeof(ImapData[]));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue("list", List.ToArray());
      }
#endif

      public override string ToString()
      {
        return string.Format("{{List:{0}}}", string.Join(", ", List.Select(d => d.ToString())));
      }
    }

#if SERIALIZATION
    [Serializable]
#endif
    private class ImapTextData : ImapData {
      public override ImapDataFormat Format => ImapDataFormat.Text;
      private readonly ByteString text;

      public ImapTextData(ByteString text)
      {
        this.text = text;
      }

#if SERIALIZATION
      private ImapTextData(SerializationInfo info, StreamingContext context)
      {
        var t = (byte[])info.GetValue("text", typeof(byte[]));

        text = ByteString.CreateImmutable(t);
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue("text", text.ToArray());
      }
#endif

      public override string GetTextAsString() => text.ToString();
      public override ByteString GetTextAsByteString() => text;

      public override void ReadText(Action<Stream> read)
      {
        if (read == null)
          throw new ArgumentNullException(nameof(read));

        using (var stream = new MemoryStream(text.Segment.Array,
                                             text.Segment.Offset,
                                             text.Segment.Count,
                                             writable: false)) {
          read(stream);
        }
      }

      public override long GetTextLength() => text.Length;

      internal override Stream GetTextAsStream()
      {
        return new MemoryStream(text.Segment.Array,
                                text.Segment.Offset,
                                text.Segment.Count,
                                false);
      }

      internal override int GetText(ref byte[] buffer)
      {
        if (buffer == null || buffer.Length < text.Length)
          buffer = new byte[text.Length];

        Buffer.BlockCopy(text.Segment.Array, text.Segment.Offset,
                         buffer, 0, text.Length);

        return text.Length;
      }

      public override string ToString()
      {
        return string.Format("{{Text:{0}}}", text);
      }
    }

#if SERIALIZATION
    [Serializable]
#endif
    private class ImapTextStreamData : ImapData {
      public override ImapDataFormat Format => ImapDataFormat.Text;
      private readonly ChunkedMemoryStream textStream;

      public ImapTextStreamData(ChunkedMemoryStream textStream)
      {
        this.textStream = textStream;
      }

#if SERIALIZATION
      private ImapTextStreamData(SerializationInfo info, StreamingContext context)
      {
        var t = (byte[])info.GetValue("text", typeof(byte[])); // XXX: this would be allocated on LOH

        using (var s = new MemoryStream(t, false)) {
          textStream = new ChunkedMemoryStream();

          s.CopyTo(textStream);
        }
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue("text", textStream.ToArray()); // XXX: this would be allocated on LOH
      }
#endif

      public override string GetTextAsString() => ByteString.CreateImmutable(textStream.ToArray()).ToString(); // XXX: this would be allocated on LOH
      public override ByteString GetTextAsByteString() => ByteString.CreateImmutable(textStream.ToArray()); // XXX: this would be allocated on LOH
      public override long GetTextLength() => textStream.Length;

      public override void ReadText(Action<Stream> read)
      {
        if (read == null)
          throw new ArgumentNullException(nameof(read));

        textStream.Position = 0L;

        using (var stream = new NonClosingStream(textStream, writable: false)) {
          read(stream);
        }
      }

      internal override Stream GetTextAsStream()
      {
        textStream.Position = 0L;

        return new NonClosingStream(textStream, writable: false);
      }

      internal override int GetText(ref byte[] buffer)
      {
        if (buffer == null || buffer.Length < textStream.Length)
          buffer = new byte[(int)textStream.Length];

        textStream.Position = 0L;

        return textStream.Read(buffer, 0, buffer.Length);
      }

      public override string ToString()
      {
        return string.Format("{{Text:<Stream ({0} octets)>}}", textStream.Length);
      }
    }
  }
}
