// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Text;

using Smdn.Collections;
using Smdn.Formats.PercentEncodings;
using Smdn.IO.Streams.LineOriented;
using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Urls;
using Smdn.Text;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol {
  partial class ImapSearchCriteria : IImapUrlSearchQuery {
    public static ImapSearchCriteria FromUri(Uri uri)
    {
      return FromUri(uri, true, false, false, out _, out _);
    }

    public static ImapSearchCriteria FromUri(Uri uri, bool splitCharset, out bool containsLiteral, out string charset)
    {
      return FromUri(uri, true, false, splitCharset, out containsLiteral, out charset);
    }

    private static readonly ByteString charsetSpecification = ByteString.CreateImmutable("CHARSET ");

    private static ImapSearchCriteria FromUri(Uri uri,
                                              bool convertLiteral,
                                              bool synchronizedLiteral,
                                              bool splitCharset,
                                              out bool containsLiteral,
                                              out string charset)
    {
      if (uri == null)
        throw new ArgumentNullException(nameof(uri));

      containsLiteral = false;
      charset = null;

      var q = uri.Query;

      if (q.Length == 0)
        return null;
      else if (q.Length == 1) // '?'
        return new ImapSearchCriteria(ToList(string.Empty));

      /*
       * http://tools.ietf.org/html/rfc5092
       * RFC 5092 - IMAP URL Scheme
       * 
       *    Note that quoted strings and non-synchronizing literals [LITERAL+]
       *    are allowed in the <enc-search> content; however, synchronizing
       *    literals are not allowed, as their presence would effectively mean
       *    that the agent interpreting IMAP URLs needs to parse an <enc-search>
       *    content, find all synchronizing literals, and perform proper command
       *    continuation request handling (see Sections 4.3 and 7 of [IMAP4]).
       */
      var query = PercentEncoding.Decode(q.Substring(1), false);
      var len = query.Length;
      var convertedQuery = new ByteStringBuilder(len);
      var queryStartIndex = 0;

      if (splitCharset) {
        var queryString = ByteString.CreateImmutable(query);

        if (queryString.StartsWithIgnoreCase(charsetSpecification)) {
          // CHARSET<SP>astring<SP>
          var posEndOfCharset = queryString.IndexOf(Ascii.Octets.SP, charsetSpecification.Length);

          if (posEndOfCharset < 0) {
            throw new ArgumentException("search criteria contains invalid charset specification", nameof(uri));
          }
          else {
            charset = queryString.ToString(charsetSpecification.Length,
                                           posEndOfCharset - charsetSpecification.Length);

            queryStartIndex = posEndOfCharset + 1;
            len = query.Length;
          }
        }
      }

      for (var i = queryStartIndex; i < len;) {
        if (query[i] == ImapOctets.DQuote) {
          /*
           * quoted
           */
          var start = i;

          for (;;) {
            if (++i == len)
              throw new ArgumentException("search criteria contains unclosed quoted string", nameof(uri));

            if (query[i] == ImapOctets.DQuote) {
              break;
            }
            else if (query[i] == ImapOctets.BackSlash) {
              if (++i == len || !(query[i] == ImapOctets.DQuote || query[i] == ImapOctets.BackSlash))
                throw new ArgumentException("search criteria contains invalid quoted string", nameof(uri));
            }
          }

          i++;

          convertedQuery.Append(query, start, i - start);
        }
        else if (query[i] == ImapOctets.OpenBrace) {
          /*
           * literal
           */
          var start = i;
          var isLiteralSynchronizing = false;
          var literalLength = 0;

          for (;;) {
            if (++i == len)
              throw new ArgumentException("search criteria contains incomplete literal", nameof(uri));

            if (Ascii.Octets.IsDecimalNumber(query[i])) {
              try {
                literalLength = checked(literalLength * 10 + (query[i] - 0x30 /* '0' */));
              }
              catch (OverflowException) {
                throw new ArgumentException("search criteria contains too large literal", nameof(uri));
              }
            }
            else if (query[i] == ImapOctets.CloseBrace) {
              // {xxx}
              isLiteralSynchronizing = true;
              break;
            }
            else if (query[i] == ImapOctets.Plus) {
              // {xxx+}
              if (++i == len || query[i] != ImapOctets.CloseBrace)
                throw new ArgumentException("search criteria contains incomplete non-synchronized literal", nameof(uri));
              isLiteralSynchronizing = false;
              break;
            }
            else {
              throw new ArgumentException("search criteria contains invalid literal", nameof(uri));
            }
          }

          if (++i == len || query[i] != Ascii.Octets.CR)
            throw new ArgumentException("search criteria contains incomplete literal (CR not found)", nameof(uri));
          if (++i == len || query[i] != Ascii.Octets.LF)
            throw new ArgumentException("search criteria contains incomplete literal (LF not found)", nameof(uri));
          if (++i == len && 0 < literalLength)
            throw new ArgumentException("search criteria contains incomplete literal (unexpected EOL)", nameof(uri));

          containsLiteral = true;

          if (convertLiteral) {
            if (synchronizedLiteral)
              convertedQuery.Append(string.Concat("{", literalLength, "}\x0d\x0a"));
            else
              convertedQuery.Append(string.Concat("{", literalLength, "+}\x0d\x0a"));

            convertedQuery.Append(query, i, literalLength);
          }
          else {
            if (synchronizedLiteral != isLiteralSynchronizing)
              throw new ArgumentException(synchronizedLiteral
                                          ? "search criteria contains non-synchronizing literal"
                                          : "search criteria contains synchronizing literal",
                                          nameof(uri));

            convertedQuery.Append(query, start, i - start + literalLength);
          }

          i += literalLength;
        }
        else {
          convertedQuery.Append(query[i++]);
        }
      }

      return new ImapSearchCriteria(Singleton.CreateList(CreateFromRawBytes(convertedQuery.ToByteArray())));
    }

    byte[] IImapUrlSearchQuery.GetEncodedQuery(Encoding charset, out bool charsetSpecified)
    {
      charsetSpecified = (charset != null && SetCharset(charset));

      Traverse(this, s => {
        if (s is ImapLiteralString literal)
          literal.Options = ImapLiteralOptions.NonSynchronizing;
      });

      return SearchCriteriaEncoder.Encode(this);
    }

    private class SearchCriteriaEncoder : ImapSender {
      public static byte[] Encode(ImapSearchCriteria criteria)
      {
        using (var stream = new MemoryStream()) {
          new SearchCriteriaEncoder(stream).EncodeCriteria(criteria);

          var ret = new byte[stream.Length - 2 /*CRLF*/];

#if NETFRAMEWORK || NETSTANDARD2_0
          Buffer.BlockCopy(stream.GetBuffer(), 0, ret, 0, ret.Length);
#else
          if (stream.TryGetBuffer(out var buffer))
            Buffer.BlockCopy(buffer.Array, buffer.Offset, ret, 0, ret.Length);
          else
            throw new ImapException("can't retrieve buffer");
#endif

          return ret;
        }
      }

      private SearchCriteriaEncoder(Stream stream)
        : base(new LineOrientedBufferedStream(stream))
      {
      }

      private void EncodeCriteria(ImapSearchCriteria criteria)
      {
        Enqueue(true, Singleton.CreateList(criteria));

        Send();
      }
    }
  }
}
