// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Linq;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // enum types:
  //   ImapStringEnum : ImapString
  //     => handles string constants
  //     ImapAuthenticationMechanism
  //       => handles authentication mechanisms
  //     ImapCapability
  //       => handles capability constants
  //     ImapCompressionMechanism
  //       => handles COMPRESS extension compression mechanism
  //     ImapMailboxFlag
  //       => handles mailbox flags
  //     ImapMessageFlag
  //       => handles system flags and keywords
  //     ImapThreadingAlgorithm
  //       => handles THREAD extension threading algorithm
  //     Protocol.ImapResponseCode
  //       => handles response codes
  //   * Protocol.ImapDataResponseType
  //       => handles server response types

#if SERIALIZATION
  [Serializable]
#endif
  public sealed class ImapServerDataResponseType : ImapEnumValue, IEquatable<ImapServerDataResponseType> {
    private static readonly ImapEnumValueSet<ImapServerDataResponseType> allTypes;
    public static IReadOnlyStringSet<ImapServerDataResponseType> AllTypes => allTypes;

    public static readonly ImapServerDataResponseType InvalidOrUnknown = new ImapServerDataResponseType("[?]");

    /*
     * RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
     * http://tools.ietf.org/html/rfc3501
     */
    // 7.2. Server Responses - Server and Mailbox Status
    // 7.2.1. CAPABILITY Response
    public static readonly ImapServerDataResponseType Capability  = new ImapServerDataResponseType("CAPABILITY");
    // 7.2.2. LIST Response
    public static readonly ImapServerDataResponseType List        = new ImapServerDataResponseType("LIST");
    // 7.2.3. LSUB Response
    public static readonly ImapServerDataResponseType Lsub        = new ImapServerDataResponseType("LSUB");
    // 7.2.4 STATUS Response
    public static readonly ImapServerDataResponseType Status      = new ImapServerDataResponseType("STATUS");
    // 7.2.5. SEARCH Response
    public static readonly ImapServerDataResponseType Search      = new ImapServerDataResponseType("SEARCH");
    // 7.2.6. FLAGS Response
    public static readonly ImapServerDataResponseType Flags       = new ImapServerDataResponseType("FLAGS");

    // 7.3. Server Responses - Mailbox Size
    // 7.3.1. EXISTS Response
    public static readonly ImapServerDataResponseType Exists      = new ImapServerDataResponseType("EXISTS");
    // 7.3.2. RECENT Response
    public static readonly ImapServerDataResponseType Recent      = new ImapServerDataResponseType("RECENT");

    // 7.4. Server Responses - Message Status
    // 7.4.1. EXPUNGE Response
    public static readonly ImapServerDataResponseType Expunge     = new ImapServerDataResponseType("EXPUNGE");
    // 7.4.2. FETCH Response
    public static readonly ImapServerDataResponseType Fetch       = new ImapServerDataResponseType("FETCH");

    internal static readonly IReadOnlyStringSet<ImapServerDataResponseType> SizeStatusTypes;
    internal static readonly IReadOnlyStringSet<ImapServerDataResponseType> AllTypesExceptSizeStatus;

    /*
     * RFC 2087 - IMAP4 QUOTA extension
     * http://tools.ietf.org/html/rfc2087
     */
    // 5.1. QUOTA Response
    public static readonly ImapServerDataResponseType Quota = new ImapServerDataResponseType("QUOTA");

    // 5.2. QUOTAROOT Response
    public static readonly ImapServerDataResponseType QuotaRoot = new ImapServerDataResponseType("QUOTAROOT");

    /*
     * RFC 2342 IMAP4 Namespace
     * http://tools.ietf.org/html/rfc2342
     */
    // 5. NAMESPACE Command
    public static readonly ImapServerDataResponseType Namespace = new ImapServerDataResponseType("NAMESPACE");

    /*
     * RFC 2971 IMAP4 ID extension
     * http://tools.ietf.org/html/rfc2971
     */
    // 3.2. ID Response
    public static readonly ImapServerDataResponseType Id = new ImapServerDataResponseType("ID");

    /*
     * RFC 4466 - Collected Extensions to IMAP4 ABNF
     * http://tools.ietf.org/html/rfc4466
     */
    // 2.6.2. ESEARCH untagged response
    public static readonly ImapServerDataResponseType ESearch = new ImapServerDataResponseType("ESEARCH");

    /*
     * RFC 5161 The IMAP ENABLE Extension
     * http://tools.ietf.org/html/rfc5161
     */
    // 3.2. The ENABLED Response
    public static readonly ImapServerDataResponseType Enabled = new ImapServerDataResponseType("ENABLED");

    /*
     * RFC 5256 Internet Message Access Protocol - SORT and THREAD Extensions
     * http://tools.ietf.org/html/rfc5256
     */
    // BASE.7.2.THREAD. THREAD Response
    // BASE.7.2.SORT. SORT Response
    public static readonly ImapServerDataResponseType Thread  = new ImapServerDataResponseType("THREAD");
    public static readonly ImapServerDataResponseType Sort    = new ImapServerDataResponseType("SORT");

    /*
     * RFC 5255 - Internet Message Access Protocol Internationalization
     * http://tools.ietf.org/html/rfc5255
     */
    // 3.3. LANGUAGE Response
    public static readonly ImapServerDataResponseType Language = new ImapServerDataResponseType("LANGUAGE");
    // 4.8. COMPARATOR Response
    public static readonly ImapServerDataResponseType Comparator = new ImapServerDataResponseType("COMPARATOR");

    /*
     * RFC 5464 - The IMAP METADATA Extension
     * http://tools.ietf.org/html/rfc5464
     */
    // 4.4. METADATA Response
    public static readonly ImapServerDataResponseType Metadata = new ImapServerDataResponseType("METADATA");

    static ImapServerDataResponseType()
    {
      allTypes = CreateDefinedConstantsSet<ImapServerDataResponseType>();

      SizeStatusTypes = new ImapEnumValueSet<ImapServerDataResponseType>(asReadOnly: true, values: new[] {
        Exists,
        Recent,
        Expunge,
        Fetch,
      });

      AllTypesExceptSizeStatus = new ImapEnumValueSet<ImapServerDataResponseType>(asReadOnly: true, values: allTypes.Except(SizeStatusTypes));
    }

    internal ImapServerDataResponseType(string type)
      : base(type)
    {
      if (type == null)
        throw new ArgumentNullException(nameof(type));
    }

    #region "equatable"
    public static bool operator ==(ImapServerDataResponseType x, ImapServerDataResponseType y) =>  Equals(x, y);
    public static bool operator !=(ImapServerDataResponseType x, ImapServerDataResponseType y) => !Equals(x, y);
    public bool Equals(ImapServerDataResponseType other) => Equals(this, other);
    public override bool Equals(object obj) => obj is ImapServerDataResponseType other ? Equals(this, other) : base.Equals(obj);
    public override int GetHashCode() => base.GetHashCode();
    #endregion
  }
}
