// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  // enum types:
  //   ImapStringEnum : ImapString
  //     => handles string constants
  //     ImapAuthenticationMechanism
  //       => handles authentication mechanisms
  //     ImapCapability
  //       => handles capability constants
  //     ImapCompressionMechanism
  //       => handles COMPRESS extension compression mechanism
  //     ImapMailboxFlag
  //       => handles mailbox flags
  //     ImapMessageFlag
  //       => handles system flags and keywords
  //   * ImapThreadingAlgorithm
  //       => handles THREAD extension threading algorithm
  //     Protocol.ImapResponseCode
  //       => handles response codes
  //     Protocol.ImapDataResponseType
  //       => handles server response types

  // http://tools.ietf.org/html/rfc5256
  // BASE.6.4.THREAD. THREAD Command
#if SERIALIZATION
  [Serializable]
#endif
  public sealed class ImapThreadingAlgorithm : ImapEnumValue, IImapExtension, IEquatable<ImapThreadingAlgorithm> {
    private static readonly ImapEnumValueSet<ImapThreadingAlgorithm> allAlgorithms;
    public static IReadOnlyStringSet<ImapThreadingAlgorithm> AllAlgorithms => allAlgorithms;

    // ORDEREDSUBJECT
    //    The ORDEREDSUBJECT threading algorithm is also referred to as
    //    "poor man's threading".  The searched messages are sorted by
    //    base subject and then by the sent date.  The messages are then
    //    split into separate threads, with each thread containing
    //    messages with the same base subject text.  Finally, the threads
    //    are sorted by the sent date of the first message in the thread.
    public static readonly ImapThreadingAlgorithm OrderedSubject = new ImapThreadingAlgorithm("ORDEREDSUBJECT", ImapCapability.ThreadOrderedSubject);

    // REFERENCES
    //    The REFERENCES threading algorithm threads the searched
    //    messages by grouping them together in parent/child
    //    relationships based on which messages are replies to others.
    //    The parent/child relationships are built using two methods:
    //    reconstructing a message's ancestry using the references
    //    contained within it; and checking the original (not base)
    //    subject of a message to see if it is a reply to (or forward of)
    //    another message.
    public static readonly ImapThreadingAlgorithm References = new ImapThreadingAlgorithm("REFERENCES", ImapCapability.ThreadReferences);

    static ImapThreadingAlgorithm()
    {
      allAlgorithms = CreateDefinedConstantsSet<ImapThreadingAlgorithm>();
    }

    IEnumerable<ImapCapability> IImapExtension.RequiredCapabilities {
      get
      {
        if (requiredCapability != null)
          yield return requiredCapability;
      }
    }

    private ImapThreadingAlgorithm(string threadingAlgorithm, ImapCapability requiredCapability)
      : base(threadingAlgorithm)
    {
      this.requiredCapability = requiredCapability;
    }

    #region "equatable"
    public static bool operator ==(ImapThreadingAlgorithm x, ImapThreadingAlgorithm y) =>  Equals(x, y);
    public static bool operator !=(ImapThreadingAlgorithm x, ImapThreadingAlgorithm y) => !Equals(x, y);
    public bool Equals(ImapThreadingAlgorithm other) => Equals(this, other);
    public override bool Equals(object obj) => obj is ImapThreadingAlgorithm other ? Equals(this, other) : base.Equals(obj);
    public override int GetHashCode() => base.GetHashCode();
    #endregion

    private readonly ImapCapability requiredCapability;
  }
}
