using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures;
using Smdn.Net.Imap4.Protocol.Session;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  public class ImapMessageBodyBaseTestsBase {
    protected class MessageBodyGenerator {
      public static readonly int DefaultBlockSize = ImapFetchMessageBodyStream.DefaultFetchBlockSize;

      public long MessageUid { get; }

      private readonly IReadOnlyList<string> messageBodyBlocks;

      private readonly int totalOctets;
      public int TotalOctets => totalOctets;

      private readonly IEnumerable<string> newFlags;

      public bool RespondRfc822Size = true;

      private int currentMessageBodyBlockIndex = 0;
      public int BlockIndex => currentMessageBodyBlockIndex;
      public int BlockCount => messageBodyBlocks.Count;

      public int GetBlockLengthOf(int index) => messageBodyBlocks[index].Length;

      public MessageBodyGenerator(long messageUid, string messageBody, IEnumerable<string> newFlags = null, bool respondRfc822Size = true)
      {
        this.MessageUid = messageUid;
        this.newFlags = newFlags;
        this.messageBodyBlocks = GenerateMessageBodyBlocks(messageBody, out this.totalOctets);
        this.RespondRfc822Size = respondRfc822Size;
      }

      private static IReadOnlyList<string> GenerateMessageBodyBlocks(string messageBody, out int totalOctets)
      {
        messageBody = messageBody.Replace("\r\n", "\n").Replace("\n", "\r\n");

        totalOctets = OctetEncoding.SevenBits.GetByteCount(messageBody);

        var blocks = new List<string>();

        for (var offset = 0; ;) {
          if (ImapFetchMessageBodyStream.DefaultFetchBlockSize < messageBody.Length - offset) {
            blocks.Add(messageBody.Substring(offset, ImapFetchMessageBodyStream.DefaultFetchBlockSize));
            offset += ImapFetchMessageBodyStream.DefaultFetchBlockSize;
          }
          else {
            blocks.Add(messageBody.Substring(offset));
            break;
          }
        }

        return blocks;
      }

      public static MessageBodyGenerator Create(long messageUid, int blockCount, IEnumerable<string> newFlags = null, bool respondRfc822Size = true)
      {
        var totalOctets = ImapFetchMessageBodyStream.DefaultFetchBlockSize * (blockCount - 1) + 32;

        return new MessageBodyGenerator(messageUid, totalOctets, setFlagsByFetch: newFlags, respondRfc822Size: respondRfc822Size);
      }

      public MessageBodyGenerator(long messageUid, int totalOctets, IEnumerable<string> setFlagsByFetch = null, bool respondRfc822Size = true)
      {
        this.MessageUid = messageUid;
        this.totalOctets = totalOctets;
        this.newFlags = setFlagsByFetch;
        this.messageBodyBlocks = GenerateMessageBodyBlocks(totalOctets);
        this.RespondRfc822Size = respondRfc822Size;
      }

      private static IReadOnlyList<string> GenerateMessageBodyBlocks(int totalOctets)
      {
        var blocks = new List<string>();

        for (var offset = 0; ;) {
          if (ImapFetchMessageBodyStream.DefaultFetchBlockSize < totalOctets - offset) {
            var marker = $"[block #{blocks.Count}]\r\n";
            blocks.Add(marker + GenerateBlock('*', ImapFetchMessageBodyStream.DefaultFetchBlockSize - marker.Length));
            offset += ImapFetchMessageBodyStream.DefaultFetchBlockSize;
          }
          else {
            var block = $"[block #{blocks.Count} (final)]\r\n" + GenerateBlock('.', ImapFetchMessageBodyStream.DefaultFetchBlockSize);
            blocks.Add(block.Substring(0, totalOctets - offset));
            break;
          }
        }

        return blocks;

        string GenerateBlock(char fillCharacter, int length)
        {
          const string newline = "\r\n";
          const int lineLength = 120;
          var builder = new StringBuilder(capacity: length, maxCapacity: length);

          for (; ; ) {
            var fillLength = Math.Min(lineLength, length) - newline.Length;

            if (fillLength < 0) {
              builder.Append('_', newline.Length + fillLength);
              break;
            }

            builder.Append(fillCharacter, fillLength);
            builder.Append(newline);

            length -= (fillLength + newline.Length);

            if (length == 0)
              break;
          }

          return builder.ToString();
        }
      }

      public string GetMessageBody()
        => string.Concat(messageBodyBlocks);

      public Stream GetMessageBodyStream()
        => new MemoryStream(OctetEncoding.SevenBits.GetBytes(GetMessageBody()), false);

      public void EmitAllResponse(ImapPseudoServer server)
      {
        for (; currentMessageBodyBlockIndex < messageBodyBlocks.Count;) {
          EmitNextResponse(server);
        }
      }

      public void EmitNextResponse(ImapPseudoServer server, Action actionBeforeRespond = null, Action actionAfterRespond = null)
      {
        if (currentMessageBodyBlockIndex == messageBodyBlocks.Count)
          throw new InvalidOperationException();

        var messageBodyBlock = messageBodyBlocks[currentMessageBodyBlockIndex];

        var rfc822Size = (currentMessageBodyBlockIndex == 0 && RespondRfc822Size) ? $"RFC822.SIZE {totalOctets} " : string.Empty;
        var flags = newFlags == null ? string.Empty : $"FLAGS ({string.Join(" ",newFlags)}) ";
        var responseText = $"* {MessageUid} FETCH ({rfc822Size}{flags}BODY[] {{{messageBodyBlock.Length}}}\r\n{messageBodyBlock})\r\n" +
                            "$tag OK done\r\n";

        server.EnqueueTaggedResponse($"UID FETCH {MessageUid}",
                                     responseText: responseText,
                                     actionBeforeRespond: actionBeforeRespond,
                                     actionAfterRespond: actionAfterRespond);

        currentMessageBodyBlockIndex++;
      }
    }

    private void WithMessage(long messageUid,
                             string fetchResponseBodyStructure,
                             string fetchResponseEnvelope,
                             DateTimeOffset fetchResponseInternalDate,
                             int fetchResponseRfc822Size,
                             IEnumerable<string> initialFlags,
                             Action<ImapPseudoServer, ImapMessage> testAction)
    {

      var fetchResp =
        $"* 1 FETCH (UID {messageUid}" +
        $" BODYSTRUCTURE ({fetchResponseBodyStructure})" +
        $" ENVELOPE {fetchResponseEnvelope}" +
        $" INTERNALDATE {ImapDateTimeFormat.ToDateTimeString(fetchResponseInternalDate)}" +
        $" RFC822.SIZE {fetchResponseRfc822Size}" +
        $" FLAGS ({string.Join(" ", initialFlags ?? Enumerable.Empty<string>())}))\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp, ImapMessageFetchAttributeOptions.AllAttributes, (server, message) => {
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.MessagePart);

        testAction(server, message);
      });
    }

    protected void WithMessageSingleTextPlain(Encoding contentTypeCharset,
                                              string messageBody,
                                              IEnumerable<string> initialFlags = null,
                                              Action<ImapPseudoServer, ImapMessage, MessageBodyGenerator> testAction = null)
    {
      WithMessageSingleTextPlain(
        contentTypeCharset: contentTypeCharset,
        messageBodyGenerator: new MessageBodyGenerator(messageUid: 1L, messageBody),
        initialFlags: initialFlags,
        testAction: testAction
      );
    }

    protected void WithMessageSingleTextPlain(Encoding contentTypeCharset,
                                              MessageBodyGenerator messageBodyGenerator,
                                              IEnumerable<string> initialFlags = null,
                                              Action<ImapPseudoServer, ImapMessage, MessageBodyGenerator> testAction = null)
    {
      if (testAction == null)
        throw new ArgumentNullException(nameof(testAction));

      var encodingCharset = contentTypeCharset ?? Encoding.ASCII;
      var bodyStructure = new SinglePartMessage(
        MimeType.TextPlain,
        mediaTypeParameters: (contentTypeCharset == null) ? null : new Dictionary<string, string> { {"charset" ,contentTypeCharset.WebName} } ,
        sizeInOctets: messageBodyGenerator.TotalOctets
      ).ToBody();
      var envelope = new MessageEnvelope().ToEnvelope();
      var internalDate = new DateTimeOffset(2019, 3, 16, 3, 20, 5, TimeSpan.FromHours(+9.0));

      WithMessage(messageBodyGenerator.MessageUid, bodyStructure, envelope, internalDate, messageBodyGenerator.TotalOctets, initialFlags, (server, message) => {
        testAction(server, message, messageBodyGenerator);
      });
    }

    protected void WithMessagePartSingleTextPlain(Encoding contentTypeCharset = null,
                                                  string contentTypeCharsetName = null,
                                                  string contentTransferEncoding = null,
                                                  string messagePartBody = null,
                                                  IEnumerable<string> initialFlags = null,
                                                  Action<ImapPseudoServer, ImapMessagePart, MessageBodyGenerator> testAction = null)
    {
      WithMessagePartSingleTextPlain(
        contentTypeCharset: contentTypeCharset,
        contentTypeCharsetName: contentTypeCharsetName,
        contentTransferEncoding: contentTransferEncoding,
        messageBodyGenerator: new MessageBodyGenerator(messageUid: 1L, messagePartBody),
        initialFlags: initialFlags,
        testAction: testAction
      );
    }

    protected void WithMessagePartSingleTextPlain(Encoding contentTypeCharset = null,
                                                  string contentTypeCharsetName = null,
                                                  string contentTransferEncoding = null,
                                                  MessageBodyGenerator messageBodyGenerator = null,
                                                  IEnumerable<string> initialFlags = null,
                                                  Action<ImapPseudoServer, ImapMessagePart, MessageBodyGenerator> testAction = null)
    {
      if (testAction == null)
        throw new ArgumentNullException(nameof(testAction));

      var encodingCharset = contentTypeCharset ?? Encoding.ASCII;
      var charsetName = contentTypeCharsetName ?? contentTypeCharset?.WebName;
      var bodyStructure = new MultiPartMessage(
        "mixed",
        new SinglePartMessage(
          MimeType.TextPlain,
          mediaTypeParameters: (charsetName == null) ? null : new Dictionary<string, string> { {"charset", charsetName} } ,
          encoding: contentTransferEncoding ?? "7bit",
          sizeInOctets: messageBodyGenerator.TotalOctets
        )
      ).ToBody();
      var envelope = new MessageEnvelope().ToEnvelope();
      var internalDate = new DateTimeOffset(2019, 3, 16, 3, 20, 5, TimeSpan.FromHours(+9.0));

      WithMessage(messageBodyGenerator.MessageUid, bodyStructure, envelope, internalDate, messageBodyGenerator.TotalOctets, initialFlags: initialFlags, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsNotNull(part);
        Assert.AreEqual(part.Length, messageBodyGenerator.TotalOctets);

        testAction(server, part, messageBodyGenerator);
      });
    }

    protected void WithMessagePartEncapsulated(Encoding contentTypeCharset = null,
                                               string contentTypeCharsetName = null,
                                               string contentTransferEncoding = null,
                                               string messagePartBody = null,
                                               Action<ImapPseudoServer, ImapMessagePart, MessageBodyGenerator> testAction = null)
    {
      WithMessagePartEncapsulated(
        contentTypeCharset: contentTypeCharset,
        contentTypeCharsetName: contentTypeCharsetName,
        contentTransferEncoding: contentTransferEncoding,
        messageBodyGenerator: new MessageBodyGenerator(messageUid: 1L, messagePartBody),
        testAction: testAction
      );
    }

    protected void WithMessagePartEncapsulated(Encoding contentTypeCharset = null,
                                               string contentTypeCharsetName = null,
                                               string contentTransferEncoding = null, 
                                               MessageBodyGenerator messageBodyGenerator = null,
                                               Action<ImapPseudoServer, ImapMessagePart, MessageBodyGenerator> testAction = null)
    {
      if (testAction == null)
        throw new ArgumentNullException(nameof(testAction));

      var encodingCharset = contentTypeCharset ?? Encoding.ASCII;
      var charsetName = contentTypeCharsetName ?? contentTypeCharset?.WebName;
      var bodyStructure = new Rfc822Message(
        new SinglePartMessage(
          MimeType.TextPlain,
          mediaTypeParameters: (charsetName == null) ? null : new Dictionary<string, string> { {"charset", charsetName} } ,
          encoding: contentTransferEncoding ?? "7bit",
          sizeInOctets: messageBodyGenerator.TotalOctets
        )
      ).ToBody();
      var envelope = new MessageEnvelope().ToEnvelope();
      var internalDate = new DateTimeOffset(2019, 3, 16, 3, 20, 5, TimeSpan.FromHours(+9.0));

      WithMessage(messageBodyGenerator.MessageUid, bodyStructure, envelope, internalDate, messageBodyGenerator.TotalOctets, initialFlags: null, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsNotNull(part);
        Assert.IsTrue(part.IsEncapsulatedMessagePart);
        Assert.AreEqual(part.Length, messageBodyGenerator.TotalOctets);

        testAction(server, part, messageBodyGenerator);
      });
    }
  }

  [TestFixture]
  public class ImapMessageBodyBaseTests : ImapMessageBodyBaseTestsBase {
    [SetUp]
    public void SetUp()
    {
      TestUtils.Encodings.RegisterCodePagesEncodingProvider();
    }

    [Test]
    public void TestReadMessageBody_AttributesMustNotBeSavedByFetchMessageBody()
      => ReadMessageBody_AttributesMustNotBeSavedByFetchMessageBody(runAsync: false);

    [Test]
    public void TestReadMessageBody_AttributesMustNotBeSavedByFetchMessageBody_Async()
      => ReadMessageBody_AttributesMustNotBeSavedByFetchMessageBody(runAsync: true);

    private void ReadMessageBody_AttributesMustNotBeSavedByFetchMessageBody(bool runAsync)
    {
      var messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      TestCases.WithOpenedMailbox("INBOX", (server, mailbox) => {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.None); // fetch no attributes (except UID)

        server.DequeueRequest();

        var octets = OctetEncoding.EightBits.GetByteCount(messageBody);

        server.EnqueueTaggedResponse($"* {message.Uid} FETCH (RFC822.SIZE {octets} BODY[] {{{octets}}}\r\n{messageBody})\r\n" +
                                      "$tag OK done\r\n");

        Assert.IsFalse(message.Client.IsBusy);

        string fetchedMessageBody = null;

        if (runAsync)
          Assert.DoesNotThrowAsync(async () => fetchedMessageBody = await message.Body.ReadAsStringAsync());
        else
          Assert.DoesNotThrow(() => fetchedMessageBody = message.Body.ReadAsString());

        Assert.IsFalse(message.Client.IsBusy);
        Assert.AreEqual(messageBody, fetchedMessageBody);

        StringAssert.Contains("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.", server.DequeueRequest());

        // test EnsureStaticAttributesFetched / FETCH (msg-att-static) must be issued
        var bodyStructure = new SinglePartMessage(
          MimeType.TextPlain,
          encoding: "7bit",
          sizeInOctets: octets,
          lineCount: messageBody.Count('\n')
        ).ToBody();
        var envelope = new MessageEnvelope(
          date: "Tue, 23 Mar 2010 09:46:52 +0900",
          subject: "test message",
          from: new[] {new Address(name: "from address", mailbox: "from", host: "mail.example.com") },
          sender: new[] {new Address(name: "sender address", mailbox: "sender", host: "mail.example.com") },
          replyTo: new[] {new Address(name: "reply-to address", mailbox: "reply-to", host: "mail.example.com") },
          to: new[] {new Address(name: "to address", mailbox: "to", host: "mail.example.com") },
          cc: new[] {new Address(name: "cc address", mailbox: "cc", host: "mail.example.com") },
          bcc: new[] {new Address(name: "bcc address", mailbox: "bcc", host: "mail.example.com") },
          inReplyTo: "<in.reply.to@mail.example.com>",
          messageId: "<message.id@mail.example.com>"
        ).ToEnvelope();
        var internalDate = "23-Mar-2010 09:46:52 +0900";
        var rfc822SizeInResponse = 316L;

        server.EnqueueTaggedResponse($"UID FETCH {message.Uid} (BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)\r\n",
                                      "* 1 FETCH " +
                                     $"(UID {message.Uid} " +
                                     $" BODYSTRUCTURE ({bodyStructure})" +
                                     $" ENVELOPE {envelope}" +
                                     $" INTERNALDATE \"{internalDate}\"" +
                                     $" RFC822.SIZE {rfc822SizeInResponse}" +
                                      ")\r\n" +
                                     "$tag OK done\r\n");

        Assert.AreEqual(rfc822SizeInResponse, message.Length); // will fetch static attributes

        StringAssert.Contains($"UID FETCH {message.Uid} (BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)",
                              server.DequeueRequest());

        // test EnsureDynamicAttributesFetched / FETCH (msg-att-dynamic) must be issued
        server.EnqueueTaggedResponse($"UID FETCH {message.Uid} (FLAGS)\r\n",
                                     $"* 1 FETCH (UID {message.Uid} FLAGS (\\Seen))\r\n" +
                                      "$tag OK done\r\n");

        Assert.IsTrue(message.IsSeen); // will fetch dynamic attributes

        StringAssert.Contains($"UID FETCH {message.Uid} (FLAGS)", server.DequeueRequest());
      });
    }

    [Test]
    public void TestReadMessageBody_Busy()
    {
      const string messageBody = "body";

      WithMessageSingleTextPlain(contentTypeCharset: null, messageBody: messageBody, testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        var ret = message.Body.ReadAs((Stream stream) => {
          Assert.IsTrue(message.Client.IsBusy);

          Assert.Throws<InvalidOperationException>(() => message.Body.ReadAsString());
          Assert.Throws<InvalidOperationException>(() => message.Body.ReadAsStringAsync());

          return new StreamReader(stream).ReadToEnd();
        });

        Assert.IsFalse(message.Client.IsBusy);

        Assert.AreEqual(messageBody, ret);
      });
    }

    [Test]
    public void TestReadMessageBody_Busy_Async()
    {
      const string messageBody = "body";

      WithMessageSingleTextPlain(contentTypeCharset: null, messageBody: messageBody, testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrowAsync(async () => {
          var ret = await message.Body.ReadAsAsync(async (Stream stream) => {
            Assert.IsTrue(message.Client.IsBusy);

            Assert.Throws<InvalidOperationException>(() => message.Body.ReadAsString());
            Assert.Throws<InvalidOperationException>(() => message.Body.ReadAsStringAsync());

            return await new StreamReader(stream).ReadToEndAsync();
          });

          Assert.IsFalse(message.Client.IsBusy);

          Assert.AreEqual(messageBody, ret);
        });
      });
    }

    private void AssertCloseStream(ImapMessageBody body, Stream stream)
    {
      Assert.IsNotNull(stream);
      Assert.IsTrue(body.Message.Client.IsBusy, "IsBusy; still open");

      Assert.DoesNotThrow(stream.Dispose, "dispose stream");

      Assert.IsFalse(body.Message.Client.IsBusy, "IsBusy; after close");

      Assert.IsFalse(stream.CanRead, "CanRead");
      Assert.IsFalse(stream.CanWrite, "CanWrite");
      Assert.IsFalse(stream.CanSeek, "CanSeek");
      //Assert.IsTrue(stream.CanTimeout, "CanTimeout");

      if (Runtime.IsRunningOnMono && !(stream is ImapFetchMessageBodyStream)) {
        // mono bug? CryptoStream.ReadByte() throws NotSupportedException
      }
      else {
        Assert.Throws<ObjectDisposedException>(() => stream.ReadByte(), "operation with disposed stream");
      }

      Assert.DoesNotThrow(stream.Dispose, "dispose stream again");

      Assert.IsFalse(body.Message.Client.IsBusy, "IsBusy; after re-close");
    }

    private void AssertCloseStream(ImapMessageBody body, StreamReader reader)
    {
      Assert.IsNotNull(reader);
      Assert.IsNotNull(reader.BaseStream);
      Assert.IsTrue(body.Message.Client.IsBusy, "IsBusy; still open");

      Assert.DoesNotThrow(reader.Dispose, "dispose stream");

      Assert.IsFalse(body.Message.Client.IsBusy, "IsBusy; after close");

      Assert.IsNull(reader.BaseStream);

      Assert.Throws<ObjectDisposedException>(() => reader.Read(), "operation with disposed stream");

      Assert.DoesNotThrow(reader.Dispose, "dispose stream again");

      Assert.IsFalse(body.Message.Client.IsBusy, "IsBusy; after re-close");
    }

    [Test]
    public void TestReadMessageBody_StreamLengthMustThrowsNotSupportedException_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_StreamLengthMustThrowsNotSupportedException(ImapMessageFetchBodyOptions.Default, messageBodyGenerator, server, message.Body);
      });
    }

    [Test]
    public void TestReadMessageBody_StreamLengthMustThrowsNotSupportedException_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_StreamLengthMustThrowsNotSupportedException(ImapMessageFetchBodyOptions.Default, messageBodyGenerator, server, messagePart.Body);
      });
    }

    [Test]
    public void TestReadMessageBody_StreamLengthMustThrowsNotSupportedException_OptionsDecodeContent_Base64()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "base64", contentTypeCharset: Encoding.UTF8, messagePartBody: @"YmFzZTY0IGVuY29kZWQg5pys5paH", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_StreamLengthMustThrowsNotSupportedException(ImapMessageFetchBodyOptions.DecodeContent, messageBodyGenerator, server, messagePart.Body);
      });
    }

    [Test]
    public void TestReadMessageBody_StreamLengthMustThrowsNotSupportedException_OptionsDecodeContent_QuotedPrintable()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "quoted-printable", contentTypeCharset: Encoding.UTF8, messagePartBody: @"quoted-printable =E6=9C=AC=E6=96=87", testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_StreamLengthMustThrowsNotSupportedException(ImapMessageFetchBodyOptions.DecodeContent, messageBodyGenerator, server, messagePart.Body);
      });
    }

    private void ReadMessageBody_StreamLengthMustThrowsNotSupportedException(ImapMessageFetchBodyOptions options, MessageBodyGenerator messageBodyGenerator, ImapPseudoServer server, ImapMessageBody body)
    {
      messageBodyGenerator.EmitNextResponse(server);

      body.Read(
        options: options,
        read: (Stream stream) => Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length))
      );

      Assert.IsFalse(body.Message.Client.IsBusy);
    }

    [Test]
    public void TestReadMessageBody_StreamLengthMustThrowsNotSupportedException_Partial()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, new MessageBodyGenerator(1L, 128), testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        message.Body.Read(
          fetchRange: new ImapFetchRange(64, 128),
          read: (Stream stream) => Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length))
        );

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_CloseStream()
    {
      const string messageBody = "body";

      WithMessageSingleTextPlain(contentTypeCharset: null, messageBody: messageBody, testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          message.Body.Read((Stream stream) => AssertCloseStream(message.Body, stream));
        });

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_CloseStream_OptionsDecodeContent()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "base64", contentTypeCharset: Encoding.UTF8, messagePartBody: @"Ym9keQ==", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (Stream stream) => AssertCloseStream(messagePart.Body, stream));
        });

        Assert.IsFalse(messagePart.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_CloseStream_ViaStreamReader()
    {
      const string messageBody = "body";

      WithMessageSingleTextPlain(contentTypeCharset: null, messageBody: messageBody, testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          message.Body.Read((StreamReader reader) => AssertCloseStream(message.Body, reader));
        });

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_CloseStream_ViaStreamReader_OptionsDecodeContent()
    {
      WithMessagePartSingleTextPlain(contentTransferEncoding: "base64", contentTypeCharset: Encoding.UTF8, messagePartBody: @"Ym9keQ==", testAction: (server, messagePart, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.DoesNotThrow(() => {
          messagePart.Body.Read(options: ImapMessageFetchBodyOptions.DecodeContent, read: (StreamReader reader) => AssertCloseStream(messagePart.Body, reader));
        });

        Assert.IsFalse(messagePart.Client.IsBusy);
      });
    }

    private class ConverterException : Exception {
    }

    [Test]
    public void TestReadMessageBody_ExceptionFromConverter()
    {
      WithMessageSingleTextPlain(contentTypeCharset: null, messageBody: "body", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.Throws<ConverterException>(() => {
          message.Body.Read((Stream stream) => throw new ConverterException());
        });

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_ExceptionFromConverter_Async()
    {
      WithMessageSingleTextPlain(contentTypeCharset: null, messageBody: "body", testAction: (server, message, messageBodyGenerator) => {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.ThrowsAsync<ConverterException>(async () => {
          await message.Body.ReadAsync((Stream stream) => throw new ConverterException());
        });

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_MailboxClosed()
      => ReadMessageBody_MailboxClosed(message => message.Body.ReadAsString());

    [Test]
    public void TestReadMessageBody_MailboxClosed_Async()
      => ReadMessageBody_MailboxClosed(message => message.Body.ReadAsStringAsync());

    private void ReadMessageBody_MailboxClosed(Action<ImapMessage> testAction)
    {
      TestCases.WithMessageMailboxClosed((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => testAction(message));

        TestUtils.Assert.IsSerializableBinaryFormat(ex);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_VanishedMessage()
      => ReadMessageBody_VanishedMessage(message => message.Body.ReadAsString());

    [Test]
    public void TestReadMessageBody_VanishedMessage_Async()
      => ReadMessageBody_VanishedMessage(message => message.Body.ReadAsStringAsync());

    private void ReadMessageBody_VanishedMessage(Action<ImapMessage> testAction)
    {
      TestCases.WithMessageVanished((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => testAction(message));

        TestUtils.Assert.IsSerializableBinaryFormat(ex);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadMessageBody_MessageVanishingByFetch()
      => ReadMessageBody_MessageVanishingByFetch(false);

    [Test]
    public void TestReadMessageBody_MessageVanishingByFetch_Async()
      => ReadMessageBody_MessageVanishingByFetch(true);

    private void ReadMessageBody_MessageVanishingByFetch(bool runAsync)
    {
      TestCases.WithMessage((server, message) => {
        Assert.AreNotEqual(ImapMessage.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsFalse(message.IsDeleted);

        // FETCH resopnse without BODY
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        var ex = runAsync
          ? Assert.ThrowsAsync<ImapMessageVanishedException>(async () => await message.Body.ReadAsStringAsync())
          : Assert.Throws<ImapMessageVanishedException>(() => message.Body.ReadAsString());

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(ex.VanishedMessage);
        Assert.AreSame(message, ex.VanishedMessage);

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.AreEqual(ImapMessage.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsTrue(message.IsDeleted);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));
        Assert.IsFalse(message.IsAvailable);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeCall_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeCall(message, message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeCall_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeCall(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeCall_Async_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeCall_Async(message, message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeCall_Async_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeCall_Async(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Cancellation_CancelledBeforeCall(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      using (var cts = new CancellationTokenSource()) {
        cts.Cancel();

        Assert.Throws<OperationCanceledException>(() => {
          body.Read(cancellationToken: cts.Token, read: (Stream stream) => Assert.Fail("converter called"));
        });

        Assert.IsFalse(message.Client.IsBusy);
      }
    }

    private void ReadMessageBody_Cancellation_CancelledBeforeCall_Async(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      using (var cts = new CancellationTokenSource()) {
        cts.Cancel();

        Assert.ThrowsAsync<TaskCanceledException>(async () => {
          await body.ReadAsync(cancellationToken: cts.Token, read: (Stream stream) => { Assert.Fail("converter called");
#if NET45
            return Task.FromResult(0);
#else
            return Task.CompletedTask;
#endif
          });
        });

        Assert.IsFalse(message.Client.IsBusy);
      }
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeRead_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeRead(message, message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeRead_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeRead(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeRead_Async_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeRead_Async(message, message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeRead_Async_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 1), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeRead_Async(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Cancellation_CancelledBeforeRead(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      using (var cts = new CancellationTokenSource()) {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.Throws<OperationCanceledException>(() => {
          body.Read(cancellationToken: cts.Token, read: (Stream stream) => {
            cts.Cancel();

            try {
              stream.ReadByte();
            }
            catch (OperationCanceledException) {
              Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
              throw;
            }
          });
        });

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");

        StringAssert.EndsWith("<0.10240>)\r\n", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest());
      }
    }

    private void ReadMessageBody_Cancellation_CancelledBeforeRead_Async(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      using (var cts = new CancellationTokenSource()) {
        messageBodyGenerator.EmitNextResponse(server);

        Assert.ThrowsAsync<OperationCanceledException>(async () => {
          await body.ReadAsync(cancellationToken: cts.Token, read: async (Stream stream) => {
            cts.Cancel();

            try {
              await stream.ReadAsync(new byte[1], 0, 1);
            }
            catch (OperationCanceledException) {
              Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
              throw;
            }
          });
        });

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");

        StringAssert.EndsWith("]<0.10240>)\r\n", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest());
      }
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeFillBuffer_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 2), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeFillBuffer(message, message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeFillBuffer_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 2), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeFillBuffer(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeFillBuffer_Async_OfMessage()
    {
      WithMessageSingleTextPlain(Encoding.ASCII, MessageBodyGenerator.Create(1L, blockCount: 2), testAction: (server, message, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeFillBuffer_Async(message, message.Body, server, messageBodyGenerator);
      });
    }

    [Test]
    public void TestReadMessageBody_Cancellation_CancelledBeforeFillBuffer_Async_OfMessagePart()
    {
      WithMessagePartSingleTextPlain(Encoding.ASCII, messageBodyGenerator: MessageBodyGenerator.Create(1L, blockCount: 2), testAction: (server, messagePart, messageBodyGenerator) => {
        ReadMessageBody_Cancellation_CancelledBeforeFillBuffer_Async(messagePart.Message, messagePart.Body, server, messageBodyGenerator);
      });
    }

    private void ReadMessageBody_Cancellation_CancelledBeforeFillBuffer(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      using (var cts = new CancellationTokenSource()) {
        messageBodyGenerator.EmitNextResponse(server, actionAfterRespond: () => cts.Cancel()); // cancel before reading block #2

        Assert.Throws<OperationCanceledException>(() => {
          body.Read(cancellationToken: cts.Token, read: (Stream stream) => {
            var buffer = new byte[ImapFetchMessageBodyStream.DefaultFetchBlockSize];

            // read #1 block (-1 bytes), fetching #1 block
            Assert.AreEqual(buffer.Length - 1,
                            stream.Read(buffer, 0, buffer.Length - 1));

            // try to read last 1 byte of #1 block and #2 block (-1 bytes), and throws OperationCanceledException
            try {
              stream.Read(buffer, 0, buffer.Length);
            }
            catch (OperationCanceledException) {
              Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
              throw;
            }
          });
        });

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");

        StringAssert.EndsWith("]<0.10240>)\r\n", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest());
      }
    }

    private void ReadMessageBody_Cancellation_CancelledBeforeFillBuffer_Async(ImapMessage message, ImapMessageBody body, ImapPseudoServer server, MessageBodyGenerator messageBodyGenerator)
    {
      using (var cts = new CancellationTokenSource()) {
        messageBodyGenerator.EmitNextResponse(server, actionAfterRespond: () => cts.Cancel()); // cancel before reading block #2

        Assert.ThrowsAsync<OperationCanceledException>(async () => {
          await body.ReadAsync(cancellationToken: cts.Token, read: async (Stream stream) => {
            var buffer = new byte[ImapFetchMessageBodyStream.DefaultFetchBlockSize];

            // read #1 block (-1 bytes), fetching #1 block
            Assert.AreEqual(buffer.Length - 1,
                            stream.Read(buffer, 0, buffer.Length - 1));

            // try to read last 1 byte of #1 block and #2 block (-1 bytes), and throws OperationCanceledException
            try {
              await stream.ReadAsync(buffer, 0, buffer.Length);
            }
            catch (OperationCanceledException) {
              Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
              throw;
            }
          });
        });

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");

        StringAssert.EndsWith("]<0.10240>)\r\n", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest());
      }
    }
  }
}
