using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessagePartOperationsFetchTests {
    [SetUp]
    public void SetUp()
    {
      TestUtils.Encodings.RegisterCodePagesEncodingProvider();
    }

    private const string dummyEnvelope = "(NIL \"test\" NIL NIL NIL NIL NIL NIL NIL NIL)";

    private void TestMessage(string fetchResponseBodyStructure,
                             Action<ImapPseudoServer, ImapMessage> testAction)
    {
      var fetchResp =
        "* 1 FETCH (UID 1 " +
        "BODYSTRUCTURE " + fetchResponseBodyStructure +
        " ENVELOPE " + dummyEnvelope +
        " INTERNALDATE \"27-Jul-2014 21:50:32 +0900\" RFC822.SIZE 42)\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp,
                             ImapMessageFetchAttributeOptions.StaticAttributes,
                             (server, message) => {
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.MessagePart);

        testAction(server, message);
      });
    }

    private void TestMessagePartPlainTextSingle(Action<ImapPseudoServer, ImapMessagePart> testAction)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        testAction(server, part);
      });
    }

    private void TestMessagePartPlainTextSingle(string messageBody, Action<ImapPseudoServer, ImapMessagePart, string> testAction)
    {
      messageBody = messageBody.Replace("\r\n", "\n").Replace("\n", "\r\n");
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsNotNull(part);
        Assert.AreEqual(part.Length, messageOctet);

        testAction(server, part, messageBody);
      });
    }

    [Obsolete]
    private static Task<int> ReadAsyncProc(Stream stream, Action<Stream> proc)
    {
      proc(stream);
      return Task.FromResult(default(int));
    }

    [Test, Obsolete]
    public void TestRead_Stream_SinglePart()
      => Read_Stream_SinglePart((part, read) => part.Body.Read(read));

    [Test, Obsolete]
    public void TestReadAsync_Stream_SinglePart()
      => Read_Stream_SinglePart((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read).Wait());

    private void Read_Stream_SinglePart(Action<ImapMessagePart, Action<Stream>> read)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        Assert.IsFalse(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreEqual(expectedMessageBody, (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_Encapsulated()
      => Read_Stream_Encapsulated((part, read) => part.Body.Read(read));

    [Test, Obsolete]
    public void TestReadAsync_Stream_Encapsulated()
      => Read_Stream_Encapsulated((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read).Wait());

    private void Read_Stream_Encapsulated(Action<ImapMessagePart, Action<Stream>> read)
    {
      const string bodyStructureEncapsulated = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "(" +
        "(\"message\" \"rfc822\" NIL NIL NIL \"7bit\" 160 " +
        dummyEnvelope + " " + 
        bodyStructureEncapsulated + " 1)" +
        " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsFalse(part.IsMultiPart);
        Assert.IsTrue(part.HasEncapsulatedMessage);

        const string expectedMessageBody = @"MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: message/rfc822

MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: text/plain

0123456789abcdef
";

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreEqual(expectedMessageBody,
                          (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_MultiPart()
      => Read_Stream_MultiPart((part, read) => part.Body.Read(read));

    [Test, Obsolete]
    public void TestReadAsync_Stream_MultiPart()
      => Read_Stream_MultiPart((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read).Wait());

    private void Read_Stream_MultiPart(Action<ImapMessagePart, Action<Stream>> read)
    {
      const string bodyStructureNested = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "((" + bodyStructureNested + " \"mixed\") \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsTrue(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        const string expectedMessageBody = @"MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: multipart/mixed; boundary=""xxxxxxxx""

This is a multi-part message in MIME format.

--xxxxxxxx
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: text/plain

0123456789abcdef
--xxxxxxxx--
";

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreEqual(expectedMessageBody,
                          (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_OmitHeader_RootMessagePart()
      => Read_Stream_OmitHeader_RootMessagePart((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.OmitHeader));

    [Test, Obsolete]
    public void TestReadAsync_Stream_OmitHeader_RootMessagePart()
      => Read_Stream_OmitHeader_RootMessagePart((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.OmitHeader).Wait());

    private void Read_Stream_OmitHeader_RootMessagePart(Action<ImapMessagePart, Action<Stream>> read)
    {
      const string respBodyStructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string body = "0123456789abcdef";

      TestMessage(respBodyStructure, (server, message) => {
        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[TEXT]<0.10240>)\r\n",
                                       "* 1 FETCH (BODY[1] {16}\r\n" + body + ")\r\n" +
                                       "$tag OK done\r\n");
        });

        read(message.MessagePart, stream => {
          responseTask.Wait();

          Assert.IsTrue(message.Client.IsBusy);

          Assert.IsNotNull(stream);
          Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length));

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_OmitHeader_SpecificMessagePart()
      => Read_Stream_OmitHeader_SpecificMessagePart((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.OmitHeader));

    [Test, Obsolete]
    public void TestReadAsync_Stream_OmitHeader_SpecificMessagePart()
      => Read_Stream_OmitHeader_SpecificMessagePart((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.OmitHeader).Wait());

    private void Read_Stream_OmitHeader_SpecificMessagePart(Action<ImapMessagePart, Action<Stream>> read)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       "* 1 FETCH (BODY[1] {16}\r\n" + body + ")\r\n" +
                                       "$tag OK done\r\n");
        });

        read(part, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length));

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_RangedFetch()
      => Read_Stream_RangedFetch((part, range, read) => part.Body.Read(read, fetchRange: range));

    [Test, Obsolete]
    public void TestReadAsync_Stream_RangedFetch()
      => Read_Stream_RangedFetch((part, range, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, fetchRange: range).Wait());

    private void Read_Stream_RangedFetch(Action<ImapMessagePart, ImapFetchRange, Action<Stream>> read)
    {
      var fetchRange = new ImapFetchRange(128, 64);
      var body = new string('x', (int)fetchRange.Length);

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<128.64>)\r\n",
                                       "* 1 FETCH (BODY[1] {64}\r\n" + body + ")\r\n" +
                                       "$tag OK done\r\n");
        });

        read(part, fetchRange, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          Assert.Throws<NotSupportedException>(() => Assert.IsNotNull(stream.Length));

          Assert.AreEqual(expectedMessageBody, (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_RangedFetch_DecodeContent()
      => Read_Stream_RangedFetch_DecodeContent((part, range, read) => part.Body.Read(read, fetchRange: range, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_Stream_RangedFetch_DecodeContent()
      => Read_Stream_RangedFetch_DecodeContent((part, range, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, fetchRange: range, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_Stream_RangedFetch_DecodeContent(Action<ImapMessagePart, ImapFetchRange, Action<Stream>> read)
    {
      var fetchRange = new ImapFetchRange(128, 64);
      var body = new string('x', (int)fetchRange.Length);

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        TestUtils.Assert.ThrowsOrAggregates<NotSupportedException>(() => read(part, fetchRange, s => Assert.IsNotNull(s)));

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_DecodeContent_CTEBase64()
      => Read_Stream_DecodeContent_CTEBase64((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_Stream_DecodeContent_CTEBase64()
      => Read_Stream_DecodeContent_CTEBase64((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_Stream_DecodeContent_CTEBase64(Action<ImapMessagePart, Action<Stream>> read)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreEqual("base64 encoded 本文",
                          (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_DecodeContent_CTEQuotedPrintable()
      => Read_Stream_DecodeContent_CTEQuotedPrintable((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_Stream_DecodeContent_CTEQuotedPrintable()
      => Read_Stream_DecodeContent_CTEQuotedPrintable((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_Stream_DecodeContent_CTEQuotedPrintable(Action<ImapMessagePart, Action<Stream>> read)
    {
      var messageBody = @"quoted-printable =E6=9C=AC=E6=96=87";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"quoted-printable\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, stream => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreEqual("quoted-printable 本文",
                          (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_DecodeContent_CloseStreamBeforeRead_Base64()
      => Read_Stream_DecodeContent_CloseStreamBeforeRead_Base64((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_Stream_DecodeContent_CloseStreamBeforeRead_Base64()
      => Read_Stream_DecodeContent_CloseStreamBeforeRead_Base64((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_Stream_DecodeContent_CloseStreamBeforeRead_Base64(Action<ImapMessagePart, Action<Stream>> read)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });


        read(part, stream => {
          responseTask.Wait();

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_Stream_DecodeContent_CloseStreamBeforeRead_QuotedPrintable()
      => Read_Stream_DecodeContent_CloseStreamBeforeRead_QuotedPrintable((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_Stream_DecodeContent_CloseStreamBeforeRead_QuotedPrintable()
      => Read_Stream_DecodeContent_CloseStreamBeforeRead_QuotedPrintable((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_Stream_DecodeContent_CloseStreamBeforeRead_QuotedPrintable(Action<ImapMessagePart, Action<Stream>> read)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"quoted-printable\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });


        read(part, stream => {
          responseTask.Wait();

          Assert.DoesNotThrow(() => stream.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestReadAs_Stream()
    {
      ReadAs_Stream((part, converter) => part.ReadAs(converter));
    }

    [Test, Obsolete]
    public void TestReadAsAsync_Stream()
    {
      ReadAs_Stream((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_Stream(Func<ImapMessagePart, Converter<Stream, byte[]>, byte[]> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAs(part, delegate(Stream stream) {
          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);

          Assert.IsTrue(part.Client.IsBusy);

          return stream.ReadToEnd();
        });

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody,
                        OctetEncoding.EightBits.GetString(ret));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAs_Stream_ExceptionInConveter()
    {
      ReadAs_Stream_ExceptionInConveter((part, converter) => part.ReadAs(converter));
    }

    [Test, Obsolete]
    public void TestReadAsAsync_Stream_ExceptionInConveter()
    {
      ReadAs_Stream_ExceptionInConveter((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_Stream_ExceptionInConveter(Func<ImapMessagePart, Converter<Stream, int>, int> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part,  expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Assert.ThrowsOrAggregates<NotSupportedException>(() => readAs(part, (Converter<Stream, int>)delegate {
          throw new NotSupportedException("exception");
        }));

        responseTask.Wait();

        Assert.AreEqual("exception", ex.Message);

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAs_Stream_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAs<string>((Converter<Stream, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAsAsync_Stream_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAsAsync<string>((Converter<Stream, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAllBytes()
    {
      ReadAllBytes(part => part.ReadAllBytes());
    }

    [Test, Obsolete]
    public void TestReadAllBytesAsync()
    {
      ReadAllBytes(part => part.ReadAllBytesAsync().Result);
    }

    private void ReadAllBytes(Func<ImapMessagePart, byte[]> readAllBytes)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part,expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAllBytes(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody,
                        OctetEncoding.EightBits.GetString(ret));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestWriteTo_Stream()
    {
      WriteTo_Stream((part, dest) => part.WriteTo(dest));
    }

    [Test, Obsolete]
    public void TestWriteToAsync_Stream()
    {
      WriteTo_Stream((part, dest) => part.WriteToAsync(dest).Wait());
    }

    private void WriteTo_Stream(Action<ImapMessagePart, Stream> writeTo)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        using (var outputStream = new MemoryStream()) {
          writeTo(part, outputStream);

          responseTask.Wait();

          outputStream.Dispose();

          //Assert.AreEqual(part.Length, outputStream.Length);
          Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        outputStream.ToArray());
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }

    [Test, Obsolete]
    public void TestWriteTo_Stream_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteTo((Stream)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestWriteTo_Stream_DestinationNotWritable()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => part.WriteTo(stream));
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }

    [Test, Obsolete]
    public void TestWriteToAsync_Stream_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteToAsync((Stream)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestWriteToAsync_Stream_DestinationNotWritable()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => part.WriteToAsync(stream));
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }

    [Test, Obsolete]
    public void TestWriteTo_BinaryWriter()
    {
      WriteTo_BinaryWriter((part, writer) => part.WriteTo(writer));
    }

    [Test, Obsolete]
    public void TestWriteToAsync_BinaryWriter()
    {
      WriteTo_BinaryWriter((part, writer) => part.WriteToAsync(writer).Wait());
    }

    private void WriteTo_BinaryWriter(Action<ImapMessagePart, BinaryWriter> writeTo)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        using (var outputStream = new MemoryStream()) {
          var writer = new BinaryWriter(outputStream);

          writeTo(part, writer);

          responseTask.Wait();

          outputStream.Dispose();

          //Assert.AreEqual(part.Length, outputStream.Length);
          Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                          outputStream.ToArray());
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }


    [Test, Obsolete]
    public void TestWriteTo_BinaryWriter_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteTo((BinaryWriter)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestWriteToAsync_BinaryWriter_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteToAsync((BinaryWriter)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestSave()
    {
      Save((part, file) => part.Save(file));
    }

    [Test, Obsolete]
    public void TestSaveAsync()
    {
      Save((part, file) => part.SaveAsync(file).Wait());
    }

    private void Save(Action<ImapMessagePart, string> save)
    {
      const string body = "0123456789abcdef";
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");
      const int existingFileLength = 10 * 1024;

      TestUtils.UsingFile(outfile, () => {
        // prepare existing file
        File.WriteAllText(outfile, new string('x', existingFileLength));

        TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
          var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
            server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
          });

          save(part, outfile);

          responseTask.Wait();

          Assert.IsTrue(File.Exists(outfile), "file existence");
          //Assert.AreEqual(part.Length, (new FileInfo(outfile)).Length);
          Assert.AreEqual(expectedMessageBody, File.ReadAllText(outfile));
          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestSave_PathNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.Save(null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestSave_PathEmpty()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentException>(() => part.Save(string.Empty));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestSaveAsync_PathNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.SaveAsync(null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestSaveAsync_PathEmpty()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentException>(() => part.SaveAsync(string.Empty));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestDownload()
    {
      Download(part => part.Download());
    }

    [Test, Obsolete]
    public void TestDownloadAsync()
    {
      Download(part => part.DownloadAsync().Result);
    }

    private void Download(Func<ImapMessagePart, Stream> download)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = download(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.IsTrue(ret.CanSeek, "CanSeek");
        Assert.IsTrue(ret.CanRead, "CanRead");
        Assert.IsTrue(ret.CanWrite, "CanWrite");
        Assert.AreEqual(0L, ret.Position, "Position");
        //Assert.AreEqual(part.Length, ret.Length, "Length");

        Assert.DoesNotThrow(() => ret.Seek(-1L, SeekOrigin.End));
        Assert.AreNotEqual(-1, ret.ReadByte());
        Assert.DoesNotThrow(() => ret.Seek(0L, SeekOrigin.Begin));

        Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        ret.ReadToEnd());

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Obsolete]
    private static Task<int> ReadAsyncProc(StreamReader reader, Action<StreamReader> proc)
    {
      proc(reader);
      return Task.FromResult(default(int));
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_NotDecodeContent()
      => Read_StreamReader_NotDecodeContent((part, read) => part.Body.Read(read));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_NotDecodeContent()
      => Read_StreamReader_NotDecodeContent((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read).Wait());

    private void Read_StreamReader_NotDecodeContent(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"0123456789abcdef";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"7bit\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, reader.BaseStream.Length);
          Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
          Assert.AreEqual(messageBody, reader.ReadToEnd());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_NotDecodeContent_CharsetMissing()
      => Read_StreamReader_NotDecodeContent_CharsetMissing((part, read) => part.Body.Read(read));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_NotDecodeContent_CharsetMissing()
      => Read_StreamReader_NotDecodeContent_CharsetMissing((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read).Wait());

    private void Read_StreamReader_NotDecodeContent_CharsetMissing(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"0123456789abcdef";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" NIL NIL NIL \"7bit\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, reader.BaseStream.Length);
          Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
          Assert.AreEqual(messageBody, reader.ReadToEnd());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_CTEBase64()
      => Read_StreamReader_DecodeContent_CTEBase64((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_CTEBase64()
      => Read_StreamReader_DecodeContent_CTEBase64((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_CTEBase64(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, reader.BaseStream.Length);
          Assert.AreEqual(Encoding.UTF8, reader.CurrentEncoding);
          Assert.AreEqual("base64 encoded 本文", reader.ReadToEnd());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_CTEQuotedPrintable()
      => Read_StreamReader_DecodeContent_CTEQuotedPrintable((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_CTEQuotedPrintable()
      => Read_StreamReader_DecodeContent_CTEQuotedPrintable((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_CTEQuotedPrintable(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"quoted-printable =E6=9C=AC=E6=96=87";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"quoted-printable\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, reader.BaseStream.Length);
          Assert.AreEqual(Encoding.UTF8, reader.CurrentEncoding);
          Assert.AreEqual("quoted-printable 本文", reader.ReadToEnd());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_CloseReaderBeforeRead_Base64()
      => Read_StreamReader_DecodeContent_CloseReaderBeforeRead_Base64((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_CloseReaderBeforeRead_Base64()
      => Read_StreamReader_DecodeContent_CloseReaderBeforeRead_Base64((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_CloseReaderBeforeRead_Base64(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_CloseReaderBeforeRead_QuotedPrintable()
      => Read_StreamReader_DecodeContent_CloseReaderBeforeRead_QuotedPrintable((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_CloseReaderBeforeRead_QuotedPrintable()
      => Read_StreamReader_DecodeContent_CloseReaderBeforeRead_QuotedPrintable((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_CloseReaderBeforeRead_QuotedPrintable(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"quoted-printable\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_CharsetMissing()
      => Read_StreamReader_DecodeContent_CharsetMissing((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_CharsetMissing()
      => Read_StreamReader_DecodeContent_CharsetMissing((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_CharsetMissing(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" NIL NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        read(part, reader => {
          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
          StringAssert.StartsWith("base64", reader.ReadToEnd());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_CharsetUnsupported()
      => Read_StreamReader_DecodeContent_CharsetUnsupported((part, read) => part.Body.Read(read, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_CharsetUnsupported()
      => Read_StreamReader_DecodeContent_CharsetUnsupported((part, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_CharsetUnsupported(Action<ImapMessagePart, Action<StreamReader>> read)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"x-unknown-charset\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.Throws<EncodingNotSupportedException>(() => read(part, r => Assert.IsNotNull(r)));

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_FallbackCharsetAvailable()
      => Read_StreamReader_DecodeContent_FallbackCharsetAvailable((part, callback, read) => part.Body.Read(read, selectFallbackCharset: callback, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_FallbackCharsetAvailable()
      => Read_StreamReader_DecodeContent_FallbackCharsetAvailable((part, callback, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, selectFallbackCharset: callback, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_FallbackCharsetAvailable(Action<ImapMessagePart, EncodingSelectionCallback, Action<StreamReader>> read)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"x-unknown-charset\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var called = false;
        var fallbackCharset = Encoding.UTF8;
        var callback = (EncodingSelectionCallback)delegate(string charset) {
          called = true;

          Assert.AreEqual("x-unknown-charset", charset);

          return fallbackCharset;
        };

        read(part, callback, reader => {
          Assert.IsTrue(called, "called back");

          responseTask.Wait();

          Assert.IsTrue(part.Client.IsBusy);

          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, stream.Length);
          Assert.AreSame(fallbackCharset, reader.CurrentEncoding);
          StringAssert.StartsWith("base64", reader.ReadToEnd());

          Assert.DoesNotThrow(() => reader.Dispose());

          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test, Obsolete]
    public void TestRead_StreamReader_DecodeContent_FallbackCharsetNotAvailable()
      => Read_StreamReader_DecodeContent_FallbackCharsetNotAvailable((part, callback, read) => part.Body.Read(read, selectFallbackCharset: callback, options: ImapMessageFetchBodyOptions.DecodeContent));

    [Test, Obsolete]
    public void TestReadAsync_StreamReader_DecodeContent_FallbackCharsetNotAvailable()
      => Read_StreamReader_DecodeContent_FallbackCharsetNotAvailable((part, callback, read) => part.Body.ReadAsAsync(converter: ReadAsyncProc, arg: read, selectFallbackCharset: callback, options: ImapMessageFetchBodyOptions.DecodeContent).Wait());

    private void Read_StreamReader_DecodeContent_FallbackCharsetNotAvailable(Action<ImapMessagePart, EncodingSelectionCallback, Action<StreamReader>> read)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"x-unknown-charset\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);
        var called = false;
        var callback = (EncodingSelectionCallback)delegate(string charset) {
          called = true;

          Assert.AreEqual("x-unknown-charset", charset);

          return null;
        };

        Assert.Throws<EncodingNotSupportedException>(() => read(part, callback, r => Assert.IsNotNull(r)));

        Assert.IsTrue(called, "called back");

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAs_StreamReader()
    {
      ReadAs_StreamReader((part, converter) => part.ReadAs(converter));
    }

    [Test, Obsolete]
    public void TestReadAsAsync_StreamReader()
    {
      ReadAs_StreamReader((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_StreamReader(Func<ImapMessagePart, Converter<StreamReader, string>, string> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAs(part, delegate(StreamReader reader) {
          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, reader.BaseStream.Length);

          Assert.IsTrue(part.Client.IsBusy);

          return reader.ReadToEnd();
        });

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody, ret);
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAs_StreamReader_ExceptionInConveter()
    {
      ReadAs_StreamReader_ExceptionInConveter((part, converter) => part.ReadAs(converter));
    }

    [Test, Obsolete]
    public void TestReadAsAsync_StreamReader_ExceptionInConveter()
    {
      ReadAs_StreamReader_ExceptionInConveter((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_StreamReader_ExceptionInConveter(Func<ImapMessagePart, Converter<StreamReader, int>, int> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Assert.ThrowsOrAggregates<NotSupportedException>(() => readAs(part, (Converter<StreamReader, int>)delegate {
          throw new NotSupportedException("exception");
        }));

        responseTask.Wait();

        Assert.AreEqual("exception", ex.Message);

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAs_StreamReader_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAs<string>((Converter<StreamReader, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAsAsync_StreamReader_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAsAsync<string>((Converter<StreamReader, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadLines()
    {
      const string body = @"1st line
2nd line
3rd line";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                     string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                   part.Length,
                                                   expectedMessageBody) +
                                     "$tag OK done\r\n");

        var lines = part.ReadLines();

        Assert.IsFalse(part.Client.IsBusy, "IsBusy; enumerator retrieved");

        Assert.IsNotNull(lines);

        CollectionAssert.AreEqual((new StringReader(body)).ReadAllLines(),
                                  lines.ToArray());

        Assert.IsFalse(part.Client.IsBusy, "IsBusy; after enumeration");
      });
    }

    [Test, Obsolete]
    public void TestReadAllLines()
    {
      ReadAllLines(part => part.ReadAllLines());
    }

    [Test, Obsolete]
    public void TestReadAllLinesAsync()
    {
      ReadAllLines(part => part.ReadAllLinesAsync().Result);
    }

    private void ReadAllLines(Func<ImapMessagePart, string[]> readAllLines)
    {
      const string body = @"1st line
2nd line
3rd line";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAllLines(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        CollectionAssert.AreEqual(expectedMessageBody.Split(new[] {"\r\n"}, StringSplitOptions.None), ret);
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestReadAllText()
    {
      ReadAllText(part => part.ReadAllText());
    }

    [Test, Obsolete]
    public void TestReadAllTextAsync()
    {
      ReadAllText(part => part.ReadAllTextAsync().Result);
    }

    private void ReadAllText(Func<ImapMessagePart, string> readAllText)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAllText(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody, ret);
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test, Obsolete]
    public void TestGetRawHeader()
    {
      GetRawHeader(part => part.GetRawHeader());
    }

    [Test, Obsolete]
    public void TestGetRawHeaderAsync()
    {
      GetRawHeader(part => part.GetRawHeaderAsync().Result);
    }

    private void GetRawHeader(Func<ImapMessagePart, string> getRawHeader)
    {
      const string bodyStructureNested = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "((" + bodyStructureNested + " \"mixed\") \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsTrue(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        var expectedMessageBody = @"MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: text/plain
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1.MIME]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1.MIME] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var headers = getRawHeader(part);

        responseTask.Wait();

        Assert.IsFalse(part.Client.IsBusy);

        Assert.AreEqual(expectedMessageBody, headers);
      });
    }

    [Test, Obsolete]
    public void TestGetHeader()
    {
      GetHeader(false, part => part.GetHeader());
    }

    [Test, Obsolete]
    public void TestGetHeader_Decode()
    {
      GetHeader(true, part => part.GetHeader(true));
    }

    [Test, Obsolete]
    public void TestGetHeaderAsync()
    {
      GetHeader(false, part => part.GetHeaderAsync().Result);
    }

    [Test, Obsolete]
    public void TestGetHeaderAsync_Decode()
    {
      GetHeader(true, part => part.GetHeaderAsync(true).Result);
    }

    private void GetHeader(bool decode, Func<ImapMessagePart, IReadOnlyList<KeyValuePair<string, string>>> getHeader)
    {
      const string bodyStructureNested = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "((" + bodyStructureNested + " \"mixed\") \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.AreSame(message, part.Message);
        Assert.IsTrue(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        var expectedMessageBody = @"Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1.MIME]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1.MIME] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var headers = getHeader(part);

        responseTask.Wait();

        Assert.IsFalse(part.Client.IsBusy);

        Assert.IsNotEmpty(headers);
        Assert.AreEqual(3, headers.Count);

        if (decode)
          Assert.AreEqual("テスト", headers.First(pair => pair.Key == "Subject").Value);
        else
          Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", headers.First(pair => pair.Key == "Subject").Value);

        Assert.AreEqual("text/plain; charset=ISO-2022-JP", headers.First(pair => pair.Key == "Content-Type").Value);
      });
    }
  }
}

