using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

#if NET45
using Array = Smdn.ArrayExtensions; // for Empty<T>()
#endif

namespace Smdn.Net.Imap4.Client {
  internal static class TestCases {
    public static void WithAuthenticatedClient(Action<ImapPseudoServer, ImapClient> testAction,
                                               [CallerMemberName] string callerMemberName = "",
                                               [CallerFilePath] string sourceFilePath = "",
                                               [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient(Array.Empty<ImapCapability>(), testAction, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithAuthenticatedClient(ImapCapability[] serverCapabilities,
                                               Action<ImapPseudoServer, ImapClient> testAction,
                                               [CallerMemberName] string callerMemberName = "",
                                               [CallerFilePath] string sourceFilePath = "",
                                               [CallerLineNumber] int sourceLineNumber = 0)
    {
      var capas = new HashSet<ImapCapability>(serverCapabilities ?? Array.Empty<ImapCapability>());

      capas.Add(ImapCapability.Imap4Rev1);

      ImapPseudoServer.Start((server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     string.Format("* CAPABILITY {0}\r\n", string.Join(" ", capas.Select(c => c.Value))) +
                                     "$tag OK done\r\n");

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms.Clear();
          client.Profile.UseTlsIfAvailable = false;

          client.Connect("pass");

          server.DequeueRequest(); // CAPABILITY
          server.DequeueRequest(); // LOGIN
          server.DequeueRequest(); // CAPABILITY

          testAction(server, client);
        }
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static Task QueueResponseTask(ImapPseudoServer server,
                                         ImapClient client,
                                         Action queueResponse)
    {
      return Task.Run(() => {
        server.WaitForClientRequest();

        var isBusy = client.IsBusy;

        queueResponse();

        Assert.IsTrue(isBusy, "IsBusy before respond");
      });
    }

    public static void WithDisconnectedClient(Action<ImapPseudoServer, ImapClient> testAction,
                                              [CallerMemberName] string callerMemberName = "",
                                              [CallerFilePath] string sourceFilePath = "",
                                              [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n",
                                     isByeResponse: true); // stop server

        client.Refresh();

        server.DequeueRequest();

        testAction(server, client);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithLoggedOutClient(Action<ImapPseudoServer, ImapClient> testAction,
                                           [CallerMemberName] string callerMemberName = "",
                                           [CallerFilePath] string sourceFilePath = "",
                                           [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK logged out\r\n");

        client.Logout();

        server.DequeueRequest();

        testAction(server, client);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailbox(string name, Action<ImapPseudoServer, ImapMailbox> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailbox(null, null, ".", name, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailbox(string hierarchyDelimiter, string name, Action<ImapPseudoServer, ImapMailbox> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailbox(null, null, hierarchyDelimiter, name, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailbox(ImapCapability[] capabilities,
                                   ImapMailboxFlag[] flags,
                                   string name,
                                   Action<ImapPseudoServer, ImapMailbox> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMailbox(capabilities, flags, ".", name, action, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMailbox(ImapCapability[] capabilities,
                                   ImapMailboxFlag[] flags,
                                   string hierarchyDelimiter,
                                   string name,
                                   Action<ImapPseudoServer, ImapMailbox> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      capabilities = capabilities ?? Array.Empty<ImapCapability>();

      WithAuthenticatedClient(capabilities, (server, client) => {
        var joinedFlags = string.Join(" ", (flags ?? Array.Empty<ImapMailboxFlag>()).Select(f => f.Value));

        server.EnqueueTaggedResponse("LIST ",
                                     string.Format("* LIST ({0}) \"{1}\" {2}\r\n", joinedFlags, hierarchyDelimiter, name) +
                                     "$tag OK done\r\n");

        var mailbox = client.GetMailbox(name);

        Assert.IsNotNull(mailbox);
        Assert.AreEqual(name, mailbox.FullName);
        Assert.AreEqual(hierarchyDelimiter, mailbox.MailboxSeparator);

        var requested = server.DequeueRequest();

        if (capabilities.Any(capa => capa == ImapCapability.ListExtended)) {
          StringAssert.Contains("LIST (", requested);
          StringAssert.Contains(string.Format(") \"\" {0}", name), requested);
        }
        else {
          StringAssert.EndsWith(string.Format("LIST \"\" {0}\r\n", name), requested);
        }

        action(server, mailbox);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithOpenedMailbox(string mailboxName,
                                         Action<ImapPseudoServer, ImapOpenMailbox> action,
                                         [CallerMemberName] string callerMemberName = "",
                                         [CallerFilePath] string sourceFilePath = "",
                                         [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithOpenedMailbox(null,
                        mailboxName,
                        "$tag OK [READ-WRITE] done\r\n",
                        action,
                        callerMemberName,
                        sourceFilePath,
                        sourceLineNumber);
    }

    public static void WithOpenedMailbox(string mailboxName,
                                         string selectResponse,
                                         Action<ImapPseudoServer, ImapOpenMailbox> action,
                                         [CallerMemberName] string callerMemberName = "",
                                         [CallerFilePath] string sourceFilePath = "",
                                         [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithOpenedMailbox(null,
                        mailboxName,
                        selectResponse,
                        action,
                        callerMemberName,
                        sourceFilePath,
                        sourceLineNumber);
    }

    public static void WithOpenedMailbox(ImapCapability[] capabilities,
                                         string mailboxName,
                                         string selectResponse,
                                         Action<ImapPseudoServer, ImapOpenMailbox> action,
                                         [CallerMemberName] string callerMemberName = "",
                                         [CallerFilePath] string sourceFilePath = "",
                                         [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithAuthenticatedClient(capabilities, (server, client) => {
        server.EnqueueTaggedResponse("LIST \"\"",
                                     string.Format("* LIST () \"\" {0}\r\n", mailboxName) +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT " + mailboxName,
                                     selectResponse);

        using (var opened = client.OpenMailbox(mailboxName)) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          try {
            Assert.IsTrue(opened.IsOpen);
            Assert.AreEqual(opened.IsReadOnly, selectResponse.Contains("OK [READ-ONLY]"));
            Assert.AreEqual(mailboxName, opened.FullName);
            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.IsNotNull(opened.PermanentFlags);

            try {
              action(server, opened);
            }
            catch (Exception ex) {
              TestContext.Out.WriteLine(ex);

              throw;
            }
          }
          finally {
            // CLOSE
            if (client.IsConnected && opened.IsOpen)
              server.EnqueueTaggedResponse("CLOSE\r\n",
                                           "$tag OK done\r\n");

          }
        }
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithClosedMailbox(string mailboxName,
                                         Action<ImapPseudoServer, ImapOpenMailbox> action,
                                         [CallerMemberName] string callerMemberName = "",
                                         [CallerFilePath] string sourceFilePath = "",
                                         [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithOpenedMailbox(mailboxName, (server, mailbox) => {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpeningMailbox);

        action(server, mailbox);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessage(Action<ImapPseudoServer, ImapMessage> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithOpenedMailbox("INBOX", (server, mailbox) => {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID)\r\n",
                                     "* 1 FETCH (UID 1)\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L);

        Assert.IsTrue(message.IsAvailable);
        Assert.AreEqual(1L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.IsFalse(message.IsDeleted);
        Assert.IsNotNull(message.Url);
        Assert.AreEqual($"{mailbox.Url.AbsoluteUri}/;UID={message.Uid}", message.Url.AbsoluteUri);
        Assert.AreSame(mailbox, message.Mailbox);

        server.DequeueRequest(); // FETCH

        action(server, message);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessageMailboxClosed(Action<ImapPseudoServer, ImapMessage> action,
                                                [CallerMemberName] string callerMemberName = "",
                                                [CallerFilePath] string sourceFilePath = "",
                                                [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMessage((server, message) => {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = message.Mailbox;

        mailbox.Close();

        server.DequeueRequest(); // CLOSE

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(message.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));

        action(server, message);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessageVanished(Action<ImapPseudoServer, ImapMessage> action,
                                           [CallerMemberName] string callerMemberName = "",
                                           [CallerFilePath] string sourceFilePath = "",
                                           [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMessage((server, message) => {
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "* 1 EXPUNGE\r\n" +
                                     "$tag OK done\r\n");

        message.Client.Refresh();

        server.DequeueRequest(); // NOOP

        Assert.IsTrue(message.IsDeleted);
        Assert.IsFalse(message.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));

        action(server, message);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessage(string fetchResponse,
                                   ImapMessageFetchAttributeOptions options,
                                   Action<ImapPseudoServer, ImapMessage> action,
                                   [CallerMemberName] string callerMemberName = "",
                                   [CallerFilePath] string sourceFilePath = "",
                                   [CallerLineNumber] int sourceLineNumber = 0)
    {
      WithMessages(fetchResponse, options, (server, messages) => {
        action(server, messages.First());
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }

    public static void WithMessages(string fetchResponse,
                                    ImapMessageFetchAttributeOptions options,
                                    Action<ImapPseudoServer, IReadOnlyList<ImapMessage>> action,
                                    [CallerMemberName] string callerMemberName = "",
                                    [CallerFilePath] string sourceFilePath = "",
                                    [CallerLineNumber] int sourceLineNumber = 0)
    {
      var selectResp = "$tag OK [READ-WRITE] done\r\n";

      WithOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                        "INBOX",
                        selectResp,
                        (server, mailbox) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) ALL\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH $ (",
                                     fetchResponse);

        var messages = mailbox.GetMessages(ImapSearchCriteria.All, options: options);

        server.DequeueRequest(); // SEARCH
        server.DequeueRequest(); // FETCH

        action(server, messages);
      }, callerMemberName, sourceFilePath, sourceLineNumber);
    }
  }
}
