using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures {
  [TestFixture]
  public class IImapBodyStructureTests {
    [Test]
    public void TestGeneratePseudoBodyStructures()
    {
      var expectedStructures = new[] {
        (section: "1", type: new MimeType("text/plain")),
        (section: "2", type: new MimeType("multipart/alternative")),
        (section: "2.1", type: new MimeType("message/rfc822")),
        (section: "2.1.1", type: new MimeType("text/plain")),
        (section: "2.2", type: new MimeType("text/html")),
        (section: "3", type: new MimeType("message/rfc822")),
        (section: "3.1", type: new MimeType("text/plain")),
      };

      var bodyStructure = new MultiPartMessage(
        "mixed",
        new SinglePartMessage(MimeType.TextPlain), // 1
        new MultiPartMessage( // 2
          "alternative",
          new Rfc822Message( // 2.1
            new SinglePartMessage(MimeType.TextPlain) // 2.1.1
          ),
          new SinglePartMessage(MimeType.CreateTextType("html")) // 2.2
        ),
        new Rfc822Message( // 3
          new SinglePartMessage(MimeType.TextPlain) // 3.1
        )
      ).ToBodyStructure();

      Assert.AreEqual(string.Empty, bodyStructure.Section, $"section root");
      Assert.AreEqual(MimeType.MultipartMixed, bodyStructure.MediaType, $"media-type root");

      var index = 0;

      bodyStructure.Traverse(s => {
        Assert.AreEqual(expectedStructures[index].section, s.Section, $"section {index}");
        Assert.AreEqual(expectedStructures[index].type, s.MediaType, $"media-type {index}");

        index++;
      });

      Assert.AreEqual(expectedStructures.Length, index, "count of structures");
    }

    [Test]
    public void TestParentStructure()
    {
      var root = new MultiPartMessage(
        "mixed",
        new SinglePartMessage(MimeType.TextPlain), // 1
        new MultiPartMessage( // 2
          "alternative",
          new Rfc822Message( // 2.1
            new SinglePartMessage(MimeType.TextPlain) // 2.1.1
          ),
          new SinglePartMessage(MimeType.CreateTextType("html")) // 2.2
        ),
        new Rfc822Message( // 3
          new SinglePartMessage(MimeType.TextPlain) // 3.1
        )
      ).ToBodyStructure();

      Assert.IsNull(root.ParentStructure);

      var part1 = (root as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part2 = (root as ImapMultiPartBodyStructure).NestedPartStructures[1];
      var part3 = (root as ImapMultiPartBodyStructure).NestedPartStructures[2];

      Assert.AreSame(root, part1.ParentStructure);
      Assert.AreSame(root, part2.ParentStructure);
      Assert.AreSame(root, part3.ParentStructure);

      var part21 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part22 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[1];

      Assert.AreSame(part2, part21.ParentStructure);
      Assert.AreSame(part2, part22.ParentStructure);

      var part211 = (part21 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.AreSame(part21, part211.ParentStructure);

      var part31 = (part3 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.AreSame(part3, part31.ParentStructure);
    }

    [Test]
    public void TestSection()
    {
      var root = new MultiPartMessage(
        "mixed",
        new SinglePartMessage(MimeType.TextPlain), // 1
        new MultiPartMessage( // 2
          "alternative",
          new Rfc822Message( // 2.1
            new SinglePartMessage(MimeType.TextPlain) // 2.1.1
          ),
          new SinglePartMessage(MimeType.CreateTextType("html")) // 2.2
        ),
        new Rfc822Message( // 3
          new SinglePartMessage(MimeType.TextPlain) // 3.1
        )
      ).ToBodyStructure();

      Assert.AreEqual(string.Empty, root.Section);

      var part1 = (root as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part2 = (root as ImapMultiPartBodyStructure).NestedPartStructures[1];
      var part3 = (root as ImapMultiPartBodyStructure).NestedPartStructures[2];

      Assert.AreEqual("1", part1.Section);
      Assert.AreEqual("2", part2.Section);
      Assert.AreEqual("3", part3.Section);

      var part21 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part22 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[1];

      Assert.AreEqual("2.1", part21.Section);
      Assert.AreEqual("2.2", part22.Section);

      var part211 = (part21 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.AreEqual("2.1.1", part211.Section);

      var part31 = (part3 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.AreEqual("3.1", part31.Section);
    }

    [Test]
    public void TestIsMultiPart()
    {
      var root = new MultiPartMessage(
        "mixed",
        new SinglePartMessage(MimeType.TextPlain), // 1
        new MultiPartMessage( // 2
          "alternative",
          new Rfc822Message( // 2.1
            new SinglePartMessage(MimeType.TextPlain) // 2.1.1
          ),
          new SinglePartMessage(MimeType.CreateTextType("html")) // 2.2
        ),
        new Rfc822Message( // 3
          new SinglePartMessage(MimeType.TextPlain) // 3.1
        )
      ).ToBodyStructure();

      Assert.IsTrue(root.IsMultiPart);

      var part1 = (root as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part2 = (root as ImapMultiPartBodyStructure).NestedPartStructures[1];
      var part3 = (root as ImapMultiPartBodyStructure).NestedPartStructures[2];

      Assert.IsFalse(part1.IsMultiPart);
      Assert.IsTrue(part2.IsMultiPart);
      Assert.IsFalse(part3.IsMultiPart);

      var part21 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part22 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[1];

      Assert.IsFalse(part21.IsMultiPart);
      Assert.IsFalse(part22.IsMultiPart);

      var part211 = (part21 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.IsFalse(part211.IsMultiPart);

      var part31 = (part3 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.IsFalse(part31.IsMultiPart);
    }

    [Test]
    public void TestHasEncapsulatedMessage()
    {
      var root = new MultiPartMessage(
        "mixed",
        new SinglePartMessage(MimeType.TextPlain), // 1
        new MultiPartMessage( // 2
          "alternative",
          new Rfc822Message( // 2.1
            new SinglePartMessage(MimeType.TextPlain) // 2.1.1
          ),
          new SinglePartMessage(MimeType.CreateTextType("html")) // 2.2
        ),
        new Rfc822Message( // 3
          new SinglePartMessage(MimeType.TextPlain) // 3.1
        )
      ).ToBodyStructure();

      Assert.IsFalse(root.HasEncapsulatedMessage);

      var part1 = (root as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part2 = (root as ImapMultiPartBodyStructure).NestedPartStructures[1];
      var part3 = (root as ImapMultiPartBodyStructure).NestedPartStructures[2];

      Assert.IsFalse(part1.HasEncapsulatedMessage);
      Assert.IsFalse(part2.HasEncapsulatedMessage);
      Assert.IsTrue(part3.HasEncapsulatedMessage);

      var part21 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[0];
      var part22 = (part2 as ImapMultiPartBodyStructure).NestedPartStructures[1];

      Assert.IsTrue(part21.HasEncapsulatedMessage);
      Assert.IsFalse(part22.HasEncapsulatedMessage);

      var part211 = (part21 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.IsFalse(part211.HasEncapsulatedMessage);

      var part31 = (part3 as ImapMessageRfc822BodyStructure).EncapsulatedMessageStructure;

      Assert.IsFalse(part31.HasEncapsulatedMessage);
    }
  }
}
