using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures {
  [TestFixture]
  public class ImapBodyStructureUtilsTests {
    private IImapBodyStructure BuildMultiPart()
    {
      // [] multipart/mixed
      // [1] text/plain
      // [2] multipart/alternative
      // [2.1] text/plain
      // [2.2] text/html
      // [3] message/rfc822
      // [3.1] text/plain
      return new MultiPartMessage(
        "mixed",
        new SinglePartMessage(MimeType.TextPlain),
        new MultiPartMessage(
          "alternative",
          new SinglePartMessage(MimeType.TextPlain),
          new SinglePartMessage(MimeType.CreateTextType("html"))
        ),
        new Rfc822Message(
          new SinglePartMessage(MimeType.TextPlain)
        )
      ).ToBodyStructure();
    }

    private IImapBodyStructure BuildNestLevel3()
    {
      // [] multipart/mixed
      // [1] multipart/alternative
      // [1.1] multipart/alternative
      // [1.1.1] text/plain
      // [1.1.2] text/html
      // [1.2] message/rfc822
      // [1.2.1] text/plain
      // [2] message/rfc822
      // [2.1] text/plain
      return new MultiPartMessage(
        "mixed",
        new MultiPartMessage(
          "alternative",
          new MultiPartMessage(
            "alternative",
            new SinglePartMessage(MimeType.TextPlain),
            new SinglePartMessage(MimeType.CreateTextType("html"))
          ),
          new Rfc822Message(
            new SinglePartMessage(MimeType.TextPlain)
          )
        ),
        new Rfc822Message(
          new SinglePartMessage(MimeType.TextPlain)
        )
      ).ToBodyStructure();
    }

    [Test]
    public void TestTraverse_SinglePart()
    {
      var single = new SinglePartMessage(MimeType.TextPlain).ToBodyStructure();

      ImapBodyStructureUtils.Traverse(single, delegate(IImapBodyStructure s) {
        Assert.Fail("enumerated");
      });
    }

    [Test]
    public void TestTraverse_MultiPart()
    {
      var expectedSections = new[] {"1", "2", "2.1", "2.2", "3", "3.1"};
      var index = 0;

      ImapBodyStructureUtils.Traverse(BuildMultiPart(), delegate(IImapBodyStructure s) {
        Assert.AreEqual(expectedSections[index++], s.Section);
      });
    }

    [Test]
    public void TestTraverse_MultiPart_NestLevel3()
    {
      var expectedSections = new[] {"1", "1.1", "1.1.1", "1.1.2", "1.2", "1.2.1", "2", "2.1"};
      var index = 0;

      ImapBodyStructureUtils.Traverse(BuildNestLevel3(), delegate (IImapBodyStructure s) {
        Assert.AreEqual(expectedSections[index++], s.Section);
      });
    }

    [Test]
    public void TestFindSection_1()
    {
      var structure = BuildMultiPart();
      var found = structure.FindSection(3, 1);

      Assert.IsNotNull(found);
      Assert.AreEqual("3.1", found.Section);
      Assert.AreEqual(MimeType.TextPlain, found.MediaType);
      Assert.AreSame(found, structure.FindSection("3.1"));

      found = structure.FindSection(4, 1);

      Assert.IsNull(found);

      Assert.IsNull(structure.FindSection());
    }

    [Test]
    public void TestFindSection_2()
    {
      var structure = BuildMultiPart();
      var found = structure.FindSection("3");

      Assert.IsNotNull(found);
      Assert.AreEqual("3", found.Section);
      Assert.AreEqual(new MimeType("message/rfc822"), found.MediaType);
      Assert.AreSame(found, structure.FindSection(3));

      found = structure.FindSection("4.1");

      Assert.IsNull(found);

      Assert.IsNull(structure.FindSection((string)null));
      Assert.IsNull(structure.FindSection(string.Empty));
    }

    [Test]
    public void TestFindSection_3()
    {
      var structure = BuildNestLevel3();
      var found = structure.FindSection("1.1.2");

      Assert.IsNotNull(found);
      Assert.AreEqual("1.1.2", found.Section);
      Assert.AreEqual(new MimeType("text/html"), found.MediaType);
      Assert.AreSame(found, structure.FindSection(1, 1, 2));
    }

    [Test]
    public void TestFindSection_InvalidSection()
    {
      var structure = BuildMultiPart();

      Assert.Throws<ArgumentException>(() => structure.FindSection("0"));
      Assert.Throws<ArgumentException>(() => structure.FindSection("foo"));
    }

    [Test]
    public void TestFind()
    {
      var structure = BuildMultiPart();
      var found = structure.Find(delegate(IImapBodyStructure s) {
        return s is ImapMessageRfc822BodyStructure;
      });

      Assert.IsNotNull(found);
      Assert.AreEqual("3", found.Section);

      found = structure.Find(delegate(IImapBodyStructure s) {
        return s is ImapExtendedMessageRfc822BodyStructure;
      });

      Assert.IsNull(found);
    }

    [Test]
    public void TestFindAll()
    {
      var structure = BuildMultiPart();
      var found = structure.FindAll(delegate(IImapBodyStructure s) {
        return !s.IsMultiPart;
      });

      var expectedSections = new[] {"1", "2.1", "2.2", "3", "3.1"};
      var index = 0;

      Assert.IsNotNull(found);

      foreach (var s in found) {
        Assert.IsNotNull(s);
        Assert.AreEqual(expectedSections[index++], s.Section);
      }

      found = structure.FindAll(delegate(IImapBodyStructure s) {
        return s is IImapExtendedBodyStructure;
      });

      Assert.IsNotNull(found);
      Assert.AreEqual(0, found.Count());
    }

    [Test]
    public void TestFind_ByMediaType()
    {
      var structure = BuildMultiPart();
      var found = structure.Find(new MimeType("MESSAGE/RFC822"));

      Assert.IsNotNull(found);
      Assert.AreEqual("3", found.Section);

      found = structure.Find(MimeType.ApplicationOctetStream);

      Assert.IsNull(found);
    }

    [Test]
    public void TestFindAll_ByMediaType()
    {
      var structure = BuildMultiPart();
      var found = structure.FindAll(MimeType.TextPlain);

      var expectedSections = new[] {"1", "2.1", "3.1"};
      var index = 0;

      Assert.IsNotNull(found);

      foreach (var s in found) {
        Assert.IsNotNull(s);
        Assert.AreEqual(expectedSections[index++], s.Section);
      }

      found = structure.FindAll(MimeType.ApplicationOctetStream);

      Assert.IsNotNull(found);
      Assert.AreEqual(0, found.Count());
    }

    [Test]
    public void TestGetRootStructure1()
    {
      var structure = BuildMultiPart();

      Assert.AreSame(structure, structure.GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2.1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2.2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("3").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("3.1").GetRootStructure());
    }

    [Test]
    public void TestGetRootStructure2()
    {
      var structure = BuildNestLevel3();

      Assert.AreSame(structure, structure.GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1.1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1.1.1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1.1.2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1.2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("1.2.1").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2").GetRootStructure());
      Assert.AreSame(structure, structure.FindSection("2.1").GetRootStructure());
    }

    [Test]
    public void TestGetValidatedSectionSpecOfSectionPart_Valid()
    {
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("10"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("100"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.1"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.10"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.100"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.1.1"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("10.1.1"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("100.1.1"));

      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("3.1"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("4"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("4.2.2.2"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("9"));
      Assert.DoesNotThrow(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("9.9"));
    }

    [Test]
    public void TestGetValidatedSectionSpecOfSectionPart_Invalid()
    {
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("0"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("01"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("001"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("."));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(".1"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.01"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.001"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("01.1"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("001.1"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.1."));

      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("１"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1１"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.１"));
      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("1.1１"));

      Assert.Throws<ArgumentException>(() => ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart("A"));
    }

    [Test]
    public void TestGetContentFileName_FromContentTypeNameParameter()
    {
      var structure1 = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        mediaTypeParameters: new Dictionary<string, string> {{ "name", "attached.jpg" } }
      ).ToBodyStructure() as ImapSinglePartBodyStructure;

      Assert.IsTrue(structure1.Parameters.ContainsKey("name"));
      Assert.AreEqual("attached.jpg", structure1.Parameters["name"]);
      Assert.AreEqual("attached.jpg", structure1.GetContentFileName());

      var structure2 = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        mediaTypeParameters: new Dictionary<string, string> {{ "name", "=?ISO-2022-JP?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=" } }
      ).ToBodyStructure() as ImapSinglePartBodyStructure;

      Assert.IsTrue(structure2.Parameters.ContainsKey("name"));
      Assert.AreEqual("=?ISO-2022-JP?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", structure2.Parameters["name"]);
      Assert.AreEqual("添付ファイル.jpg", structure2.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_FromContentDispositionFilenameParameter()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        extended: true,
        disposition: new MessageContentDispsition("attached", new Dictionary<string, string> { { "filename", "attached.jpg" } })
      ).ToBodyStructure() as ImapExtendedSinglePartBodyStructure;

      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.AreEqual("attached.jpg", structure.Disposition.Parameters["filename"]);
      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename*"));
      Assert.AreEqual("attached.jpg", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_FromContentDispositionExtendedFilenameParameter()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        extended: true,
        disposition: new MessageContentDispsition("attached", new Dictionary<string, string> { { "filename*", "ISO-2022-JP''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67" } })
      ).ToBodyStructure() as ImapExtendedSinglePartBodyStructure;

      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename*"));
      Assert.AreEqual("ISO-2022-JP''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.Disposition.Parameters["filename*"]);
      Assert.AreEqual("添付ファイル.jpg", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_NoNameOrDisposition()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        extended: true
      ).ToBodyStructure() as ImapExtendedSinglePartBodyStructure;

      Assert.IsFalse(structure.Parameters.ContainsKey("name"));
      Assert.IsNull(structure.Disposition);

      Assert.IsNull(structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_WithExtendedMultiPartBodyStructure()
    {
      var structure = new MultiPartMessage(
        "mixed",
        new[] {new SinglePartMessage(new MimeType("image", "jpeg")) },
        extended: true,
        disposition: new MessageContentDispsition("attached", new Dictionary<string, string> { { "filename", "attached.eml" } })
      ).ToBodyStructure() as ImapExtendedMultiPartBodyStructure;

      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.AreEqual("attached.eml", structure.Disposition.Parameters["filename"]);
      Assert.AreEqual("attached.eml", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_WithExtendedMessageRfc822BodyStructure()
    {
      var structure = new Rfc822Message(
        new SinglePartMessage(new MimeType("image", "jpeg")),
        extended: true,
        disposition: new MessageContentDispsition("attached", new Dictionary<string, string> { { "filename", "attached.eml" } })
      ).ToBodyStructure() as ImapExtendedMessageRfc822BodyStructure;

      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.AreEqual("attached.eml", structure.Disposition.Parameters["filename"]);
      Assert.AreEqual("attached.eml", structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_NoNameOrFilenameParameter()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        extended: true,
        disposition: new MessageContentDispsition("attached", null)
      ).ToBodyStructure() as ImapExtendedSinglePartBodyStructure;

      Assert.IsFalse(structure.Parameters.ContainsKey("name"));
      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename"));

      Assert.IsNull(structure.GetContentFileName());
    }

    [Test]
    public void TestGetContentFileName_NameParameter_InvalidMimeEncoding()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        mediaTypeParameters: new Dictionary<string, string> { { "name", "=?ISO-2022-JP?X?xxxxxx?=" } }
      ).ToBodyStructure() as ImapSinglePartBodyStructure;

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", structure.GetContentFileName());
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", structure.GetContentFileName(null, false));

      Assert.Throws<FormatException>(() => structure.GetContentFileName(null, true));
    }

    [Test]
    public void TestGetContentFileName_NameParameter_InvalidCharset()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        mediaTypeParameters: new Dictionary<string, string> { { "name", "=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=" } }
      ).ToBodyStructure() as ImapSinglePartBodyStructure;

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", structure.GetContentFileName());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", structure.GetContentFileName(null, false));

      Assert.Throws<EncodingNotSupportedException>(() => structure.GetContentFileName(null, true));

      Assert.AreEqual("添付ファイル.jpg", structure.GetContentFileName(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return TestUtils.Encodings.Jis;
      }, true));
    }

    [Test]
    public void TestGetContentFileName_FilenameParameter_InvalidCharset()
    {
      var structure = new SinglePartMessage(
        new MimeType("image", "jpeg"),
        extended: true,
        disposition: new MessageContentDispsition("attached", new Dictionary<string, string>() { { "filename*", "X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67" } })
      ).ToBodyStructure() as ImapExtendedSinglePartBodyStructure;

      Assert.IsFalse(structure.Disposition.Parameters.ContainsKey("filename"));
      Assert.IsTrue(structure.Disposition.Parameters.ContainsKey("filename*"));
      Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.Disposition.Parameters["filename*"]);
      Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.GetContentFileName());
      Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", structure.GetContentFileName(null, false));

      Assert.Throws<EncodingNotSupportedException>(() => structure.GetContentFileName(null, true));

      Assert.AreEqual("添付ファイル.jpg", structure.GetContentFileName(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return TestUtils.Encodings.Jis;
      }, true));
    }
  }
}
