using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Protocol.DataStructures.BodyStructures {
  internal static class Generator {
    internal const string NIL = "NIL";

    public static IImapBodyStructure GenerateBodyStructure(MessageBase message)
    {
      var fetchResponse = $"* 1 FETCH (BODYSTRUCTURE ({message.ToBody()}))\r\n";

      const int maxRetry = 20;

      using (var receiver = new ImapPseudoResponseReceiver(fetchResponse)) {
        ImapResponse response = null;

        for (var n = 0; n < maxRetry; n++) {
          response = receiver.ReceiveResponse();

          if (response != null)
            break;
        }

        var data = response as ImapServerDataResponse;

        try {
          var staticAttrs = ImapServerDataResponseParser.FromFetch<ImapMessageStaticAttributes>(data);

          return staticAttrs.BodyStructure;
        }
        catch (ImapMalformedDataException) {
          Console.WriteLine($"malformed response: {fetchResponse}");
          throw;
        }
      }
    }

    internal static string EscapeOrNil(string val)
    {
      if (val == null)
        return NIL;

      if (val.Contains("\r") || val.Contains("\n")) {
        var length = val.Length;

        return $"~{{{length}}}\r\n{val}";
      }

      return "\"" + val.Replace("\"", "\\\"") + "\"";
    }
  }

  public class Address {
    public string Name { get; }
    public string AtDomainList { get; }
    public string Mailbox { get; }
    public string Host { get; }

    public Address(string name = null,
                          string atDomainList = null,
                          string mailbox = null,
                          string host = null)
    {
      Name = name;
      AtDomainList = atDomainList;
      Mailbox = mailbox;
      Host = host;
    }

    internal string ToAddress()
    {
      return $"({Generator.EscapeOrNil(Name)} {Generator.EscapeOrNil(AtDomainList)} {Generator.EscapeOrNil(Mailbox)} {Generator.EscapeOrNil(Host)})";
    }
  }

  public class MessageEnvelope {
    public static readonly MessageEnvelope NIL = new MessageEnvelope();

    public string Date { get; }
    public string Subject { get; }
    public IReadOnlyList<Address> From { get; }
    public IReadOnlyList<Address> Sender { get; }
    public IReadOnlyList<Address> ReplyTo { get; }
    public IReadOnlyList<Address> To { get; }
    public IReadOnlyList<Address> CC { get; }
    public IReadOnlyList<Address> BCC { get; }
    public string InReplyTo { get; }
    public string MessageId { get; }

    public MessageEnvelope(string date = null,
                           string subject = null,
                           IReadOnlyList<Address> from = null,
                           IReadOnlyList<Address> sender = null,
                           IReadOnlyList<Address> replyTo = null,
                           IReadOnlyList<Address> to = null,
                           IReadOnlyList<Address> cc = null,
                           IReadOnlyList<Address> bcc = null,
                           string inReplyTo = null,
                           string messageId = null)
    {
      Date = date;
      Subject = subject;
      From = from;
      Sender = sender;
      ReplyTo = replyTo;
      To = to;
      CC = cc;
      BCC = bcc;
      InReplyTo = inReplyTo;
      MessageId = messageId;
    }

    internal string ToEnvelope()
    {
      return $"({Generator.EscapeOrNil(Date)} {Generator.EscapeOrNil(Subject)} {ToAddressList(From)} {ToAddressList(Sender)} {ToAddressList(ReplyTo)} {ToAddressList(To)} {ToAddressList(CC)} {ToAddressList(BCC)} {Generator.EscapeOrNil(InReplyTo)} {Generator.EscapeOrNil(MessageId)})";
    }

    private static string ToAddressList(IReadOnlyList<Address> addressList)
    {
      if (addressList == null)
        return Generator.NIL;

      return "(" + string.Join(" ", addressList.Select(addr => addr.ToAddress())) + ")";
    }
  }

  public class MessageContentDispsition {
    public string Type { get; }
    public IReadOnlyDictionary<string, string> Parameters { get; }

    public MessageContentDispsition(string type, IReadOnlyDictionary<string, string>  parameters = null)
    {
      Type = type ?? throw new ArgumentNullException(nameof(type));
      Parameters = parameters;
    }

    internal string ToBodyFldDsp()
    {
      var bodyFldParam = Parameters == null ? null : "(" + string.Join(" ", Parameters.Select(pair => string.Concat(pair.Key, " ", pair.Value))) + ")";

      return $"({Type} {bodyFldParam ?? Generator.NIL})";
    }
  }

  public abstract class MessageBase {
    public IImapBodyStructure ToBodyStructure()
    {
      return Generator.GenerateBodyStructure(this);
    }

    internal abstract string ToBody();

    protected string BodyFldParam(IReadOnlyDictionary<string, string> parameters)
    {
      if (parameters == null)
        return Generator.NIL;

      // TODO: escape
      return "(" + string.Join(" ", parameters.Select(pair => string.Concat(pair.Key, " ", pair.Value))) + ")";
    }
  }

  public class SinglePartMessage : MessageBase {
    public MimeType MediaType { get; }
    public IReadOnlyDictionary<string, string> MediaTypeParameters { get; }
    public string Id { get; }
    public string Description { get; }
    public string Encoding { get; }
    public long? SizeInOctets { get; }
    public long? LineCount { get; }
    public string MD5 { get; }
    public MessageContentDispsition Disposition { get; }
    public string Language { get; }
    public Uri Location { get; }
    public IReadOnlyDictionary<string, string> Extensions { get; }

    private readonly bool extended;

    public SinglePartMessage(MimeType mediaType,
                             IReadOnlyDictionary<string, string> mediaTypeParameters = null,
                             string id = null,
                             string description = null,
                             string encoding = "7bit",
                             long? sizeInOctets = null,
                             long? lineCount = null,
                             bool extended = false,
                             string md5 = null,
                             MessageContentDispsition disposition = null,
                             string language = null,
                             Uri location = null,
                             IReadOnlyDictionary<string, string> extensions = null)
    {
      MediaType = mediaType ?? throw new ArgumentNullException(nameof(mediaType));
      MediaTypeParameters = mediaTypeParameters;
      Id = id;
      Description = description;
      Encoding = encoding;
      SizeInOctets = sizeInOctets;
      LineCount = lineCount;
      MD5 = md5;
      Disposition = disposition;
      Language = language;
      Location = location;
      Extensions = extensions;

      this.extended = extended;
    }

    internal override string ToBody()
    {
      var basicFields = $"{MediaType.Type} {MediaType.SubType} {BodyFldParam(MediaTypeParameters)} {Id ?? Generator.NIL} {Description ?? Generator.NIL} {Encoding ?? Generator.NIL} {SizeInOctets?.ToString() ?? "0"}";

      if (MediaType.TypeEqualsIgnoreCase("text"))
        basicFields += $" {LineCount?.ToString() ?? "0"}";

      var hasExtension = extended;
      var extensionFields = new StringBuilder();

      if (hasExtension || Extensions != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(Extensions == null ? Generator.NIL : throw new NotSupportedException())}");
      }

      if (hasExtension || Language != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(Language == null ? Generator.NIL : throw new NotSupportedException())}");
      }

      if (hasExtension || Disposition != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {Disposition?.ToBodyFldDsp() ?? Generator.NIL}");
      }

      if (hasExtension || MD5 != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {MD5 ?? Generator.NIL}");
      }

      if (hasExtension)
        return basicFields + extensionFields;

      return basicFields;
    }
  }

  public class Rfc822Message : SinglePartMessage {
    public MessageBase EncapsulatedMessage { get; }
    public MessageEnvelope EncapsulatedMessageEnvelope { get; }

    private readonly bool extended;

    public Rfc822Message(MessageBase encapsulatedMessage,
                         MessageEnvelope encapsulatedMessageEnvelope = null,
                         IReadOnlyDictionary<string, string> mediaTypeParameters = null,
                         string id = null,
                         string description = null,
                         string encoding = "7bit",
                         long? sizeInOctets = null,
                         string envelope = null,
                         long? lineCount = null,
                         bool extended = false,
                         string md5 = null,
                         MessageContentDispsition disposition = null,
                         string language = null,
                         Uri location = null,
                         IReadOnlyDictionary<string, string> extensions = null)
      : base(MimeType.MessageRfc822,
             mediaTypeParameters: mediaTypeParameters,
             id: id,
             description: description,
             encoding: encoding,
             sizeInOctets: sizeInOctets,
             lineCount: lineCount,
             md5: md5,
             disposition: disposition,
             language: language,
             location: location,
             extensions: extensions)
    {
      EncapsulatedMessage = encapsulatedMessage ?? throw new ArgumentNullException(nameof(encapsulatedMessage));
      EncapsulatedMessageEnvelope = encapsulatedMessageEnvelope;

      this.extended = extended;
    }

    internal override string ToBody()
    {
      var basicFields = $"message rfc822 {BodyFldParam(MediaTypeParameters)} {Id ?? Generator.NIL} {Description ?? Generator.NIL} {Encoding ?? Generator.NIL} {SizeInOctets?.ToString() ?? "0"} {(EncapsulatedMessageEnvelope ?? MessageEnvelope.NIL).ToEnvelope()} ({EncapsulatedMessage.ToBody()}) {LineCount?.ToString() ?? "0"}";

      var hasExtension = extended;
      var extensionFields = new StringBuilder();

      if (hasExtension || Extensions != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(Extensions == null ? Generator.NIL : throw new NotSupportedException())}");
      }

      if (hasExtension || Location != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {Location?.ToString() ?? Generator.NIL}");
      }

      if (hasExtension || Language != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(Language == null ? Generator.NIL : throw new NotSupportedException())}");
      }

      if (hasExtension || Disposition != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {Disposition?.ToBodyFldDsp() ?? Generator.NIL}");
      }

      if (hasExtension || MD5 != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {MD5 ?? Generator.NIL}");
      }

      if (hasExtension)
        return basicFields + extensionFields;

      return basicFields;
    }
  }

  public class MultiPartMessage : MessageBase {
    public MimeType MediaType { get; }
    public IReadOnlyList<MessageBase> MessageParts { get; }
    public IReadOnlyDictionary<string, string> MediaTypeParameters { get; }
    public MessageContentDispsition Disposition { get; }
    public string Language { get; }
    public Uri Location { get; }
    public IReadOnlyDictionary<string, string> Extensions { get; }

    private readonly bool extended = false;

    public MultiPartMessage(string multiPartSubType,
                            params MessageBase[] messageParts)
    {
      MediaType = MimeType.CreateMultipartType(multiPartSubType);
      MessageParts = messageParts ?? Enumerable.Empty<MessageBase>().ToArray();

      this.extended = false;
    }

    public MultiPartMessage(string multiPartSubType,
                            IReadOnlyList<MessageBase> messageParts,
                            bool extended = false,
                            IReadOnlyDictionary<string, string> mediaTypeParameters = null,
                            MessageContentDispsition disposition = null,
                            string language = null,
                            Uri location = null,
                            IReadOnlyDictionary<string, string> extensions = null)
    {
      MediaType = MimeType.CreateMultipartType(multiPartSubType);
      MessageParts = messageParts ?? Enumerable.Empty<MessageBase>().ToArray();
      MediaTypeParameters = mediaTypeParameters;
      Disposition = disposition;
      Language = language;
      Location = location;
      Extensions = extensions;

      this.extended = extended;
    }

    internal override string ToBody()
    {
      var basicFields = $"({string.Join(") (", MessageParts.Select(part => part.ToBody()))}) {MediaType.SubType}";
      var body = basicFields;

      var hasExtension = extended;
      var extensionFields = new StringBuilder();

      if (hasExtension || Extensions != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(Extensions == null ? Generator.NIL : throw new NotSupportedException())}");
      }

      if (hasExtension || Location != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {Location?.ToString() ?? Generator.NIL}");
      }

      if (hasExtension || Language != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(Language == null ? Generator.NIL : throw new NotSupportedException())}");
      }

      if (hasExtension || Disposition != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {Disposition?.ToBodyFldDsp() ?? Generator.NIL}");
      }

      if (hasExtension || MediaTypeParameters != null) {
        hasExtension = true;
        extensionFields.Insert(0, $" {(MediaTypeParameters == null ? Generator.NIL : string.Join(" ", MediaTypeParameters.Select(pair => pair.Key + " " + pair.Value)))}");
      }

      if (hasExtension)
        return basicFields + extensionFields;

      return body;
    }
  }
}
