using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  [TestFixture]
  public class ImapFetchRangeTests {
    [Test]
    public void TestAll()
    {
      Assert.AreEqual(0L, ImapFetchRange.All.Start);
      Assert.IsNull(ImapFetchRange.All.Length);
      Assert.IsFalse(ImapFetchRange.All.IsPartial);

      Assert.IsTrue(ImapFetchRange.All.Equals(new ImapFetchRange()));
      Assert.IsTrue(ImapFetchRange.All.Equals(default(ImapFetchRange)));

      Assert.AreEqual(ImapFetchRange.All.GetHashCode(), new ImapFetchRange().GetHashCode());
      Assert.AreEqual(ImapFetchRange.All.ToString(), new ImapFetchRange().ToString());
    }

    [Test]
    public void TestConstruct()
    {
      ImapFetchRange r;

      r = new ImapFetchRange(0L);

      Assert.AreEqual(0L, r.Start);
      Assert.IsNull(r.Length);
      Assert.IsFalse(r.IsLengthSpecified);
      Assert.IsFalse(r.IsPartial);
      Assert.AreEqual("<0>", r.ToString());
      Assert.AreEqual("/;PARTIAL=0", r.ToString("u"));

      r = new ImapFetchRange(0L, 1024L);

      Assert.AreEqual(0L, r.Start);
      Assert.AreEqual(1024L, r.Length);
      Assert.IsTrue(r.IsLengthSpecified);
      Assert.IsTrue(r.IsPartial);
      Assert.AreEqual("<0.1024>", r.ToString());
      Assert.AreEqual("<0.1024>", r.ToString("f"));
      Assert.AreEqual("/;PARTIAL=0.1024", r.ToString("u"));

      r = new ImapFetchRange(0L, (long)0xffffffff);

      Assert.AreEqual(0L, r.Start);
      Assert.AreEqual(4294967295L, r.Length);
      Assert.IsTrue(r.IsLengthSpecified);
      Assert.IsTrue(r.IsPartial);
      Assert.AreEqual("<0.4294967295>", r.ToString());
      Assert.AreEqual("<0.4294967295>", r.ToString("f"));
      Assert.AreEqual("/;PARTIAL=0.4294967295", r.ToString("u"));

      r = new ImapFetchRange((long)0xfffffffe, 1L);

      Assert.AreEqual(4294967294L, r.Start);
      Assert.AreEqual(1L, r.Length);
      Assert.IsTrue(r.IsLengthSpecified);
      Assert.IsTrue(r.IsPartial);
      Assert.AreEqual("<4294967294.1>", r.ToString());
      Assert.AreEqual("<4294967294.1>", r.ToString("f"));
      Assert.AreEqual("/;PARTIAL=4294967294.1", r.ToString("u"));
    }

    [TestCase(-1L)]
    [TestCase(0x100000000)]
    public void TestConstruct_OutOfRange_Start(long start)
    {
      var ex = Assert.Throws<ArgumentOutOfRangeException>(() => new ImapFetchRange(start, 1L));

      Assert.AreEqual("start", ex.ParamName);
    }

    [TestCase(-1L)]
    [TestCase(0L)]
    [TestCase(0x100000000)]
    public void TestConstruct_OutOfRange_Length(long length)
    {
      var ex = Assert.Throws<ArgumentOutOfRangeException>(() => new ImapFetchRange(0L, length));

      Assert.AreEqual("length", ex.ParamName);
    }

    [TestCase(0L, (long)0x100000000)]
    [TestCase(1L, (long)0x0ffffffff)]
    [TestCase((long)0x0ffffffff, 1L)]
    public void TestConstruct_OutOfRange_End(long start, long length)
    {
      Assert.Throws<ArgumentOutOfRangeException>(() => new ImapFetchRange(start, length));
    }

    [Test]
    public void TestEquals()
    {
      var r = new ImapFetchRange(0L, 1024L);
      var r2 = r;

      Assert.IsTrue(r.Equals(r2));
      Assert.IsTrue(r.Equals(new ImapFetchRange(0L, 1024L)));
      Assert.IsTrue((new ImapFetchRange(0L, 1024L)).Equals(r));

      Assert.IsFalse(r.Equals(0));
      Assert.IsFalse(r.Equals("<0.1024>"));
      Assert.IsFalse(r.Equals(new ImapFetchRange(1L, 1024L)));
      Assert.IsFalse(r.Equals(new ImapFetchRange(0L, 1023L)));
    }

    [Test]
    public void TestToString_FormatException()
    {
      foreach (var pair in new[] {
        new {Format = "f", Range = new ImapFetchRange(1024L)},
        new {Format = "f", Range = default(ImapFetchRange)},
      }) {
        Assert.Throws<FormatException>(() => pair.Range.ToString(pair.Format), "format = '{0}', range = {1}", pair.Format, pair.Range);
      }
    }
  }
}
