using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  [TestFixture]
  public class ImapMatchedSequenceSetTests {
    private ImapMatchedSequenceSet GetMatchedSingle() => GetMatchedSequenceSet("* SEARCH 1\r\n", false);
    private ImapMatchedSequenceSet GetMatchedEmpty() => GetMatchedSequenceSet("* SEARCH\r\n", false);
    private ImapMatchedSequenceSet GetMatched(string numberSet, bool uid = false) => GetMatchedSequenceSet($"* SEARCH {numberSet}\r\n", uid);

    private ImapMatchedSequenceSet GetMatchedSequenceSet(string searchResponse, bool uid)
    {
      ImapMatchedSequenceSet matched = null;

      Session.TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("SEARCH UNSEEN\r\n",
                                     searchResponse +
                                     "$tag OK SEARCH completed\r\n");

        if (uid)
          session.UidSearch(out matched, ImapSearchCriteria.Unseen);
        else
          session.Search(out matched, ImapSearchCriteria.Unseen);
      });

      return matched;
    }

    private ImapMatchedSequenceSet GetSavedResult(bool uid = false)
    {
      ImapMatchedSequenceSet matched = null;

      Session.TestCases.WithMailboxSelectedSession(new[] { "ESEARCH", "SEARCHRES" }, (session, server) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) UNSEEN\r\n",
                                     "$tag OK SEARCH completed, result saved\r\n");

        if (uid)
          session.UidESearch(out matched, ImapSearchCriteria.Unseen, ImapSearchResultOptions.Save);
        else
          session.ESearch(out matched, ImapSearchCriteria.Unseen, ImapSearchResultOptions.Save);
      });

      return matched;
    }

    private void AssertSerialization(ImapMatchedSequenceSet s)
    {
      TestUtils.Assert.IsSerializableBinaryFormat(s, d => {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual(s.IsSavedResult, d.IsSavedResult);
        Assert.AreEqual(s.Expression, d.Expression);
        Assert.AreEqual(s.IsUidSet, d.IsUidSet);
        Assert.AreEqual(s.IsEmpty, d.IsEmpty);
        Assert.AreEqual(s.IsSingle, d.IsSingle);
        Assert.AreEqual(s.ConcreteCount, d.ConcreteCount);
        Assert.AreEqual(s.Tag, d.Tag);
        Assert.AreEqual(s.HighestModSeq, d.HighestModSeq);
        Assert.AreEqual(s.Min, d.Min);
        Assert.AreEqual(s.Count, d.Count);
      });
    }

    [Test] 
    public void TestIsSavedResult() =>  Assert.IsFalse(GetMatched("1 2 3").IsSavedResult);

    [Test]
    public void TestIsSavedResult_SavedResult() =>  Assert.IsTrue(GetSavedResult().IsSavedResult);

    [Test]
    public void TestSerializeBinary()
    {
      AssertSerialization(GetMatched("1 2 3", uid: false));
      AssertSerialization(GetMatched("1 2 3", uid: true));
    }

    [Test]
    public void TestSerializeBinary_SavedResult()
    {
      AssertSerialization(GetSavedResult(uid: false));
      AssertSerialization(GetSavedResult(uid: true));
    }

    [Test]
    public void TestExpression()
    {
      Assert.AreEqual("1", GetMatched("1").Expression);
      Assert.AreEqual("1,2,3", GetMatched("1 2 3").Expression);
    }

    [Test]
    public void TestExpression_SavedResult() => Assert.AreEqual("$", GetSavedResult().Expression);

    [Test]
    public void TestIsUidSet()
    {
      Assert.IsTrue(GetMatched("1 2 3", uid: true).IsUidSet);
      Assert.IsFalse(GetMatched("1 2 3", uid: false).IsUidSet);
    }

    [Test]
    public void TestIsUidSet_SavedResult()
    {
      Assert.IsTrue(GetSavedResult(uid: true).IsUidSet);
      Assert.IsFalse(GetSavedResult(uid: false).IsUidSet);
    }

    [Test]
    public void TestIsEmpty()
    {
      Assert.IsFalse(GetMatched("1 2 3").IsEmpty);
      Assert.IsTrue(GetMatchedEmpty().IsEmpty);
    }

    [Test]
    public void TestIsEmpty_SavedResult() => Assert.IsFalse(GetSavedResult().IsEmpty);

    [Test]
    public void TestIsSingle()
    {
      Assert.IsFalse(GetMatched("1 2 3").IsSingle);
      Assert.IsTrue(GetMatchedSingle().IsSingle);
    }

    [Test]
    public void TestIsSingle_SavedResult() => Assert.IsFalse(GetSavedResult().IsSingle);

    [Test]
    public void TestConcreteCount()
    {
      Assert.AreEqual(3L, GetMatched("1 2 3").ConcreteCount);
      Assert.AreEqual(1L, GetMatchedSingle().ConcreteCount);
    }

    [Test]
    public void TestConcreteCount_SavedResult() => Assert.Less(GetSavedResult().ConcreteCount, 0);

    [Test]
    public void TestGetEnumerator()
    {
      CollectionAssert.AreEqual(new[] { 1L, 2L, 3L }, GetMatched("1 2 3"));
    }

    [Test]
    public void TestGetEnumerator_SavedResult()
    {
      Assert.Throws<NotSupportedException>(() => {
        foreach (var num in GetSavedResult()) {
          Assert.Fail("enumerated: {0}", num);
        }
      });
    }

    [Test]
    public void TestSplitIntoEach()
    {
      var matched = GetMatched("1 3 5 7 9");
      var expectedExpression = new[] {
        "1,3,5",
        "7,9",
      };
      var index = 0;

      foreach (var splitted in matched.SplitIntoEach(3)) {
        Assert.AreEqual(expectedExpression[index++], splitted.Expression);
      }

      Assert.AreEqual(expectedExpression.Length, index);
    }

    [Test]
    public void TestSplitIntoEach_SavedResult()
    {
      var matched = GetSavedResult();
      var expectedExpression = new[] {
        "$",
      };
      var index = 0;

      foreach (var splitted in matched.SplitIntoEach(3)) {
        Assert.AreEqual(expectedExpression[index++], splitted.Expression);
      }

      Assert.AreEqual(expectedExpression.Length, index);

      index = 0;

      foreach (var splitted in matched.SplitIntoEach(10)) {
        Assert.AreEqual(expectedExpression[index++], splitted.Expression);
      }

      Assert.AreEqual(expectedExpression.Length, index);
    }
  }
}
