using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Protocol.DataStructures {
  [TestFixture]
  public class ImapMetadataTests {
    [Test]
    public void TestJoinEntryName()
    {
      Assert.AreEqual("/shared/comment", ImapMetadata.JoinEntryName("shared", "comment"));
      Assert.AreEqual("/shared/vendor/foo/bar", ImapMetadata.JoinEntryName("shared", "vendor", "foo", "bar"));
    }

    [Test]
    public void TestJoinEntryName_ArgumentException_ComponentsTooFew()
    {
      const string expectedParamName = "hierarchicalEntryNames";
      ArgumentException ex;

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared"));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("private"));
      Assert.AreEqual(expectedParamName, ex.ParamName);
    }

    [Test]
    public void TestJoinEntryName_ArgumentException_ContainsNull()
    {
      const string expectedParamName = "hierarchicalEntryNames";
      ArgumentException ex;

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", null));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", new string[] { null }));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", "vendor", null));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", new[] { "vendor", null }));
      Assert.AreEqual(expectedParamName, ex.ParamName);
    }

    [Test]
    public void TestJoinEntryName_ArgumentException_ContainsInvalidEntryNameChars()
    {
      const string expectedParamName = "hierarchicalEntryNames";
      ArgumentException ex;

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", "%"));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", new string[] { "/" }));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", "vendor", "*"));
      Assert.AreEqual(expectedParamName, ex.ParamName);

      ex = Assert.Throws<ArgumentException>(() => ImapMetadata.JoinEntryName("shared", new[] { "vendor", "%" }));
      Assert.AreEqual(expectedParamName, ex.ParamName);
    }

    [Test]
    public void TestSplitEntryName()
    {
      Assert.AreEqual(new[] {"shared", "comment"},
                      ImapMetadata.SplitEntryName("/shared/comment"));
      Assert.AreEqual(new[] {"shared", "vendor", "foo", "bar"},
                      ImapMetadata.SplitEntryName("/shared/vendor/foo/bar"));

      Assert.Throws<ArgumentException>(() => ImapMetadata.SplitEntryName("shared/vendor"));
    }

    [Test]
    public void TestIsShared()
    {
      Assert.IsTrue(ImapMetadata.CreateQuoted("/shared/admin", "admin@example.net").IsShared);
      Assert.IsTrue(ImapMetadata.CreateQuoted("/SHARED/ADMIN", "admin@example.net").IsShared);
      Assert.IsFalse(ImapMetadata.CreateQuoted("/private/admin", "admin@example.net").IsShared);
    }

    [Test]
    public void TestIsPrivate()
    {
      Assert.IsTrue(ImapMetadata.CreateQuoted("/private/comment", "Really useful mailbox").IsPrivate);
      Assert.IsTrue(ImapMetadata.CreateQuoted("/PRIVATE/COMMENT", "Really useful mailbox").IsPrivate);
      Assert.IsFalse(ImapMetadata.CreateQuoted("/shared/comment", "Really useful mailbox").IsPrivate);
    }
  }
}
