using System;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Transmission;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCommandsIdleTests {
    private readonly TimeSpan testCaseIdleInterval = TimeSpan.FromMilliseconds(500);
    private readonly TimeSpan testCaseResponseInterval = TimeSpan.FromMilliseconds(50);

    [Test]
    public void TestIdle_ArgumentOutOfRange_MillisecondsTimeout()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => session.Idle(-2));
        Assert.Throws<ArgumentOutOfRangeException>(() => session.Idle(-2, response => true));
      });
    }

    [Test]
    public void TestIdle_ArgumentOutOfRange_Timeout()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.Throws<ArgumentOutOfRangeException>(() => session.Idle(TimeSpan.FromMilliseconds(-2)));
        Assert.Throws<ArgumentOutOfRangeException>(() => session.Idle(TimeSpan.FromMilliseconds(-2), response => true));
      });
    }

    [Test]
    public void TestIdle_Argument_InfinitePredicate()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.Throws<ArgumentException>(() => session.Idle(null));
        Assert.Throws<ArgumentException>(() => session.Idle(Timeout.Infinite, null));
        Assert.Throws<ArgumentException>(() => session.Idle(TimeSpan.FromMilliseconds(-1), null));
      });
    }

    [Test]
    public void TestIdleAsync_Argument_InfinitePredicate()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.Throws<ArgumentNullException>(() => session.IdleAsync(null));
      });
    }

    [Test]
    public void TestIdle()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "$tag OK done.\r\n");
        }, server);

        var interval = testCaseIdleInterval;

        TestUtils.Assert.Elapses(interval, () => {
          Assert.IsTrue(session.Idle(interval).IsSuccess);
        });

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdleAsync()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "$tag OK done.\r\n");
        }, server);

        using (var cts = new CancellationTokenSource()) {
          var idleTask = session.IdleAsync(cts.Token);

          Assert.IsTrue(session.IsIdling, "IsIdling after IdleAsync()");
          Assert.IsTrue(session.IsTransactionProceeding, "IsTransactionProceeding after IdleAsync()");
          Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after IdleAsync()");
          Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout, "ReceiveTimeout after IdleAsync()");
          Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after IdleAsync()");

          var interval = testCaseIdleInterval;

          TestUtils.Assert.Elapses(interval, () => {
            Assert.IsFalse(idleTask.Wait(interval));
          });

          cts.Cancel();

          Assert.IsTrue(idleTask.Result.IsSuccess);
        }

        Assert.IsFalse(session.IsIdling, "IsIdling after done idle");
        Assert.IsFalse(session.IsTransactionProceeding, "IsTransactionProceeding after done idle");
        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_No()
    {
      Idle_No((session, interval) => session.Idle(interval));
    }

    [Test]
    public void TestIdleAsync_No()
    {
      Idle_No((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          return session.IdleAsync(cts.Token).Result;
        }
      });
    }

    private void Idle_No(Func<ImapSession, TimeSpan, ImapCommandResult> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        server.EnqueueTaggedResponse("IDLE",
                                     "$tag NO not allowed at this time.\r\n");

        var interval = testCaseIdleInterval;

        TestUtils.Assert.NotElapse(interval, () => {
          Assert.IsFalse(testCode(session, interval).IsSuccess);
        });

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");
      });
    }

    [Test]
    public void TestIdle_StatusUpdate()
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        var initialExistsMessage = session.SelectedMailbox.ExistsMessage;

        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "* 2 EXPUNGE\r\n" +
                            "+ idling\r\n" +
                            "* 3 EXISTS\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "* 4 EXISTS\r\n" +
                                  "$tag OK IDLE terminated\r\n");
        }, server);

        var interval = testCaseIdleInterval;
        var receivedResponseCount = 0;

        TestUtils.Assert.Elapses(interval, () => {
          Assert.IsTrue(session.Idle(interval, resp => {
            receivedResponseCount++;

            // mailbox size is not updated while idling
            Assert.AreEqual(initialExistsMessage, session.SelectedMailbox.ExistsMessage);

            if (receivedResponseCount == 2)
              // receive timeout is infinite while idling
              Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);

            return true;
          }).IsSuccess);
        });

        Assert.IsFalse(session.IsIdling);
        Assert.AreEqual(3, receivedResponseCount, nameof(receivedResponseCount));
        Assert.AreEqual(4, session.SelectedMailbox.ExistsMessage);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");

        serverTask.Wait();
      });
    }

    private class IdleState {
      public ImapSession Session;
      public long CurrentMessageCount;
    }

    [Test]
    public void TestIdle_KeepIdleCallback()
    {
      TestCases.WithAuthenticatedSession(new[] { "IDLE" }, (session, server) => {
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "* FLAGS (\\Deleted \\Seen)\r\n" +
                                     "* 4 EXISTS\r\n" +
                                     "* 0 RECENT\r\n" +
                                     "* OK [UIDVALIDITY 1]\r\n" +
                                     "$tag OK SELECT completed\r\n");

        Assert.IsTrue(session.Select("INBOX").IsSuccess);

        Assert.AreEqual(4, session.SelectedMailbox.ExistsMessage);

        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var responseInterval = TimeSpan.FromMilliseconds(100);

        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "* 2 EXPUNGE\r\n" +
                            "* 3 EXISTS\r\n" +
                            "+ idling\r\n");

          Thread.Sleep(responseInterval);

          // ...time passes; another client expunges message 3...
          s.EnqueuePushResponse("* 3 EXPUNGE\r\n" +
                                "* 2 EXISTS\r\n");

          Thread.Sleep(responseInterval);

          // ...time passes; new mail arrives...
          s.EnqueuePushResponse("* 3 EXISTS\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "$tag OK IDLE terminated\r\n");
        }, server);

        var idleState = new IdleState();

        idleState.Session = session;
        idleState.CurrentMessageCount = session.SelectedMailbox.ExistsMessage;

        var interval = TimeSpan.FromMilliseconds(1000);
        var receivedResponseCount = 0;

        TestUtils.Assert.Elapses(responseInterval + responseInterval,
                                 interval,
                                 () => {
          var result = session.Idle(interval, receivedResponse => {
            receivedResponseCount++;

            if (3 <= receivedResponseCount && receivedResponseCount <= 5)
              // receive timeout is infinite while idling
              Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);

            var data = receivedResponse as ImapServerDataResponse;

            if (data == null)
              return true;

            if (data.Type == ImapServerDataResponseType.Expunge) {
              idleState.CurrentMessageCount--;
            }
            else if (data.Type == ImapServerDataResponseType.Exists) {
              var exists = ImapServerDataResponseParser.FromExists(data);

              if (idleState.CurrentMessageCount < exists) {
                return false;
              }
              else {
                idleState.CurrentMessageCount = exists;
              }
            }

            return true;
          });

          Assert.IsTrue(result.IsSuccess);
        });

        Assert.AreEqual(5, receivedResponseCount, nameof(receivedResponseCount));
        Assert.AreEqual(3L, session.SelectedMailbox.ExistsMessage);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK CLOSE completed\r\n");

        Assert.IsTrue(session.Close().IsSuccess);

        session.Disconnect(false);

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_KeepIdleCallback_StopIdleBeforeIdleStarted()
    {
      var callback = false;

      Idle_KeepIdleCallback_StopIdleBeforeIdleStarted((session, interval) => session.Idle(interval, response => {
        callback = true;
        return false;
      }));

      Assert.IsTrue(callback, "KeepIdleCallback called");
    }

    [Test]
    public void TestIdleAsync_KeepIdleCallback_StopIdleBeforeIdleStarted()
    {
      var callback = false;

      Idle_KeepIdleCallback_StopIdleBeforeIdleStarted((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          return session.IdleAsync(response => {
            callback = true;
            return false;
          }, cts.Token).Result;
        }
      });

      Assert.IsTrue(callback, "KeepIdleCallback called");
    }

    private void Idle_KeepIdleCallback_StopIdleBeforeIdleStarted(Func<ImapSession, TimeSpan, ImapCommandResult> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        var initialReceiveTimeout = Math.Max(1000, (int)testCaseResponseInterval.TotalMilliseconds * 2);
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "* 2 EXISTS\r\n"); // respond before continuation request

          Thread.Sleep(testCaseResponseInterval);

          s.EnqueuePushResponse("+ idling\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "$tag OK done.\r\n");
        }, server);

        var interval = testCaseIdleInterval;

        TestUtils.Assert.Elapses(testCaseResponseInterval, interval, () => {
          Assert.IsTrue(testCode(session, interval).IsSuccess);
        });

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_KeepIdleCallback_Timeout()
    {
      var callback = false;

      Idle_KeepIdleCallback_Timeout((session, interval) => session.Idle(interval, response => {
        callback = true;
        return true;
      }));

      Assert.IsTrue(callback, "KeepIdleCallback called");
    }

    [Test]
    public void TestIdleAsync_KeepIdleCallback_Timeout()
    {
      var callback = false;

      Idle_KeepIdleCallback_Timeout((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          return session.IdleAsync(response => {
            callback = true;
            return true;
          }, cts.Token).Result;
        }
      });

      Assert.IsTrue(callback, "KeepIdleCallback called");
    }

    private void Idle_KeepIdleCallback_Timeout(Func<ImapSession, TimeSpan, ImapCommandResult> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          Thread.Sleep(testCaseResponseInterval);

          s.EnqueuePushResponse("* 3 EXPUNGE\r\n" +
                                "* 2 EXISTS\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "$tag OK done.\r\n");
        }, server);

        var interval = testCaseIdleInterval;

        TestUtils.Assert.Elapses(interval, () => {
          Assert.IsTrue(testCode(session, interval).IsSuccess);
        });

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_KeepIdleCallback_Exception()
    {
      var callback = false;

      Idle_KeepIdleCallback_Exception((session, interval) => session.Idle(interval, response => {
        callback = true;
        throw new NullReferenceException();
      }));

      Assert.IsTrue(callback, "KeepIdleCallback called");
    }

    [Test]
    public void TestIdleAsync_KeepIdleCallback_Exception()
    {
      var callback = false;

      Idle_KeepIdleCallback_Exception((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          session.IdleAsync(response => {
            callback = true;
            throw new NullReferenceException();
          }, cts.Token).Wait();
        }
      });

      Assert.IsTrue(callback, "KeepIdleCallback called");
    }

    private void Idle_KeepIdleCallback_Exception(Action<ImapSession, TimeSpan> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          Thread.Sleep(testCaseResponseInterval);

          s.EnqueuePushResponse("* 3 EXPUNGE\r\n" +
                                "* 2 EXISTS\r\n");
        }, server);

        var interval = testCaseIdleInterval;

        TestUtils.Assert.NotElapse(interval, () => TestUtils.Assert.ThrowsOrAggregates<NullReferenceException>(() => testCode(session, interval)));

        ImapSessionTests.AssertSessionDisposed(session, server);

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_Bye()
    {
      Idle_Bye((session, interval) => session.Idle(interval));
    }

    [Test]
    public void TestIdleAsync_Bye()
    {
      Idle_Bye((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          return session.IdleAsync(cts.Token).Result;
        }
      });
    }

    private void Idle_Bye(Func<ImapSession, TimeSpan, ImapCommandResult> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          Thread.Sleep(testCaseResponseInterval);

          s.EnqueuePushResponse("* BYE Server shutting down.\r\n", false);
        }, server);

        var interval = testCaseIdleInterval;

        TestUtils.Assert.NotElapse(interval, () => {
          Assert.IsFalse(testCode(session, interval).IsSuccess);
        });

        ImapSessionTests.AssertSessionDisposed(session, server);

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_DisconnectedFromServer()
    {
      Idle_DisconnectedFromServer((session, interval) => session.Idle(interval));
    }

    [Test]
    public void TestIdleAsync_DisconnectedFromServer()
    {
      Idle_DisconnectedFromServer((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          session.IdleAsync(cts.Token).Wait();
        }
      });
    }

    private void Idle_DisconnectedFromServer(Action<ImapSession, TimeSpan> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          Thread.Sleep(testCaseResponseInterval);

          s.EnqueuePushResponse("* OK still here\r\n", true);
        }, server);

        var interval = testCaseIdleInterval;

        TestUtils.Assert.NotElapse(interval, () => {
          TestUtils.Assert.ThrowsOrAggregates<ImapConnectionException>(() => testCode(session, interval));
        });

        ImapSessionTests.AssertSessionDisposed(session, server);

        serverTask.Wait();
      });
    }

    [Test]
    public void TestIdle_SessionIsInNonAuthenticatedState()
    {
      Idle_SessionIsInNonAuthenticatedState((session, interval) => session.Idle(interval));
    }

    [Test]
    public void TestIdleAsync_SessionIsInNonAuthenticatedState()
    {
      Idle_SessionIsInNonAuthenticatedState((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          session.IdleAsync(cts.Token).Wait();
        }
      });
    }

    private void Idle_SessionIsInNonAuthenticatedState(Action<ImapSession, TimeSpan> testCode)
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        var initialSessionState = session.State;
        var interval = testCaseIdleInterval;

        TestUtils.Assert.NotElapse(interval, () => {
          Assert.Throws<ImapProtocolViolationException>(() => testCode(session, interval));
        });

        Assert.IsFalse(session.IsTransactionProceeding);
        Assert.AreEqual(initialSessionState, session.State);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");
      });
    }

    [Test]
    public void TestIdle_IdleIncapable()
    {
      Idle_IdleIncapable((session, interval) => session.Idle(interval));
    }

    [Test]
    public void TestIdleAsync_IdleIncapable()
    {
      Idle_IdleIncapable((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          session.IdleAsync(cts.Token).Wait();
        }
      });
    }

    private void Idle_IdleIncapable(Action<ImapSession, TimeSpan> testCode)
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.IsFalse(session.IsIdling);

        const int initialSendTimeout = 1001;
        const int initialReceiveTimeout = 1002;
        const int initialTransactionTimeout = 1003;

        session.SendTimeout = initialSendTimeout;
        session.ReceiveTimeout = initialReceiveTimeout;
        session.TransactionTimeout = initialTransactionTimeout;

        var initialSessionState = session.State;
        var interval = testCaseIdleInterval;

        TestUtils.Assert.NotElapse(interval, () => {
          var ex = Assert.Throws<ImapIncapableException>(() => testCode(session, interval));

          CollectionAssert.Contains(ex.RequiredCapabilities, ImapCapability.Idle);
        });

        Assert.IsFalse(session.IsTransactionProceeding);
        Assert.AreEqual(initialSessionState, session.State);

        Assert.AreEqual(initialSendTimeout, session.SendTimeout, "SendTimeout after done idle");
        Assert.AreEqual(initialReceiveTimeout, session.ReceiveTimeout, "ReceiveTimeout after done idle");
        Assert.AreEqual(initialTransactionTimeout, session.TransactionTimeout, "TransactionTimeout after done idle");
      });
    }

    [Test]
    public void TestIdle_AlreadyIdling()
    {
      Idle_AlreadyIdling((session, interval) => session.Idle(interval));
    }

    [Test]
    public void TestIdleAsync_AlreadyIdling()
    {
      Idle_AlreadyIdling((session, interval) => {
        using (var cts = new CancellationTokenSource(interval)) {
          session.IdleAsync(cts.Token).Wait();
        }
      });
    }

    private void Idle_AlreadyIdling(Action<ImapSession, TimeSpan> testCode)
    {
      TestCases.WithMailboxSelectedSession(new[] { "IDLE" }, (session, server) => {
        Assert.IsFalse(session.IsIdling);

        // IDLE
        var serverTask = Task.Factory.StartNew((state) => {
          var s = state as ImapPseudoServer;

          s.WaitForClientRequest();

          s.EnqueueResponse("IDLE\r\n",
                            "+ idling\r\n");

          s.WaitForClientRequest();

          s.EnqueueTaggedResponse("DONE\r\n",
                                  "$tag OK done.\r\n");
        }, server);

        using (var cts = new CancellationTokenSource()) {
          var idleTask = session.IdleAsync(cts.Token);

          Assert.IsTrue(session.IsIdling);
          Assert.IsTrue(session.IsTransactionProceeding);

          // attempt to start new IDLE transaction
          var interval = testCaseIdleInterval;

          TestUtils.Assert.NotElapse(interval, () => {
            Assert.Throws<InvalidOperationException>(() => testCode(session, interval));
          });

          Assert.IsTrue(session.IsIdling);
          Assert.IsTrue(session.IsTransactionProceeding);

          // receive timeout is infinite while idling
          Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);

          cts.Cancel();

          Assert.IsTrue(idleTask.Result.IsSuccess);

          Assert.IsFalse(session.IsIdling);
          Assert.IsFalse(session.IsTransactionProceeding);
        }

        serverTask.Wait();
      });
    }
  }
}
