using System;
using System.IO;
using System.Net;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Connection;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;
using Smdn.Text;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCommandsNonAuthenticatedStateTests {
    class NullCredential : ICredentialsByHost {
      public NetworkCredential GetCredential(string host, int port, string authType)
      {
        return null;
      }
    }

    [Test]
    public void TestLogin()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          session.HandlesIncapableAsException = true;

          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag NO incorrect\r\n");

          Assert.IsFalse(session.Login(new NetworkCredential("user", "pass")).IsSuccess);

          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag OK authenticated\r\n");

          Assert.IsTrue(session.Login(new NetworkCredential("user", "pass")).IsSuccess);

          Assert.AreEqual(string.Format("imap://user@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        }
      });
    }

    [Test]
    public void TestLogin_SelectAppropriateCredential_UsernameSpecified()
    {
      LoginSelectAppropriateCredential(true);
    }

    [Test]
    public void TestLogin_SelectAppropriateCredential_UsernameNotSpecified()
    {
      LoginSelectAppropriateCredential(false);
    }

    private void LoginSelectAppropriateCredential(bool specifyUsername)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          session.HandlesIncapableAsException = true;

          var credentials = new CredentialCache();

          credentials.Add("imap.example.net", 143, "LOGIN", new NetworkCredential("user", "pass1"));
          credentials.Add(server.Host, server.Port, "LOGIN", new NetworkCredential("user", "pass2"));
          credentials.Add(server.Host, server.Port, "PLAIN", new NetworkCredential("user", "pass3"));
          credentials.Add(server.Host, server.Port, string.Empty, new NetworkCredential("user", "pass4"));

          server.EnqueueTaggedResponse("LOGIN user pass4\r\n",
                                       "$tag OK authenticated\r\n");

          if (specifyUsername)
            Assert.IsTrue(session.Login(credentials, "user").IsSuccess);
          else
            Assert.IsTrue(session.Login(credentials).IsSuccess);

          Assert.AreEqual(string.Format("imap://user@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        }
      });
    }

    [Test]
    public void TestLogin_SelectAppropriateCredential_NotFound()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          session.HandlesIncapableAsException = true;

          var credentials = new CredentialCache();

          credentials.Add("imap.example.net", 143, "LOGIN", new NetworkCredential("user", "pass1"));
          credentials.Add(server.Host, server.Port, "LOGIN", new NetworkCredential("user", "pass2"));
          credentials.Add(server.Host, server.Port, "PLAIN", new NetworkCredential("user", "pass3"));
          credentials.Add(server.Host, server.Port, string.Empty, new NetworkCredential("user", "pass4"));

          var result = session.Login(credentials, "xxxx");

          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        }
      });
    }

    [Test]
    public void TestLogin_Credential_NotFound()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsFalse(session.Login(new NullCredential()).IsSuccess);
          Assert.IsFalse(session.Login(new NullCredential(), "user").IsSuccess);
        }
      });
    }

    [Test]
    public void TestLogin_Credential_UsernameNull()
    {
      LoginCredentialPropertyNull(new NetworkCredential(null, "password"),
                                  "\"\" password");
    }

    [Test]
    public void TestLogin_Credential_PasswordNull()
    {
      LoginCredentialPropertyNull(new NetworkCredential("username", (string)null),
                                  "username \"\"");
    }

    private void LoginCredentialPropertyNull(ICredentialsByHost credential, string expectedArgs)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");
        server.EnqueueTaggedResponse(string.Format("LOGIN {0}\r\n", expectedArgs),
                                     "$tag NO failed\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsFalse(session.Login(credential).IsSuccess);
        }
      });
    }

    [Test]
    public void TestLogin_ReissueCapability_DataAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK authenticated\r\n");

          Assert.IsTrue(session.Login(new NetworkCredential("user", "pass"), reissueCapability: true).IsSuccess);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestLogin_ReissueCapability_ResponseCodeAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag OK [CAPABILITY IMAP4rev1] authenticated\r\n");

          Assert.IsTrue(session.Login(new NetworkCredential("user", "pass"), reissueCapability: true).IsSuccess);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestLogin_ReissueCapability_NotAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag OK authenticated\r\n");
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(session.Login(new NetworkCredential("user", "pass"), reissueCapability: true).IsSuccess);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestLogin_LoginDisabled()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 LOGINDISABLED] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.Throws<ImapIncapableException>(() => session.Login(new NetworkCredential("user", "pass")));
        }
      });
    }

    [Test]
    public void TestLogin_HomeServerReferral_AsException()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: true)) {
          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag NO [REFERRAL IMAP://MIKE@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

          var ex = Assert.Throws<ImapLoginReferralException>(() => session.Login(new NetworkCredential("user", "pass")));

          Assert.IsTrue(ex.Message.Contains("Specified user is invalid on this server. Try SERVER2."));

          StringAssert.AreEqualIgnoringCase("IMAP://MIKE@SERVER2/", ex.ReferToUri.AbsoluteUri);

          TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate (ImapLoginReferralException deserialized) {
            Assert.IsNotNull(deserialized.ReferToUri);
            Assert.AreEqual(ex.ReferToUri, deserialized.ReferToUri);
          });
        }
      });
    }

    [Test]
    public void TestLogin_HomeServerReferral_AsError()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: false)) {
          server.EnqueueTaggedResponse("$tag NO [REFERRAL IMAP://MIKE@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

          Assert.IsFalse(session.Login(new NetworkCredential("user", "pass")).IsSuccess);
        }
      });
    }

    [Test]
    public void TestLogin_MailboxReferral_AsException()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: true)) {
          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag OK [REFERRAL IMAP://MATTHEW@SERVER2/] Specified user's personal mailboxes located on Server2, but public mailboxes are available.\r\n");

          var ex = Assert.Throws<ImapLoginReferralException>(() => session.Login(new NetworkCredential("user", "pass")));

          Assert.IsTrue(ex.Message.Contains("Specified user's personal mailboxes located on Server2, but public mailboxes are available."));

          StringAssert.AreEqualIgnoringCase("IMAP://MATTHEW@SERVER2/", ex.ReferToUri.AbsoluteUri);

          TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate (ImapLoginReferralException deserialized) {
            Assert.IsNotNull(deserialized.ReferToUri);
            Assert.AreEqual(ex.ReferToUri, deserialized.ReferToUri);
          });
        }
      });
    }

    [Test]
    public void TestLogin_MailboxReferral_AsError()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: false)) {
          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "$tag OK [REFERRAL IMAP://MATTHEW@SERVER2/] Specified user's personal mailboxes located on Server2, but public mailboxes are available.\r\n");

          Assert.IsTrue(session.Login(new NetworkCredential("user", "pass")).IsSuccess);
        }
      });
    }

    [Test]
    public void TestAuthenticate_MailboxReferral_AsException()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN LOGIN-REFERRALS] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: true)) {
          server.EnqueueTaggedResponse("AUTHENTICATE LOGIN\r\n",
                                       "$tag NO [REFERRAL IMAP://user;AUTH=GSSAPI@SERVER2/] Specified user is invalid on this server. Try SERVER2.\r\n");

          var ex = Assert.Throws<ImapLoginReferralException>(() => session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass")));

          Assert.IsTrue(ex.Message.Contains("Specified user is invalid on this server. Try SERVER2."));

          StringAssert.AreEqualIgnoringCase("IMAP://user;AUTH=GSSAPI@SERVER2/", ex.ReferToUri.AbsoluteUri);

          TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate (ImapLoginReferralException deserialized) {
            Assert.IsNotNull(deserialized.ReferToUri);
            Assert.AreEqual(ex.ReferToUri, deserialized.ReferToUri);
          });
        }
      });
    }

    [Test]
    public void TestAuthenticate_AuthLogin()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE LOGIN\r\n",
                                                    "+ " + Convert.ToBase64String(OctetEncoding.EightBits.GetBytes("Username:")) + "\r\n",
                                                    true);
          server.EnqueueLiteralContinuationResponse("+ " + Convert.ToBase64String(OctetEncoding.EightBits.GetBytes("Password:")) + "\r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK authenticated\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass")).IsSuccess);

          Assert.AreEqual("0000 AUTHENTICATE LOGIN\r\n",
                          server.DequeueRequest());

          var requested = server.DequeueRequest();

          Assert.AreEqual("user",
                          OctetEncoding.EightBits.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

          requested = server.DequeueRequest();

          Assert.AreEqual("pass",
                          OctetEncoding.EightBits.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

          Assert.AreEqual(string.Format("imap://user;AUTH=LOGIN@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_SelectAppropriateCredential_UsernameSpecified()
    {
      AuthenticateSelectAppropriateCredential(true);
    }

    [Test]
    public void TestAuthenticate_SelectAppropriateCredential_UsernameNotSpecified()
    {
      AuthenticateSelectAppropriateCredential(false);
    }

    private void AuthenticateSelectAppropriateCredential(bool specifyUsername)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE LOGIN\r\n",
                                                    "+ " + Convert.ToBase64String(OctetEncoding.EightBits.GetBytes("Username:")) + "\r\n",
                                                    true);
          server.EnqueueLiteralContinuationResponse("+ " + Convert.ToBase64String(OctetEncoding.EightBits.GetBytes("Password:")) + "\r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK authenticated\r\n");

          var credentials = new CredentialCache();

          credentials.Add("imap.example.net", 143, "LOGIN", new NetworkCredential("user", "pass1"));
          credentials.Add(server.Host, server.Port, "LOGIN", new NetworkCredential("user", "pass2"));
          credentials.Add(server.Host, server.Port, "PLAIN", new NetworkCredential("user", "pass3"));
          credentials.Add(server.Host, server.Port, string.Empty, new NetworkCredential("user", "pass4"));

          if (specifyUsername)
            Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Login, credentials, "user").IsSuccess);
          else
            Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Login, credentials).IsSuccess);

          Assert.AreEqual("0000 AUTHENTICATE LOGIN\r\n",
                          server.DequeueRequest());

          var requested = server.DequeueRequest();

          Assert.AreEqual("user",
                          OctetEncoding.EightBits.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

          requested = server.DequeueRequest();

          Assert.AreEqual("pass2",
                          OctetEncoding.EightBits.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))));

          Assert.AreEqual(string.Format("imap://user;AUTH=LOGIN@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_SelectAppropriateCredential_NotFound()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=LOGIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          var credentials = new CredentialCache();

          credentials.Add("imap.example.net", 143, "LOGIN", new NetworkCredential("user", "pass1"));
          credentials.Add(server.Host, server.Port, "LOGIN", new NetworkCredential("user", "pass2"));
          credentials.Add(server.Host, server.Port, "PLAIN", new NetworkCredential("user", "pass3"));
          credentials.Add(server.Host, server.Port, string.Empty, new NetworkCredential("user", "pass4"));

          var result = session.Authenticate(ImapAuthenticationMechanism.Login, credentials, "xxxx");

          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_AuthPlain()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE PLAIN\r\n",
                                                    "+ \r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK authenticated\r\n");

          var credential = new NetworkCredential("user", "pass", server.Host);

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Plain, credential).IsSuccess);

          Assert.AreEqual("0000 AUTHENTICATE PLAIN\r\n",
                          server.DequeueRequest());

          var requested = server.DequeueRequest();
          var userpass = OctetEncoding.EightBits.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))).Split(new[] { "\0" }, StringSplitOptions.RemoveEmptyEntries);

          Assert.AreEqual(credential.Domain, userpass[0]);
          Assert.AreEqual(credential.UserName, userpass[1]);
          Assert.AreEqual(credential.Password, userpass[2]);

          Assert.AreEqual(string.Format("imap://user;AUTH=PLAIN@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_AuthPlain_SaslIRCapable()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN SASL-IR] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueTaggedResponse("AUTHENTICATE PLAIN dGVzdAB0ZXN0AHRlc3Q=\r\n",
                                       "$tag OK authenticated\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Plain,
                                             new NetworkCredential("test", "test", "test")).IsSuccess);

          Assert.AreEqual(string.Format("imap://test;AUTH=PLAIN@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
        }
      });
    }

    [Test]
    public void TestAuthenticate_AuthCramMd5()
    {
      Authenticate_AuthCramMd5(false);
    }

    [Test]
    public void TestAuthenticate_AuthCramMd5_SaslIRCapable()
    {
      Authenticate_AuthCramMd5(true);
    }

    private void Authenticate_AuthCramMd5(bool saslIRCapable)
    {
      ImapPseudoServer.Start((server) => {
        if (saslIRCapable)
          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5] ImapSimulatedServer ready\r\n");
        else
          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 SASL-IR] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          var timestamp = ((long)DateTime.Now.Subtract(new DateTime(1970, 1, 1)).TotalMilliseconds).ToString();
          var challenge = OctetEncoding.EightBits.GetBytes(string.Format("<{0}@{1}>", timestamp, server.Host));

          server.EnqueueLiteralContinuationResponse("AUTHENTICATE CRAM-MD5\r\n",
                                                    "+ " + Convert.ToBase64String(challenge) + "\r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK authenticated\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.CRAMMD5, new NetworkCredential("user", "pass")).IsSuccess);

          /*
          var requested = server.DequeueRequest();
          var userkeyed = OctetEncoding.EightBits.GetString(Convert.FromBase64String(requested.Substring(0, requested.Length - 2))).Split(' ');

          Assert.AreEqual(username, userkeyed[0]);

          var keyed = new List<byte>();

          for (var index = 0; index < userkeyed[1].Length; index += 2) {
            keyed.Add(Convert.ToByte(userkeyed[1].Substring(index, 2), 16));
          }

          Assert.AreEqual(Convert.ToBase64String((new HMACMD5(OctetEncoding.EightBits.GetBytes(password))).ComputeHash(challenge)),
                          Convert.ToBase64String(keyed.ToArray()));
          */

          Assert.AreEqual(string.Format("imap://user;AUTH=CRAM-MD5@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_MechanismIncapable()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          session.HandlesIncapableAsException = true;

          var authMechanism = ImapAuthenticationMechanism.Plain;

          Assert.IsFalse(session.ServerCapabilities.IsCapable(authMechanism));

          var ex = Assert.Throws<ImapIncapableException>(() => session.Authenticate(authMechanism, new NetworkCredential("user", "pass")));

          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_AuthMechanismNotSupportedByClient()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          var authMechanism = new ImapAuthenticationMechanism("X-UNSUPPORTED");

          var ex = Assert.Throws<SaslMechanismNotSupportedException>(() => session.Authenticate(authMechanism, new NetworkCredential("user", "pass")));

          Assert.AreEqual("X-UNSUPPORTED", ex.MechanismName);

          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        }
      });
    }

    [Test]
    public void TestAuthenticate_InvalidResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE DIGEST-MD5\r\n",
                                                    "+ xxxxx-invalid-response-xxxxx\r\n",
                                                    true);

          Assert.Throws<FormatException>(() => session.Authenticate(ImapAuthenticationMechanism.DigestMD5,
                                                                    new NetworkCredential("test", "test", "test")));

          ImapSessionTests.AssertSessionDisposed(session, server);
        }
      });
    }

    [SaslMechanism("X-FAIL-INITIAL-EXCHANGE", true)]
    private class FailInitialExchangeSaslMechanism : SaslClientMechanism {
      public override bool ClientFirst => true;

      public override SaslExchangeStatus GetInitialResponse(out byte[] initialClientResponse)
      {
        initialClientResponse = null;
        return SaslExchangeStatus.Failed;
      }

      protected override SaslExchangeStatus Exchange(ByteString serverChallenge, out ByteString clientResponse) => throw new NotImplementedException();
    }

    [Test]
    public void TestAuthenticate_InitialExchangeFailed_SaslIRCapable()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 SASL-IR] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          var mechanism = new FailInitialExchangeSaslMechanism();

          mechanism.Credential = new NetworkCredential("test", "test", "test");

          var result = session.Authenticate(mechanism);

          Assert.IsTrue(result.IsConcludedByClientSide);
          Assert.IsFalse(result.IsSuccess);
        }
      });
    }

    [Test]
    public void TestAuthenticate_CancelExchanging()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE DIGEST-MD5\r\n",
                                                    "+ eD0xLHk9Mix6PTM=\r\n", // x=1,y=2,z=3
                                                    true);
          server.EnqueueTaggedResponse("$tag NO AUTHENTICATE failed.\r\n");

          session.Authenticate(ImapAuthenticationMechanism.DigestMD5,
                               new NetworkCredential("test", "test", "test"));

          Assert.AreEqual("0000 AUTHENTICATE DIGEST-MD5\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("*\r\n",
                          server.DequeueRequest());
        }
      });
    }

    [Test]
    public void TestAuthenticate_CancellationRequested()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          var cts = new CancellationTokenSource();

          server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                                 "+ eD0xLHk9Mix6PTM=\r\n", // x=1,y=2,z=3
                                 () => cts.Cancel(),
                                 null);
          server.EnqueueTaggedResponse("$tag BAD AUTHENTICATE cancelled.\r\n");

          var result = session.Authenticate(ImapAuthenticationMechanism.DigestMD5,
                                            new NetworkCredential("test", "test", "test"),
                                            cancellationToken: cts.Token);

          Assert.IsFalse(result.IsSuccess);
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

          Assert.AreEqual("0000 AUTHENTICATE DIGEST-MD5\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("*\r\n",
                          server.DequeueRequest());
        }
      });
    }

    [Test]
    public void TestAuthenticate_Credential_NotFound()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsFalse(session.Authenticate(ImapAuthenticationMechanism.DigestMD5, new NullCredential()).IsSuccess);
          Assert.IsFalse(session.Authenticate(ImapAuthenticationMechanism.DigestMD5, new NullCredential(), "user").IsSuccess);
        }
      });
    }

    [Test]
    public void TestAuthenticate_Credential_UsernameNull()
    {
      AuthenticateCredentialPropertyNull(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestAuthenticate_Credential_PasswordNull()
    {
      AuthenticateCredentialPropertyNull(new NetworkCredential("username", (string)null));
    }

    private void AuthenticateCredentialPropertyNull(ICredentialsByHost credential)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");
        server.EnqueueLiteralContinuationResponse("AUTHENTICATE PLAIN\r\n",
                                                  "+ \r\n",
                                                  true);
        server.EnqueueTaggedResponse("$tag NO failed\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsFalse(session.Authenticate(ImapAuthenticationMechanism.Plain, credential).IsSuccess);

          Assert.AreEqual("0000 AUTHENTICATE PLAIN\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("*\r\n",
                          server.DequeueRequest());
        }
      });
    }

    [Test]
    public void TestAuthenticate_Credential_UsernameNull_SaslIRCapable()
    {
      AuthenticateCredentialPropertyNullSaslIRCapable(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestAuthenticate_Credential_PasswordNull_SaslIRCapable()
    {
      AuthenticateCredentialPropertyNullSaslIRCapable(new NetworkCredential("username", (string)null));
    }

    private void AuthenticateCredentialPropertyNullSaslIRCapable(ICredentialsByHost credential)
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN SASL-IR] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsFalse(session.Authenticate(ImapAuthenticationMechanism.Plain, credential).IsSuccess);
        }
      });
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism()
    {
      using (var authMechanism = new SaslPseudoMechanism(2)) {
        authMechanism.Credential = new NetworkCredential("user", (string)null);

        AuthenticateSpecificMechanism(false, authMechanism);
      }
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism_SaslIRCapable()
    {
      using (var authMechanism = new SaslPseudoMechanism(2)) {
        authMechanism.Credential = new NetworkCredential("user", (string)null);

        AuthenticateSpecificMechanism(true, authMechanism);
      }
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism_WithNoCredential()
    {
      using (var authMechanism = new SaslPseudoMechanism(2)) {
        AuthenticateSpecificMechanism(false, authMechanism);
      }
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism_AlreadyExchanged()
    {
      using (var authMechanism = new SaslPseudoMechanism(2)) {
        byte[] clientResponse;

        authMechanism.Exchange(null, out clientResponse);

        Assert.AreNotEqual(SaslExchangeStatus.None,
                           authMechanism.ExchangeStatus);

        AuthenticateSpecificMechanism(true, authMechanism);
      }
    }

    private void AuthenticateSpecificMechanism(bool saslIRCapable, SaslClientMechanism authMechanism)
    {
      ImapPseudoServer.Start((server) => {
        if (saslIRCapable)
          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 SASL-IR] ImapSimulatedServer ready\r\n");
        else
          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          session.HandlesIncapableAsException = true;

          if (!saslIRCapable)
            server.EnqueueLiteralContinuationResponse("+ \r\n",
                                                      true);

          server.EnqueueLiteralContinuationResponse("+ \r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK done\r\n");

          Assert.IsTrue(session.Authenticate(authMechanism).IsSuccess);

          if (saslIRCapable) {
            Assert.AreEqual("0000 AUTHENTICATE X-PSEUDO-MECHANISM c3RlcDA=\r\n",
                            server.DequeueRequest());
            Assert.AreEqual("c3RlcDE=\r\n",
                            server.DequeueRequest());
          }
          else {
            Assert.AreEqual("0000 AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                            server.DequeueRequest());
            Assert.AreEqual("c3RlcDA=\r\n",
                            server.DequeueRequest());
            Assert.AreEqual("c3RlcDE=\r\n",
                            server.DequeueRequest());
          }

          // not disposed
          Assert.AreEqual(SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);

          Assert.AreEqual(string.Format("imap://{0};AUTH=X-PSEUDO-MECHANISM@{1}:{2}/",
                                        (authMechanism.Credential == null) ? null : authMechanism.Credential.UserName,
                                        server.Host,
                                        server.Port),
                          session.Authority.AbsoluteUri);
        }
      });
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism_ArgumentNull()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          session.HandlesIncapableAsException = true;

          SaslClientMechanism authMechanism = null;

          Assert.Throws<ArgumentNullException>(() => session.Authenticate(authMechanism));
        }
      });
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism_CancellationRequested()
    {
      using (var authMechanism = new SaslPseudoMechanism(1)) {
        ImapPseudoServer.Start((server) => {
          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapSimulatedServer ready\r\n");

          using (var session = new ImapSession(server.Host, server.Port)) {
            session.HandlesIncapableAsException = true;

            var cts = new CancellationTokenSource();

            server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                                   "+ \r\n",
                                   () => cts.Cancel(),
                                   null);
            server.EnqueueTaggedResponse("$tag BAD AUTHENTICATE cancelled\r\n");

            Assert.IsFalse(session.Authenticate(authMechanism, false, cts.Token).IsSuccess);
            Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

            Assert.AreEqual("0000 AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                          server.DequeueRequest());
            Assert.AreEqual("*\r\n",
                          server.DequeueRequest());

            // not started
            Assert.AreEqual(SaslExchangeStatus.None,
                          authMechanism.ExchangeStatus);
          }
        });
      }
    }

    [Test]
    public void TestAuthenticate_ReissueCapability_DataAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE LOGIN\r\n",
                                                    "+ \r\n",
                                                    true);
          server.EnqueueLiteralContinuationResponse("+ \r\n",
                                                    true);
          server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK authenticated\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass"), reissueCapability: true).IsSuccess);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestAuthenticate_ReissueCapability_ResponseCodeAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE LOGIN\r\n",
                                                    "+ \r\n",
                                                    true);
          server.EnqueueLiteralContinuationResponse("+ \r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK [CAPABILITY IMAP4rev1] authenticated\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass"), reissueCapability: true).IsSuccess);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestAuthenticate_ReissueCapability_NotAdvertised()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          server.EnqueueLiteralContinuationResponse("AUTHENTICATE LOGIN\r\n",
                                                    "+ \r\n",
                                                    true);
          server.EnqueueLiteralContinuationResponse("+ \r\n",
                                                    true);
          server.EnqueueTaggedResponse("$tag OK authenticated\r\n");
          server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass"), reissueCapability: true).IsSuccess);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestAuthenticate_OkWithCapabilityResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=PLAIN SASL-IR] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(3, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.SaslIR));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

          server.EnqueueTaggedResponse("AUTHENTICATE PLAIN dGVzdAB0ZXN0AHRlc3Q=\r\n",
                                       "$tag OK [CAPABILITY IMAP4rev1] authenticated\r\n");

          Assert.IsTrue(session.Authenticate(ImapAuthenticationMechanism.Plain,
                                             new NetworkCredential("test", "test", "test")).IsSuccess);

          Assert.AreEqual(string.Format("imap://test;AUTH=PLAIN@{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);

          Assert.AreEqual(1, session.ServerCapabilities.Count);
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestLogin_InAuthenticatedState()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var result = session.Login(new NetworkCredential("user", "pass"));

        Assert.IsTrue(result.IsConcludedByClientSide);
        Assert.IsTrue(result.IsSuccess);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      });
    }

    [Test]
    public void TestAuthenticate_InAuthenticatedState()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        var result = session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass"));

        Assert.IsTrue(result.IsConcludedByClientSide);
        Assert.IsTrue(result.IsSuccess);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
      });
    }

    [Test]
    public void TestLogin_InSelectedState()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var selectedMailbox = session.SelectedMailbox;

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        var result = session.Login(new NetworkCredential("user", "pass"));

        Assert.IsTrue(result.IsConcludedByClientSide);
        Assert.IsTrue(result.IsSuccess);

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        Assert.AreSame(selectedMailbox, session.SelectedMailbox);
      });
    }

    [Test]
    public void TestAuthenticate_InSelectedState()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var selectedMailbox = session.SelectedMailbox;

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        var result = session.Authenticate(ImapAuthenticationMechanism.Login, new NetworkCredential("user", "pass"));

        Assert.IsTrue(result.IsConcludedByClientSide);
        Assert.IsTrue(result.IsSuccess);

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.IsNotNull(session.SelectedMailbox);

        Assert.AreSame(selectedMailbox, session.SelectedMailbox);
      });
    }

    [Test]
    public void TestStartTls()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.HandlesIncapableAsException = true;
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        Assert.IsTrue(session.Capability().IsSuccess);

        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK Begin TLS negotiation now\r\n");

        var prevAuthority = session.Authority.AbsoluteUri;
        var streamUpgraded = false;

        Assert.IsFalse(session.IsSecureConnection);
        Assert.IsTrue(session.StartTls((connection, baseStream, cancellationToken) => {
          streamUpgraded = true;
          return Task.FromResult<Stream>(baseStream); // TODO: return SslStream
        }).IsSuccess);

        Assert.IsTrue(streamUpgraded, "stream upgraded");
        Assert.IsTrue(session.IsSecureConnection, "IsSecureConnection");

        Assert.AreEqual(prevAuthority, session.Authority.AbsoluteUri);
      });
    }

    [Test]
    public void TestStartTls_ExceptionWhileUpgrading()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.HandlesIncapableAsException = true;
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        Assert.IsTrue(session.Capability().IsSuccess);

        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK Begin TLS negotiation now\r\n");

        var ex = Assert.Throws<ImapSecureConnectionException>(delegate {
          session.StartTls((connection, baseStream, cancellationToken) => {
            throw new System.Security.Authentication.AuthenticationException("failed");
          });
        });

        Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);

        ImapSessionTests.AssertSessionDisposed(session, server);
      });
    }

    [Test]
    public void TestStartTls_Incapable()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        Assert.IsTrue(session.Capability().IsSuccess);

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(delegate {
          session.StartTls((connection, baseStream, cancellationToken) => {
            streamUpgraded = true;
            return Task.FromResult<Stream>(baseStream); // TODO: return SslStream
          });
        });

        Assert.IsFalse(streamUpgraded, "stream upgraded");
        Assert.IsFalse(session.IsSecureConnection, "IsSecureConnection");
      });
    }

    [Test]
    public void TestStartTls_BadResponse()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag BAD command unknown\r\n");

        var prevAuthority = session.Authority.AbsoluteUri;

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        Assert.IsFalse(session.StartTls((connection, baseStream, cancellationToken) => {
          streamUpgraded = true;
          return Task.FromResult<Stream>(baseStream); // TODO: return SslStream
        }).IsSuccess);

        Assert.IsFalse(streamUpgraded, "stream upgraded");

        Assert.AreEqual(prevAuthority, session.Authority.AbsoluteUri);
        Assert.IsFalse(session.IsSecureConnection);
      });
    }
  }
}
