using System;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.DataStructures;
using Smdn.Net.Imap4.Protocol.DataStructures.MessageAttributes;
using Smdn.Net.Imap4.Protocol.Transmission;
using Smdn.Net.Imap4.Urls;

namespace Smdn.Net.Imap4.Protocol.Session {
  [TestFixture]
  public class ImapSessionCommandsSelectedStateTests {
    [Test]
    public void TestClose_InNonAuthenticatedState()
    {
      TestCases.WithConnectedSession(session => {
        Assert.Throws<ImapProtocolViolationException>(() => session.Close());
      });
    }

    [Test]
    public void TestClose_InAuthenticatedState()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);

        var result = session.Close();

        Assert.IsTrue(result.IsConcludedByClientSide);
        Assert.IsTrue(result.IsSuccess);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      });
    }

    [Test]
    public void TestCheck_StatusUpdate()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // CHECK transaction
        server.EnqueueTaggedResponse("CHECK\r\n",
                                     "* 23 EXISTS\r\n" +
                                     "* 1 RECENT\r\n" +
                                     "$tag OK CHECK completed\r\n");

        Assert.IsTrue(session.Check().IsSuccess);

        Assert.AreEqual(23L, session.SelectedMailbox.ExistsMessage, "selected mailbox exist message count");
        Assert.AreEqual(1L, session.SelectedMailbox.RecentMessage, "selected mailbox recent message count");
      });
    }

    [Test]
    public void TestUnselect()
    {
      TestCases.WithMailboxSelectedSession(new[] {"UNSELECT"}, (session, server) => {
        // EXPUNGE transaction
        server.EnqueueTaggedResponse("UNSELECT\r\n",
                                     "$tag OK UNSELECT completed\r\n");

        Assert.IsTrue(session.Unselect().IsSuccess);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      });
    }

    [Test]
    public void TestSearch()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // SEARCH transaction
        server.EnqueueTaggedResponse("SEARCH UNSEEN\r\n",
                                     "* SEARCH 2 84 882\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.Unseen).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 2, 84, 882 }, matched);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.AreEqual("2,84,882", d.Expression);
        });
      });
    }

    [Test]
    public void TestSearch_WithLiteral_DefaultCharset()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("SEARCH CHARSET us-ascii ",
                                     "* SEARCH 2 84 882\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.Subject("subject")).IsSuccess);

        Assert.AreEqual("0004 SEARCH CHARSET us-ascii SUBJECT {7}\r\n" +
                        "subject\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("2,84,882", d.Expression);
        });
      });
    }

    [Test]
    public void TestSearch_WithLiteral_SpecifiedCharset()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("SEARCH CHARSET shift_jis ",
                                     "* SEARCH 2 84 882\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.Subject("件名"), TestUtils.Encodings.ShiftJis).IsSuccess);

        Assert.AreEqual("0004 SEARCH CHARSET shift_jis SUBJECT {4}\r\n" +
                        "\x008c\x008f\x0096\x00bc\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("2,84,882", d.Expression);
        });
      });
    }

    [Test]
    public void TestSearch_NothingMatched()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("SEARCH CHARSET utf-8 ",
                                     "* SEARCH\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.Text("string not in mailbox"), Encoding.UTF8).IsSuccess);

        Assert.AreEqual("0004 SEARCH CHARSET utf-8 TEXT {21}\r\n" +
                        "string not in mailbox\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsTrue(d.IsEmpty);
        });
      });
    }

    [Test]
    public void TestSearch_WithCriteriaFromUri()
    {
      TestCases.WithMailboxSelectedSession(new[] {"LITERAL+"}, (session, server) => {
        var b = new ImapUriBuilder(session.SelectedMailbox.Url);

        b.SearchCriteria = ImapSearchCriteria.From("差出人") & (ImapSearchCriteria.Seen | ImapSearchCriteria.Subject("件名"));
        b.Charset = Encoding.UTF8;

        server.EnqueueTaggedResponse("SEARCH CHARSET utf-8 ",
                                     "* SEARCH 2 84 882\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.FromUri(b.Uri)).IsSuccess);

        Assert.AreEqual("0004 SEARCH CHARSET utf-8 FROM {9+}\r\n" +
                        "\x00E5\x00B7\x00AE\x00E5\x0087\x00BA\x00E4\x00BA\x00BA OR (SEEN) (SUBJECT {6+}\r\n" +
                        "\x00E4\x00BB\x00B6\x00E5\x0090\x008D)\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
        });
      });
    }

    [Test]
    public void TestSearch_BadCharset()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("SEARCH CHARSET shift_jis BODY {8}",
                                     "$tag NO [BADCHARSET (ONDUL EUC-JP)] Unknown charset\r\n");

        var result = session.Search(out var matched, ImapSearchCriteria.Body("ほげほげ"), TestUtils.Encodings.ShiftJis);

        Assert.IsFalse(result.IsSuccess);
        Assert.AreEqual(ImapResponseCode.BadCharset, result.TaggedStatusResponse.ResponseText.Code);

        var supportedCharsets = ImapResponseTextParser.FromBadCharset(result.TaggedStatusResponse.ResponseText);

        Assert.AreEqual(2, supportedCharsets.Count);
        Assert.AreEqual("ONDUL", supportedCharsets[0]);
        Assert.AreEqual("EUC-JP", supportedCharsets[1]);
      });
    }

    [Test]
    public void TestSearch_CondStore_ModSeq()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // SEARCH transaction
        server.EnqueueTaggedResponse("SEARCH MODSEQ 620162338\r\n",
                                     "* SEARCH 2 5 6 (MODSEQ 917162500)\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.ModSeq(620162338UL)).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        CollectionAssert.AreEqual(new long[] { 2, 5, 6 }, matched);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.AreEqual("2,5,6", d.Expression);
          Assert.AreEqual(917162500UL, d.HighestModSeq);
        });
      });
    }

    [Test]
    public void TestSearch_CondStore_ModSeq_WithEntry()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // SEARCH transaction
        server.EnqueueTaggedResponse("SEARCH MODSEQ \"/flags/\\\\Draft\" all 620162338\r\n",
                                     "* SEARCH 2 5 6 7 11 12 18 19 20 23 (MODSEQ 917162500)\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        Assert.IsTrue(session.Search(out var matched, ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft)).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        CollectionAssert.AreEqual(new long[] { 2, 5, 6, 7, 11, 12, 18, 19, 20, 23 }, matched);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.AreEqual("2,5,6,7,11,12,18,19,20,23", d.Expression);
          Assert.AreEqual(917162500UL, d.HighestModSeq);
        });
      });
    }

    [Test]
    public void TestSearch_CondStore_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ImapIncapableException>(() => session.Search(out _, ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft)));
      });
    }

    [Test]
    public void TestESearch()
    {
      TestCases.WithMailboxSelectedSession(new[] {"ESEARCH"}, (session, server) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (MIN COUNT) CHARSET us-ascii FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                     "* ESEARCH (TAG \"0004\") MIN 2 COUNT 3\r\n" +
                                     "$tag OK SEARCH completed\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        Assert.IsTrue(session.ESearch(out var matched, criteria, ImapSearchResultOptions.Min + ImapSearchResultOptions.Count).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);
        Assert.IsNotNull(matched.Tag);
        Assert.AreEqual("0004", matched.Tag);
        Assert.IsNotNull(matched.Min);
        Assert.AreEqual(2, matched.Min);
        Assert.IsNotNull(matched.Count);
        Assert.AreEqual(3, matched.Count);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsTrue(d.IsEmpty);
          Assert.IsNotNull(d.Tag);
          Assert.AreEqual("0004", d.Tag);
          Assert.IsNotNull(d.Min);
          Assert.AreEqual(2, d.Min);
          Assert.IsNotNull(d.Count);
          Assert.AreEqual(3, d.Count);
        });
      });
    }

    [Test]
    public void TestESearch_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ImapIncapableException>(() => session.ESearch(out _, ImapSearchCriteria.All, ImapSearchResultOptions.Min));
      });
    }

    [Test]
    public void TestESearch_Searchres()
    {
      TestCases.WithMailboxSelectedSession(new[] {"ESEARCH", "SEARCHRES"}, (session, server) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) CHARSET us-ascii FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                     "$tag OK SEARCH completed, result saved\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        Assert.IsTrue(session.ESearch(out var matched, criteria, ImapSearchResultOptions.Save).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.IsTrue(d.IsSavedResult);
        });

        server.EnqueueTaggedResponse("FETCH $ (UID)\r\n",
                                     "* 2 FETCH (UID 14)\r\n" +
                                     "* 84 FETCH (UID 100)\r\n" +
                                     "* 882 FETCH (UID 1115)\r\n" +
                                     "$tag OK completed\r\n");

        Assert.IsTrue(session.Fetch(matched, ImapFetchDataItems.Uid, out ImapMessageFullAttributes[] _).IsSuccess);
      });
    }

    [Test]
    public void TestESearch_Searchres_WithSavedResult()
    {
      TestCases.WithMailboxSelectedSession(new[] {"ESEARCH", "SEARCHRES"}, (session, server) => {
        server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) CHARSET us-ascii FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                     "$tag OK SEARCH completed, result saved\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        Assert.IsTrue(session.ESearch(out var matched, criteria, ImapSearchResultOptions.Save).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.IsTrue(d.IsSavedResult);
        });

        server.EnqueueTaggedResponse("UID SEARCH UID $ SMALLER 4096\r\n",
                                     "* SEARCH 17 900 901\r\n" +
                                     "$tag OK completed\r\n");

        Assert.IsTrue(session.UidSearch(out var matched2, ImapSearchCriteria.Uid(matched) & ImapSearchCriteria.Smaller(4096)).IsSuccess);

        Assert.IsTrue(matched2.IsUidSet);
        Assert.IsFalse(matched2.IsEmpty);

        CollectionAssert.AreEqual(new long[] { 17, 900, 901 }, matched2);

        TestUtils.Assert.IsSerializableBinaryFormat(matched2, delegate(ImapMatchedSequenceSet d) {
          Assert.IsTrue(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("17,900,901", d.Expression);
        });
      });
    }

    [Test]
    public void TestESearch_Searchres_Incapable()
    {
      TestCases.WithMailboxSelectedSession(new[] {"ESEARCH"}, (session, server) => {
        // SEARCH transaction
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        Assert.Throws<ImapIncapableException>(() => session.ESearch(out _, criteria, ImapSearchResultOptions.Save));
      });
    }

    [Test]
    public void TestSort()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT"}, (session, server) => {
        // SORT transaction
        server.EnqueueTaggedResponse("SORT (REVERSE SIZE DATE) utf-8 ALL\r\n",
                                     "* SORT 5 3 4 1 2\r\n" +
                                     "$tag OK SORT completed\r\n");

        Assert.IsTrue(session.Sort(out var matched, ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);

        CollectionAssert.AreEqual(new long[] { 5, 3, 4, 1, 2 }, matched);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("5,3,4,1,2", d.Expression);
        });
      });
    }

    [Test]
    public void TestSort_WithLiteral_DefaultCharset()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT"}, (session, server) => {
        server.EnqueueTaggedResponse("SORT ",
                                     "* SORT 5 3 4 1 2\r\n" +
                                     "$tag OK SORT completed\r\n");

        Assert.IsTrue(session.Sort(out var matched,
                                   ImapSortCriteria.SizeReverse + ImapSortCriteria.Date,
                                   ImapSearchCriteria.Subject("件名")).IsSuccess);

        Assert.AreEqual("0004 SORT (REVERSE SIZE DATE) utf-8 SUBJECT {6}\r\n" +
                        "件名\r\n",
                        server.DequeueRequest(Encoding.UTF8));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
        });
      });
    }

    [Test]
    public void TestSort_WithLiteral_SpecifiedCharset()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT"}, (session, server) => {
        server.EnqueueTaggedResponse("SORT ",
                                     "* SORT 5 3 4 1 2\r\n" +
                                     "$tag OK SORT completed\r\n");

        Assert.IsTrue(session.Sort(out var matched,
                                   ImapSortCriteria.SizeReverse + ImapSortCriteria.Date,
                                   ImapSearchCriteria.Subject("件名"),
                                   TestUtils.Encodings.ShiftJis).IsSuccess);

        Assert.AreEqual("0004 SORT (REVERSE SIZE DATE) shift_jis SUBJECT {4}\r\n" +
                        "\x008c\x008f\x0096\x00bc\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
        });
      });
    }

    [Test]
    public void TestSort_NothingMatched()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT"}, (session, server) => {
        // SORT transaction
        server.EnqueueTaggedResponse("SORT (DATE) utf-8 ALL\r\n",
                                     "* SORT\r\n" +
                                     "$tag OK SORT completed\r\n");

        Assert.IsTrue(session.Sort(out var matched, ImapSortCriteria.Date, ImapSearchCriteria.All).IsSuccess);

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);

        TestUtils.Assert.IsSerializableBinaryFormat(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsTrue(d.IsEmpty);
        });
      });
    }

    [Test]
    public void TestSort_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ImapIncapableException>(() => session.Sort(out _, ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All));
      });
    }

    [Test]
    public void TestESort()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT", "ESORT"}, (session, server) => {
        // SORT transaction
        server.EnqueueTaggedResponse("UID SORT RETURN (ALL) (REVERSE DATE) utf-8 UNDELETED UNKEYWORD $Junk\r\n",
                                     "* ESEARCH (TAG \"0004\") UID ALL 23765,23764,23763,23761\r\n" +
                                     "$tag OK SORT completed\r\n");

        Assert.IsTrue(session.UidESort(out var matched,
                                       ImapSortCriteria.DateReverse,
                                       ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                                       ImapSearchResultOptions.All).IsSuccess);

        Assert.IsTrue(matched.IsUidSet);
        Assert.AreEqual("0004", matched.Tag);

        CollectionAssert.AreEqual(new long[] { 23765, 23764, 23763, 23761 }, matched);
      });
    }

    [Test]
    public void TestESort_Incapable()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT"}, (session, server) => {
        // SORT transaction
        Assert.Throws<ImapIncapableException>(() => {
          session.UidESort(out _,
                           ImapSortCriteria.DateReverse,
                           ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                           ImapSearchResultOptions.All);
        });
      });
    }

    [Test]
    public void TestSort_DisplayIncapable()
    {
      TestCases.WithMailboxSelectedSession(new[] {"SORT"}, (session, server) => {
        var ex = Assert.Throws<ImapIncapableException>(() => session.Sort(out _, ImapSortCriteria.Date + ImapSortCriteria.DisplayFrom, ImapSearchCriteria.All));

        CollectionAssert.Contains(ex.RequiredCapabilities, ImapCapability.SortDisplay);
      });
    }

    [Test]
    public void TestThread()
    {
      TestCases.WithMailboxSelectedSession(new[] {"THREAD=REFERENCES"}, (session, server) => {
        // THREAD transaction
        server.EnqueueTaggedResponse("THREAD REFERENCES utf-8 ALL\r\n",
                                     "* THREAD (166)(167)(168)(169)(172)(170)(171)(173)(174 (175)(176)(178)(181)(180))(179)(177 (183)(182)(188)(184)(185)(186)(187)(189))(190)(191)(192)(193)(194 195)(196 (197)(198))(199)(200 202)(201)(203)(204)(205)(206 207)(208)\r\n" +
                                     "$tag OK THREAD completed\r\n");

        Assert.IsTrue(session.Thread(out var threadList, ImapThreadingAlgorithm.References, ImapSearchCriteria.All).IsSuccess);

        Assert.AreEqual(25, threadList.Children.Count);
      });
    }

    [Test]
    public void TestThread_NothingMatched()
    {
      TestCases.WithMailboxSelectedSession(new[] {"THREAD=REFERENCES"}, (session, server) => {
        // THREAD transaction
        server.EnqueueTaggedResponse("THREAD REFERENCES utf-8 ALL\r\n",
                                     "* THREAD \r\n" +
                                     "$tag OK THREAD completed\r\n");

        Assert.IsTrue(session.Thread(out var threadList, ImapThreadingAlgorithm.References, ImapSearchCriteria.All).IsSuccess);

        Assert.IsNotNull(threadList);
        Assert.AreEqual(0, threadList.Children.Count);
      });
    }

    [Test]
    public void TestThread_WithLiteral_DefaultCharset()
    {
      TestCases.WithMailboxSelectedSession(new[] {"THREAD=REFERENCES"}, (session, server) => {
        server.EnqueueTaggedResponse("THREAD ",
                                     "* THREAD (206 207)(208)\r\n" +
                                     "$tag OK THREAD completed\r\n");

        Assert.IsTrue(session.Thread(out var threadList,
                                     ImapThreadingAlgorithm.References,
                                     ImapSearchCriteria.Subject("件名")).IsSuccess);

        Assert.AreEqual("0004 THREAD REFERENCES utf-8 SUBJECT {6}\r\n" +
                        "件名\r\n",
                        server.DequeueRequest(Encoding.UTF8));

        Assert.IsNotNull(threadList);
        Assert.AreEqual(2, threadList.Children.Count);
      });
    }

    [Test]
    public void TestThread_WithLiteral_SpecifiedCharset()
    {
      TestCases.WithMailboxSelectedSession(new[] {"THREAD=REFERENCES"}, (session, server) => {
        server.EnqueueTaggedResponse("THREAD ",
                                     "* THREAD (206 207)(208)\r\n" +
                                     "$tag OK THREAD completed\r\n");

        Assert.IsTrue(session.Thread(out var threadList,
                                     ImapThreadingAlgorithm.References,
                                     ImapSearchCriteria.Subject("件名"),
                                     TestUtils.Encodings.ShiftJis).IsSuccess);

        Assert.AreEqual("0004 THREAD REFERENCES shift_jis SUBJECT {4}\r\n" +
                        "\x008c\x008f\x0096\x00bc\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(threadList);
      });
    }

    [Test]
    public void TestThread_Incapable()
    {
      TestCases.WithMailboxSelectedSession(new[] {"THREAD=ORDEREDSUBJECT"}, (session, server) => {
        Assert.Throws<ImapIncapableException>(() => session.Thread(out _, ImapThreadingAlgorithm.References, ImapSearchCriteria.All));
      });
    }

    [Test]
    public void TestStore()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // STORE
        server.EnqueueTaggedResponse("STORE 2:4 +FLAGS (\\Deleted)\r\n",
                                     "* 2 FETCH (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* 3 FETCH (FLAGS (\\Deleted))\r\n" +
                                     "* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n" +
                                     "$tag OK STORE completed\r\n");

        Assert.IsTrue(session.Store(ImapSequenceSet.Range(2, 4),
                                    ImapStoreDataItems.AddFlags(ImapMessageFlag.Deleted),
                                    out var messagesAttrs).IsSuccess);

        Assert.AreEqual(3, messagesAttrs.Length);

        Assert.AreEqual(2, messagesAttrs[0].Sequence);
        Assert.AreEqual(0UL, messagesAttrs[0].ModSeq);
        Assert.AreEqual(2, messagesAttrs[0].Flags.Count);
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messagesAttrs[1].Sequence);
        Assert.AreEqual(0UL, messagesAttrs[1].ModSeq);
        Assert.AreEqual(1, messagesAttrs[1].Flags.Count);
        Assert.IsTrue(messagesAttrs[1].Flags.Contains(ImapMessageFlag.Deleted));

        Assert.AreEqual(4, messagesAttrs[2].Sequence);
        Assert.AreEqual(0UL, messagesAttrs[2].ModSeq);
        Assert.AreEqual(3, messagesAttrs[2].Flags.Count);
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Seen));
      });
    }

    [Test]
    public void TestStore_NotRetrieveResult()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // STORE
        server.EnqueueTaggedResponse("STORE 2:4 +FLAGS (\\Deleted)\r\n",
                                     "* 2 FETCH (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* 3 FETCH (FLAGS (\\Deleted))\r\n" +
                                     "* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n" +
                                     "$tag OK STORE completed\r\n");

        Assert.IsTrue(session.Store(ImapSequenceSet.Range(2, 4),
                                    ImapStoreDataItems.AddFlags(ImapMessageFlag.Deleted)).IsSuccess);
      });
    }

    [Test]
    public void TestStore_EmptySequenceSet()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ArgumentException>(delegate {
          session.Store(ImapSequenceSet.Empty(), ImapStoreDataItems.AddFlags(ImapMessageFlag.Deleted));
        });
      });
    }

    [Test]
    public void TestStore_EmptyUidSet()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ArgumentException>(delegate {
          session.Store(ImapUidSet.Empty(), ImapStoreDataItems.AddFlags(ImapMessageFlag.Deleted));
        });
      });
    }

    [Test]
    public void TestStore_UnchangedSince()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // STORE
        server.EnqueueTaggedResponse("STORE 7,5,9 (UNCHANGEDSINCE 320162338) +FLAGS.SILENT (\\Deleted)\r\n",
                                     "* 7 FETCH (MODSEQ (320162342) FLAGS (\\Seen \\Deleted))\r\n" +
                                     "* 5 FETCH (MODSEQ (320162350))\r\n" +
                                     "* 9 FETCH (MODSEQ (320162349) FLAGS (\\Answered))\r\n" +
                                     "$tag OK [MODIFIED 7,9] Conditional STORE failed\r\n");

        Assert.IsTrue(session.StoreUnchangedSince(ImapSequenceSet.Of(7, 5, 9),
                                                  ImapStoreDataItems.AddFlagsSilent(ImapMessageFlag.Deleted),
                                                  320162338UL,
                                                  out var messagesAttrs,
                                                  out var failedMessageSet).IsSuccess);

        Assert.AreEqual(3, messagesAttrs.Length);

        Assert.AreEqual(7, messagesAttrs[0].Sequence);
        Assert.AreEqual(320162342UL, messagesAttrs[0].ModSeq);
        Assert.AreEqual(2, messagesAttrs[0].Flags.Count);
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Deleted));

        Assert.AreEqual(5, messagesAttrs[1].Sequence);
        Assert.AreEqual(320162350UL, messagesAttrs[1].ModSeq);
        Assert.AreEqual(0, messagesAttrs[1].Flags.Count);

        Assert.AreEqual(9, messagesAttrs[2].Sequence);
        Assert.AreEqual(320162349UL, messagesAttrs[2].ModSeq);
        Assert.AreEqual(1, messagesAttrs[2].Flags.Count);
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Answered));

        CollectionAssert.AreEqual(new long[] { 7, 9 }, failedMessageSet);
      });
    }

    [Test]
    public void TestStore_UnchangedSince_NotRetrieveResult()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // STORE
        server.EnqueueTaggedResponse("STORE 7,5,9 (UNCHANGEDSINCE 320162338) +FLAGS.SILENT (\\Deleted)\r\n",
                                     "* 7 FETCH (MODSEQ (320162342) FLAGS (\\Seen \\Deleted))\r\n" +
                                     "* 5 FETCH (MODSEQ (320162350))\r\n" +
                                     "* 9 FETCH (MODSEQ (320162349) FLAGS (\\Answered))\r\n" +
                                     "$tag OK [MODIFIED 7,9] Conditional STORE failed\r\n");

        Assert.IsTrue(session.StoreUnchangedSince(ImapSequenceSet.Of(7, 5, 9),
                                                  ImapStoreDataItems.AddFlagsSilent(ImapMessageFlag.Deleted),
                                                  320162338UL).IsSuccess);
      });
    }


    [Test]
    public void TestStore_UnchangedSince_TaggedNoResponse()
    {
      TestCases.WithMailboxSelectedSession(new[] {"CONDSTORE"}, (session, server) => {
        // STORE
        server.EnqueueTaggedResponse("STORE 1:7 (UNCHANGEDSINCE 320172338) +FLAGS (\\Seen)\r\n",
                                     "* 1 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                                     "* 3 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                                     "$tag NO [MODIFIED 2] Some of the messages no longer exist.\r\n");

        Assert.IsFalse(session.StoreUnchangedSince(ImapSequenceSet.Range(1, 7),
                                                   ImapStoreDataItems.AddFlags(ImapMessageFlag.Seen),
                                                   320172338UL,
                                                   out var messagesAttrs,
                                                   out var failedMessageSet).IsSuccess);

        Assert.IsNotNull(messagesAttrs);
        Assert.AreEqual(2, messagesAttrs.Length);

        Assert.AreEqual(1, messagesAttrs[0].Sequence);
        Assert.AreEqual(320172342UL, messagesAttrs[0].ModSeq);
        Assert.AreEqual(1, messagesAttrs[0].Flags.Count);
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messagesAttrs[1].Sequence);
        Assert.AreEqual(320172342UL, messagesAttrs[1].ModSeq);
        Assert.AreEqual(1, messagesAttrs[1].Flags.Count);
        Assert.IsTrue(messagesAttrs[1].Flags.Contains(ImapMessageFlag.Seen));

        CollectionAssert.AreEqual(new long[] { 2 }, failedMessageSet);
      });
    }

    [Test]
    public void TestStore_UnchangedSince_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // STORE
        Assert.Throws<ImapIncapableException>(delegate {
          session.StoreUnchangedSince(ImapSequenceSet.Range(1, 7),
                                      ImapStoreDataItems.AddFlags(ImapMessageFlag.Seen),
                                      320172338UL,
                                      out _,
                                      out _);
        });
      });
    }

    [Test]
    public void TestExpunge()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // EXPUNGE transaction
        server.EnqueueTaggedResponse("EXPUNGE\r\n",
                                     "* 3 EXPUNGE\r\n" +
                                     "* 4 EXPUNGE\r\n" +
                                     "* 7 EXPUNGE\r\n" +
                                     "* 11 EXPUNGE\r\n" +
                                     "$tag OK EXPUNGE completed\r\n");

        Assert.IsTrue(session.Expunge(out var expunged).IsSuccess);

        Assert.AreEqual(4, expunged.Count);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(4,  expunged[1]);
        Assert.AreEqual(7,  expunged[2]);
        Assert.AreEqual(11, expunged[3]);
      });
    }

    [Test]
    public void TestExpunge_NotRetrieveResult()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // EXPUNGE transaction
        server.EnqueueTaggedResponse("EXPUNGE\r\n",
                                     "* 3 EXPUNGE\r\n" +
                                     "* 4 EXPUNGE\r\n" +
                                     "* 7 EXPUNGE\r\n" +
                                     "* 11 EXPUNGE\r\n" +
                                     "$tag OK EXPUNGE completed\r\n");

        Assert.IsTrue(session.Expunge().IsSuccess);
      });
    }

    [Test]
    public void TestUidExpunge()
    {
      TestCases.WithMailboxSelectedSession(new[] {"UIDPLUS"}, (session, server) => {
        // EXPUNGE transaction
        server.EnqueueTaggedResponse("UID EXPUNGE 1:*\r\n",
                                     "* 3 EXPUNGE\r\n" +
                                     "* 3 EXPUNGE\r\n" +
                                     "* 3 EXPUNGE\r\n" +
                                     "$tag OK EXPUNGE completed\r\n");

        Assert.IsTrue(session.UidExpunge(ImapUidSet.From(1), out var expunged).IsSuccess);

        Assert.AreEqual(3, expunged.Count);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(3,  expunged[1]);
        Assert.AreEqual(3,  expunged[2]);
      });
    }

    [Test]
    public void TestUidExpunge_NotRetrieveResult()
    {
      TestCases.WithMailboxSelectedSession(new[] {"UIDPLUS"}, (session, server) => {
        // EXPUNGE transaction
        server.EnqueueTaggedResponse("UID EXPUNGE 1:*\r\n",
                                     "* 3 EXPUNGE\r\n" +
                                     "* 3 EXPUNGE\r\n" +
                                     "* 3 EXPUNGE\r\n" +
                                     "$tag OK EXPUNGE completed\r\n");

        Assert.IsTrue(session.UidExpunge(ImapUidSet.From(1)).IsSuccess);
      });
    }

    [Test]
    public void TestUidExpunge_SequenceSet()
    {
      TestCases.WithMailboxSelectedSession(new[] {"UIDPLUS"}, (session, server) => {
        Assert.Throws<ArgumentException>(() => session.UidExpunge(ImapSequenceSet.From(1)));
      });
    }

    [Test]
    public void TestUidExpunge_EmptyUidSet()
    {
      TestCases.WithMailboxSelectedSession(new[] {"UIDPLUS"}, (session, server) => {
        Assert.Throws<ArgumentException>(() => session.UidExpunge(ImapUidSet.Empty()));
      });
    }

    [Test]
    public void TestUidExpunge_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.UidPlus));

        Assert.Throws<ImapIncapableException>(() => session.UidExpunge(ImapUidSet.From(1)));
      });
    }

    [Test]
    public void TestCopy()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // COPY transaction
        server.EnqueueTaggedResponse("COPY 1:10 INBOX\r\n",
                                     "$tag OK COPY completed\r\n");

        Assert.IsTrue(session.Copy(ImapSequenceSet.Range(1, 10), "INBOX").IsSuccess);
      });
    }

    [Test]
    public void TestCopy_MailboxNameEmpty()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var set = ImapSequenceSet.Range(1, 10);

        var ex1 = Assert.Throws<ArgumentException>(() => session.Copy(set, string.Empty));
        Assert.AreEqual("mailboxName", ex1.ParamName);

        var ex2 = Assert.Throws<ArgumentException>(() => session.Copy(set, string.Empty, out ImapCopiedUidSet _));
        Assert.AreEqual("mailboxName", ex2.ParamName);
      });
    }

    [Test]
    public void TestCopy_EmptySequenceSet()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ArgumentException>(() => session.Copy(ImapSequenceSet.Empty(), "INBOX"));
      });
    }

    [Test]
    public void TestCopy_EmptyUidSet()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.Throws<ArgumentException>(() => session.Copy(ImapUidSet.Empty(), "INBOX"));
      });
    }

    [Test]
    public void TestCopyTryCreateMailbox()
    {
      TestCases.WithMailboxSelectedSession((session, server, expectedAuthority) => {
        // COPY transaction
        server.EnqueueTaggedResponse("UID COPY 1:10 INBOX.copyto\r\n",
                                     "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
        server.EnqueueTaggedResponse("CREATE INBOX.copyto\r\n",
                                     "$tag OK CREATE completed\r\n");
        server.EnqueueTaggedResponse("UID COPY 1:10 INBOX.copyto\r\n",
                                     "$tag OK COPY completed\r\n");

        Assert.IsTrue(session.CopyTryCreateMailbox(ImapUidSet.Range(1, 10), "INBOX.copyto", out var created).IsSuccess);

        Assert.AreEqual("INBOX.copyto", created.Name);
        Assert.AreEqual(expectedAuthority + "INBOX.copyto",
                        created.Url.AbsoluteUri);
        Assert.IsNotNull(created.Flags);
        Assert.IsNotNull(created.ApplicableFlags);
        Assert.IsNotNull(created.PermanentFlags);
      });
    }

    [Test]
    public void TestCopyTryCreateMailbox_MailboxNameEmpty()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var set = ImapSequenceSet.Range(1, 10);

        var ex1 = Assert.Throws<ArgumentException>(() => session.CopyTryCreateMailbox(set, string.Empty, out ImapMailboxStatus _));
        Assert.AreEqual("mailboxName", ex1.ParamName);

        var ex2 = Assert.Throws<ArgumentException>(() => session.CopyTryCreateMailbox(set, string.Empty, out _, out _));
        Assert.AreEqual("mailboxName", ex2.ParamName);
      });
    }

    [Test]
    public void TestCopyTryCreateMailbox_NoReferralResponseCode()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // COPY transaction
        server.EnqueueTaggedResponse("UID COPY 1:10 SHARED/STUFF\r\n",
                                     "$tag NO [REFERRAL IMAP://user;AUTH=*@SERVER2/SHARED/STUFF] Unable to copy message(s) to SERVER2.\r\n");

        Assert.IsFalse(session.CopyTryCreateMailbox(ImapUidSet.Range(1, 10), "SHARED/STUFF", out var created).IsSuccess);

        Assert.IsNull(created);
      });
    }

    [Test]
    public void TestCopy_DontTryCreate()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // COPY transaction
        server.EnqueueTaggedResponse("COPY 1:10 INBOX.copyto\r\n",
                                     "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");

        Assert.IsFalse(session.Copy(ImapSequenceSet.Range(1, 10), "INBOX.copyto").IsSuccess);
      });
    }

    [Test]
    public void TestCopy_WithAppendUidResponseCode()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        // COPY transaction
        server.EnqueueTaggedResponse("COPY 2:4 meeting\r\n",
                                     "$tag OK [COPYUID 38505 304,319:320 3956:3958] Done\r\n");

        ImapCopiedUidSet copied;

        Assert.IsTrue(session.Copy(ImapSequenceSet.Range(2, 4), "meeting", out copied).IsSuccess);

        Assert.IsNotNull(copied);
        Assert.AreEqual(38505, copied.UidValidity);

        CollectionAssert.AreEqual(new long[] { 304, 319, 320 }, copied.CopiedUidSet);
        CollectionAssert.AreEqual(new long[] { 3956, 3957, 3958 }, copied.AssignedUidSet);
        CollectionAssert.AreEqual(copied, copied.AssignedUidSet);

        TestUtils.Assert.IsSerializableBinaryFormat(copied, delegate(ImapCopiedUidSet deserialized) {
          Assert.AreEqual(38505L, deserialized.UidValidity);
          Assert.AreEqual("304,319:320", deserialized.CopiedUidSet.Expression);
          Assert.AreEqual("3956:3958", deserialized.AssignedUidSet.Expression);
        });
      });
    }

    [Test]
    public void TestMove()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MOVE"}, (session, server) => {
        // MOVE transaction
        server.EnqueueTaggedResponse("MOVE 1:10 INBOX\r\n",
                                     "$tag OK MOVE completed\r\n");

        Assert.IsTrue(session.Move(ImapSequenceSet.Range(1, 10), "INBOX").IsSuccess);
      });
    }

    [Test]
    public void TestMove_MailboxNameEmpty()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var set = ImapSequenceSet.Range(1, 10);

        var ex1 = Assert.Throws<ArgumentException>(() => session.Move(set, string.Empty));
        Assert.AreEqual("mailboxName", ex1.ParamName);

        var ex2 = Assert.Throws<ArgumentException>(() => session.Move(set, string.Empty, out ImapCopiedUidSet _));
        Assert.AreEqual("mailboxName", ex2.ParamName);
      });
    }

    [Test]
    public void TestMove_EmptySequenceSet()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MOVE"}, (session, server) => {
        Assert.Throws<ArgumentException>(() => session.Move(ImapSequenceSet.Empty(), "INBOX"));
      });
    }

    [Test]
    public void TestMove_EmptyUidSet()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MOVE"}, (session, server) => {
        Assert.Throws<ArgumentException>(() => session.Move(ImapUidSet.Empty(), "INBOX"));
      });
    }

    [Test]
    public void TestMoveTryCreateMailbox()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MOVE"}, (session, server, expectedAuthority) => {
        // MOVE transaction
        server.EnqueueTaggedResponse("UID MOVE 1:10 INBOX.copyto\r\n",
                                     "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
        server.EnqueueTaggedResponse("CREATE INBOX.copyto\r\n",
                                     "$tag OK CREATE completed\r\n");
        server.EnqueueTaggedResponse("UID MOVE 1:10 INBOX.copyto\r\n",
                                     "$tag OK COPY completed\r\n");

        Assert.IsTrue(session.MoveTryCreateMailbox(ImapUidSet.Range(1, 10), "INBOX.copyto", out var created).IsSuccess);

        Assert.AreEqual("INBOX.copyto", created.Name);
        Assert.AreEqual(expectedAuthority + "INBOX.copyto",
                        created.Url.AbsoluteUri);
        Assert.IsNotNull(created.Flags);
        Assert.IsNotNull(created.ApplicableFlags);
        Assert.IsNotNull(created.PermanentFlags);
      });
    }

    [Test]
    public void TestMoveTryCreateMailbox_MailboxNameEmpty()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        var set = ImapSequenceSet.Range(1, 10);

        var ex1 = Assert.Throws<ArgumentException>(() => session.MoveTryCreateMailbox(set, string.Empty, out ImapMailboxStatus _));
        Assert.AreEqual("mailboxName", ex1.ParamName);

        var ex2 = Assert.Throws<ArgumentException>(() => session.MoveTryCreateMailbox(set, string.Empty, out _, out _));
        Assert.AreEqual("mailboxName", ex2.ParamName);
      });
    }

    [Test]
    public void TestMove_DontTryCreate()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MOVE"}, (session, server) => {
        // MOVE transaction
        server.EnqueueTaggedResponse("MOVE 1:10 INBOX.copyto\r\n",
                                     "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");

        Assert.IsFalse(session.Move(ImapSequenceSet.Range(1, 10), "INBOX.copyto").IsSuccess);
      });
    }

    [Test]
    public void TestMove_WithCopyUidResponseCode()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MOVE"}, (session, server) => {
        // MOVE transaction
        server.EnqueueTaggedResponse("MOVE 42:69 foo\r\n",
                                     "$tag OK [COPYUID 432432 42:69 1202:1229] Done\r\n");

        ImapCopiedUidSet copied;

        Assert.IsTrue(session.Move(ImapSequenceSet.Range(42, 69), "foo", out copied).IsSuccess);

        Assert.IsNotNull(copied);
        Assert.AreEqual(432432L, copied.UidValidity);
        Assert.AreEqual("42:69", copied.CopiedUidSet.Expression);
        Assert.AreEqual("1202:1229", copied.AssignedUidSet.Expression);

        TestUtils.Assert.IsSerializableBinaryFormat(copied, delegate(ImapCopiedUidSet deserialized) {
          Assert.AreEqual(432432L, deserialized.UidValidity);
          Assert.AreEqual("42:69", deserialized.CopiedUidSet.Expression);
          Assert.AreEqual("1202:1229", deserialized.AssignedUidSet.Expression);
        });
      });
    }

    [Test]
    public void TestMove_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Move));

        Assert.Throws<ImapIncapableException>(() => session.Move(ImapUidSet.From(1), "foo"));
      });
    }
  }
}
