using System;
using System.IO;
using System.Text;

using NUnit.Framework;

using Smdn.IO.Streams.LineOriented;
using Smdn.Net.Imap4.Protocol.DataStructures;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  [TestFixture]
  public class ImapResponseReceiverTests {
    private class PseudoResponseReceiver : IDisposable {
      public bool ReceiveResponseAsUTF8 {
        get => receiver.ReceiveResponseAsUTF8;
        set => receiver.ReceiveResponseAsUTF8 = value;
      }

      public PseudoResponseReceiver(string responseLines)
      {
        baseStream = new MemoryStream(Encoding.UTF8.GetBytes(responseLines), false);
        baseStream.Seek(0L, SeekOrigin.Begin);

        stream = new LineOrientedBufferedStream(baseStream);

        receiver = new ImapResponseReceiver(stream);
      }

      public ImapResponse ReceiveResponse(int maxTryReceive = 1)
      {
        ImapResponse resp = null;

        for (var act = 0; ; act++) {
          if (maxTryReceive < act)
            Assert.Fail("ReadResponse failed");

          resp = receiver.ReceiveResponse();

          if (resp != null)
            return resp;

          if (!receiver.IsResponseContinuing)
            break;
        }

        return resp;
      }

      public void Dispose()
      {
        baseStream.Dispose();
        stream.Dispose();
      }

      private ImapResponseReceiver receiver;
      private MemoryStream baseStream;
      private LineOrientedBufferedStream stream;
    }

    private ImapResponse ReceiveResponse(string responseLines, int maxTryReceive = 1)
    {
      using (var receiver = new PseudoResponseReceiver(responseLines)) {
        return receiver.ReceiveResponse(maxTryReceive);
      }
    }

#if false
    [Test, Ignore("can't test")]
    public void TestDividedStatusResponse()
    {
      ImapResponse response;

      WriteResponse("* OK IMAP4");
      WriteResponse("rev1 Service Ready");
      WriteResponse("\r\n");

      response = ReadResponse();

      Assert.IsNotNull(response);
      Assert.IsTrue(response is ImapUntaggedStatusResponse, "response type");

      var status = (response as ImapUntaggedStatusResponse);

      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("IMAP4rev1 Service Ready", status.ResponseText.Text);
    }

    [Test, Ignore("can't test")]
    public void TestDividedDataResponse()
    {
      ImapResponse response;

      WriteResponse("* SEARCH 2 4 5");
      WriteResponse(" 6 9 10 13");
      WriteResponse(" 15\r\n");

      response = ReadResponse();

      Assert.IsNotNull(response);
      Assert.IsTrue(response is ImapServerDataResponse, "response type");

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Search ==  dataResponse.Type);

      Assert.AreEqual(8, dataResponse.Data.Length);
      //Assert.AreEqual("* SEARCH 2 4 5 6 9 10 13 15\r\n", dataResponse.RawResponse);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.Data[0].Format);
      Assert.AreEqual("2", dataResponse.Data[0].GetTextAsString());
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.Data[7].Format);
      Assert.AreEqual("15", dataResponse.Data[7].GetTextAsString());
    }

    [Test, Ignore("can't test")]
    public void TestDividedDataResponseAtLastCrLf()
    {
      ImapResponse response;

      WriteResponse("* 12 FETCH (BODY[HEADER] {342}\r\n" +
                                "Date: Wed, 17 Jul 1996 02:23:25 -0700 (PDT)\r\n" + 
                                "From: Terry Gray <gray@cac.washington.edu>\r\n" + 
                                "Subject: IMAP4rev1 WG mtg summary and minutes\r\n" + 
                                "To: imap@cac.washington.edu\r\n" + 
                                "cc: minutes@CNRI.Reston.VA.US, John Klensin <KLENSIN@MIT.EDU>\r\n" + 
                                "Message-Id: <B27397-0100000@cac.washington.edu>\r\n" + 
                                "MIME-Version: 1.0\r\n" + 
                                 "Content-Type: TEXT/PLAIN; CHARSET=US-ASCII\r\n" + 
                                "\r\n" +
                                ")");
      WriteResponse("\r\n" +
                                "* 13 FETCH (BODY[HEADER] {342}\r\n" +
                                "Date: Wed, 17 Jul 1996 02:23:25 -0700 (PDT)\r\n" + 
                                "From: Terry Gray <gray@cac.washington.edu>\r\n" + 
                                "Subject: IMAP4rev1 WG mtg summary and minutes\r\n" + 
                                "To: imap@cac.washington.edu\r\n" + 
                                "cc: minutes@CNRI.Reston.VA.US, John Klensin <KLENSIN@MIT.EDU>\r\n" + 
                                "Message-Id: <B27397-0100000@cac.washington.edu>\r\n" + 
                                 "MIME-Version: 1.0\r\n" + 
                                "Content-Type: TEXT/PLAIN; CHARSET=US-ASCII\r\n" + 
                                "\r\n" +
                                ")\r\n");

      response = ReadResponse();

      Assert.IsTrue(response is ImapServerDataResponse, "response0 type");

      ImapServerDataResponse data = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Fetch == data.Type);

      Assert.AreEqual(2, data.Data.Length);
      Assert.AreEqual(ImapDataFormat.Text, data.Data[0].Format);
      Assert.AreEqual("12", data.Data[0].GetTextAsString());
      Assert.AreEqual(ImapDataFormat.List, data.Data[1].Format);

      response = ReadResponse();

      Assert.IsTrue(response is ImapServerDataResponse, "response1 type");

      data = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Fetch == data.Type);

      Assert.AreEqual(2, data.Data.Length);
      Assert.AreEqual(ImapDataFormat.Text, data.Data[0].Format);
      Assert.AreEqual("13", data.Data[0].GetTextAsString());
      Assert.AreEqual(ImapDataFormat.List, data.Data[1].Format);
    }

    [Test, Ignore("can't test")]
    public void TestDividedQuotedDataResponse()
    {
      ImapResponse response;

      WriteResponse("* LIST \".");
      WriteResponse("\" \"INB");
      WriteResponse("OX\"\r\n");

      response = ReadResponse();

      Assert.IsNotNull(response);
      Assert.IsTrue(response is ImapServerDataResponse, "response type");

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.List ==  dataResponse.Type);

      Assert.AreEqual(ImapDataFormat.Text, dataResponse.Data[0].Format);
      Assert.AreEqual(".", dataResponse.Data[0].GetTextAsString());

      Assert.AreEqual(ImapDataFormat.Text, dataResponse.Data[1].Format);
      Assert.AreEqual("INBOX", dataResponse.Data[1].GetTextAsString());
    }

    [Test, Ignore("can't test")]
    public void TestDividedParenthesizedDataResponse()
    {
      ImapResponse response;

      WriteResponse("* NAMESPACE");
      WriteResponse(" (");
      WriteResponse("(\"INBOX.\" ");
      WriteResponse("\".\")) NIL NIL\r\n");

      response = ReadResponse();

      Assert.IsNotNull(response);
      Assert.IsTrue(response is ImapServerDataResponse, "response type");

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Namespace ==  dataResponse.Type);

      var data = (response as ImapServerDataResponse).Data;

      Assert.AreEqual(ImapDataFormat.List, data[0].Format);
      Assert.AreEqual(ImapDataFormat.Nil, data[1].Format);
      Assert.AreEqual(ImapDataFormat.Nil, data[2].Format);
    }

    [Test, Ignore("can't test")]
    public void TestDividedLiteralDataResponse()
    {
      ImapResponse response;

      WriteResponse("* 12 FETCH (BODY[HEADER] {342}\r\n" +
                                 "Date: Wed, 17 Jul 1996 02:23:25 -0700 (PDT)\r\n" + 
                                 "From: Terry Gray <gray@cac.washington.edu>\r\n" + 
                                 "Subject: IMAP4rev1 WG");
      WriteResponse(" mtg summary and minutes\r\n" + 
                                 "To: imap@cac.washington.edu\r\n" + 
                                 "cc: minutes@CNRI.Reston.VA.US, John Klensin <KLENSIN@MIT.EDU>\r\n" + 
                                 "Message-Id: <B27397-0100000@cac.washington.edu>\r\n" + 
                                 "MIME-Version: 1.0\r");
      WriteResponse("\nContent-Type: TEXT/PLAIN; CHARSET=US-ASCII\r\n" + 
                                 "\r\n");
      WriteResponse(")\r\n");

      response = ReadResponse();

      Assert.IsNotNull(response);
      Assert.IsTrue(response is ImapServerDataResponse, "response type");

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Fetch ==  dataResponse.Type);

      Assert.AreEqual(2, dataResponse.Data.Length);
      Assert.AreEqual(2, dataResponse.Data[1].List.Length);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.Data[1].List[0].Format);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.Data[1].List[1].Format);
      Assert.AreEqual(342, dataResponse.Data[1].List[1].GetTextLength());
    }
#endif

    [TestCase(" OK\r\n")]
    [TestCase(" NO done\r\n")]
    [TestCase(" NO [INVALID] done\r\n")]
    public void TestReceiveResponse_ResponseDataResponseDone_StartsWithoutTagOrAsterisk(string resp)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      Assert.IsNull(ex.InnerException);
    }

    [TestCase("tag\r\n")]
    [TestCase("*\r\n")]
    public void TestReceiveResponse_ResponseDataResponseDone_WithoutSP(string resp)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      Assert.IsNull(ex.InnerException);
    }

    [TestCase("tag OK\r\n")]
    [TestCase("tag NO\r\n")]
    [TestCase("tag BAD\r\n")]
    public void TestReceiveResponse_ResponseTagged_WithoutSP(string resp)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      Assert.IsNull(ex.InnerException);
    }

    [TestCase("* OK [CAPABILITY\r\n")]
    [TestCase("tag OK [CAPABILITY\r\n")]
    public void TestReceiveResponse_Malformed_RespText_UnclosedBracket(string resp)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      Assert.IsNull(ex.InnerException);
    }

    [TestCase("tag GOOD done\r\n")]
    [TestCase("tag FATAL failed\r\n")]
    public void TestReceiveResponse_Malformed_UndefinedCondition(string resp)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      Assert.IsNull(ex.InnerException);
    }

    [TestCase("tag OK [ALERT\r\n")]
    [TestCase("tag OK [ALERT \r\n")]
    [TestCase("* OK [ALERT\r\n")]
    [TestCase("* OK [ALERT \r\n")]
    public void TestReceiveResponse_Malformed_RespTextCode_UnclosedBracket(string resp)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      Assert.IsNull(ex.InnerException);
    }

    [TestCase("tag OK [CAPABILITY (]\r\n", false)]
    [TestCase("tag OK [CAPABILITY )]\r\n", true)]
    [TestCase("tag OK [CAPABILITY {]\r\n", true)]
    [TestCase("tag OK [CAPABILITY ~{]\r\n", true)]
    public void TestReceiveResponse_Malformed_RespTextCode(string resp, bool throwsMalformedDataException)
    {
      var ex = Assert.Throws<ImapMalformedResponseException>(() => ReceiveResponse(resp));

      if (throwsMalformedDataException)
        Assert.IsInstanceOf<ImapMalformedDataException>(ex.InnerException);
      else
        Assert.IsNull(ex.InnerException);
    }

    [TestCase("* GOOD done\r\n", "GOOD")]
    [TestCase("* FATAL failed\r\n", "FATAL")]
    public void TestReceiveResponse_Ambiguous_UntaggedAndUndefinedCondition(string resp, string expectedDataType)
    {
      var response = ReceiveResponse(resp);

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapServerDataResponse>(response);

      var data = response as ImapServerDataResponse;

      Assert.AreEqual(ImapServerDataResponseType.InvalidOrUnknown, data.Type);
      Assert.GreaterOrEqual(data.DataList.Count, 1);
      Assert.AreEqual(ImapDataFormat.Text, data.DataList[0].Format);
      Assert.AreEqual(expectedDataType, data.DataList[0].GetTextAsString());
    }

    [TestCase("* 5 EXISTS\r\n", "EXISTS")]
    [TestCase("* 5 RECENT\r\n", "RECENT")]
    [TestCase("* 5 EXPUNGE\r\n", "EXPUNGE")]
    public void TestReceiveResponse_ServerDataResponse_SizeStatus(string resp, string expectedType)
    {
      var response = ReceiveResponse(resp);

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapServerDataResponse>(response);

      var dataResponse = response as ImapServerDataResponse;

      Assert.AreEqual(expectedType, dataResponse.Type);

      Assert.IsNotNull(dataResponse.SizeStatusNumber);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.SizeStatusNumber.Format);
      Assert.AreEqual("5", dataResponse.SizeStatusNumber.GetTextAsString());

      Assert.AreEqual(0, dataResponse.DataList.Count);
    }

    [Test]
    public void TestReceiveResponse_ServerDataResponse_SizeStatus_Fetch()
    {
      var response = ReceiveResponse("* 5 FETCH (UID 5)\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapServerDataResponse>(response);

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Fetch == dataResponse.Type);

      Assert.IsNotNull(dataResponse.SizeStatusNumber);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.SizeStatusNumber.Format);
      Assert.AreEqual("5", dataResponse.SizeStatusNumber.GetTextAsString());

      Assert.AreEqual(1, dataResponse.DataList.Count);
      Assert.AreEqual(ImapDataFormat.List, dataResponse.DataList[0].Format);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.DataList[0].List[0].Format);
      Assert.AreEqual("UID", dataResponse.DataList[0].List[0].GetTextAsString());
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.DataList[0].List[1].Format);
      Assert.AreEqual("5", dataResponse.DataList[0].List[1].GetTextAsString());
    }

    [Test]
    public void TestReceiveResponse_ServerDataResponse_LiteralString()
    {
      var response = ReceiveResponse("* 12 FETCH (BODY[HEADER] {342}\r\n" +
                                     "Date: Wed, 17 Jul 1996 02:23:25 -0700 (PDT)\r\n" + 
                                     "From: Terry Gray <gray@cac.washington.edu>\r\n" + 
                                     "Subject: IMAP4rev1 WG mtg summary and minutes\r\n" + 
                                     "To: imap@cac.washington.edu\r\n" + 
                                     "cc: minutes@CNRI.Reston.VA.US, John Klensin <KLENSIN@MIT.EDU>\r\n" + 
                                     "Message-Id: <B27397-0100000@cac.washington.edu>\r\n" + 
                                     "MIME-Version: 1.0\r\n" + 
                                     "Content-Type: TEXT/PLAIN; CHARSET=US-ASCII\r\n" + 
                                     "\r\n" +
                                     ")\r\n",
                                     maxTryReceive: 11);

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapServerDataResponse>(response);

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Fetch == dataResponse.Type);

      Assert.IsNotNull(dataResponse.SizeStatusNumber);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.SizeStatusNumber.Format);
      Assert.AreEqual("12", dataResponse.SizeStatusNumber.GetTextAsString());

      Assert.AreEqual(1, dataResponse.DataList.Count);
      Assert.AreEqual(ImapDataFormat.List, dataResponse.DataList[0].Format);

      Assert.AreEqual(2, dataResponse.DataList[0].List.Count);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.DataList[0].List[0].Format);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.DataList[0].List[1].Format);
      Assert.AreEqual(342, dataResponse.DataList[0].List[1].GetTextLength());
    }

    [Test]
    public void TestReceiveResponse_ServerDataResponse_Literal8String()
    {
      var response = ReceiveResponse("* 12 FETCH (BINARY[1.1] ~{32}\r\n" +
                                     "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f" +
                                     "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f" +
                                     ")\r\n",
                                     maxTryReceive: 1);

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapServerDataResponse>(response);

      var dataResponse = response as ImapServerDataResponse;

      Assert.IsTrue(ImapServerDataResponseType.Fetch == dataResponse.Type);

      Assert.IsNotNull(dataResponse.SizeStatusNumber);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.SizeStatusNumber.Format);
      Assert.AreEqual("12", dataResponse.SizeStatusNumber.GetTextAsString());

      Assert.AreEqual(1, dataResponse.DataList.Count);
      Assert.AreEqual(ImapDataFormat.List, dataResponse.DataList[0].Format);

      Assert.AreEqual(2, dataResponse.DataList[0].List.Count);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.DataList[0].List[0].Format);
      Assert.AreEqual(ImapDataFormat.Text, dataResponse.DataList[0].List[1].Format);
      Assert.AreEqual(32, dataResponse.DataList[0].List[1].GetTextLength());
    }

    [Test]
    public void TestReceiveResponse_ServerDataResponse_MailboxData_ContainsNoSp()
    {
      using (var receiver = new PseudoResponseReceiver("* SEARCH\r\n0006 OK Search completed.\r\n")) {
        var response = receiver.ReceiveResponse();

        Assert.IsInstanceOf<ImapServerDataResponse>(response, "responses[0] type");

        var dataResponse = response as ImapServerDataResponse;

        Assert.IsTrue(ImapServerDataResponseType.Search == dataResponse.Type);

        Assert.IsNull(dataResponse.SizeStatusNumber);

        var data = dataResponse.DataList;

        Assert.AreEqual(0, data.Count);


        response = receiver.ReceiveResponse();

        Assert.IsInstanceOf<ImapTaggedStatusResponse>(response, "responses[1] type");

        var status = response as ImapTaggedStatusResponse;

        Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
        Assert.AreEqual("Search completed.", status.ResponseText.Text);
      }
    }

    [Test]
    public void TestReceiveResponse_ServerDataResponse_ContainsSpecialCharacters()
    {
      // http://tools.ietf.org/html/rfc2683
      // RFC 2683 - IMAP4 Implementation Recommendations
      // 3.4.2. Special Characters

      var respLines = "* LIST () \"\" INBOX\r\n" + 
                      "* LIST () \"\\\\\" TEST\r\n" + 
                      "* LIST () \"\\\\\" {12}\r\n" + 
                      "\"My\" mailbox\r\n" + 
                      "* LIST () \"\\\\\" {17}\r\n" + 
                      "\"My\" mailbox\\Junk\r\n";

      using (var receiver = new PseudoResponseReceiver(respLines)) {
        ImapResponse response;
        ImapListedMailbox list;

        // 1
        response = receiver.ReceiveResponse();

        Assert.IsInstanceOf<ImapServerDataResponse>(response, "response type #1");

        list = ImapServerDataResponseParser.FromList(response as ImapServerDataResponse);

        Assert.AreEqual(string.Empty, list.HierarchyDelimiter, "LIST response #1 delimiter");
        Assert.AreEqual("INBOX", list.Name, "LIST response #1 name");

        // 2
        response = receiver.ReceiveResponse();

        Assert.IsInstanceOf<ImapServerDataResponse>(response, "response type #2");

        list = ImapServerDataResponseParser.FromList(response as ImapServerDataResponse);

        Assert.AreEqual("\\", list.HierarchyDelimiter, "LIST response #2 delimiter");
        Assert.AreEqual("TEST", list.Name, "LIST response #2 name");

        // 3
        response = receiver.ReceiveResponse();

        Assert.IsInstanceOf<ImapServerDataResponse>(response, "response type #3");

        list = ImapServerDataResponseParser.FromList(response as ImapServerDataResponse);

        Assert.AreEqual("\\", list.HierarchyDelimiter, "LIST response #3 delimiter");
        Assert.AreEqual(@"""My"" mailbox", list.Name, "LIST response #3 name");

        // 4
        response = receiver.ReceiveResponse();

        Assert.IsInstanceOf<ImapServerDataResponse>(response, "response type #4");

        list = ImapServerDataResponseParser.FromList(response as ImapServerDataResponse);

        Assert.AreEqual("\\", list.HierarchyDelimiter, "LIST response #4 delimiter");
        Assert.AreEqual(@"""My"" mailbox\Junk", list.Name, "LIST response #4 name");
      }
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_Ok()
    {
      var response = ReceiveResponse("* OK IMAP4rev1 server ready\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("IMAP4rev1 server ready", status.ResponseText.Text);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_No()
    {
      var response = ReceiveResponse("A223 NO COPY failed: disk is full\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapTaggedStatusResponse>(response);

      var status = response as ImapTaggedStatusResponse;

      Assert.IsTrue(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.No, status.Condition);
      Assert.AreEqual("COPY failed: disk is full", status.ResponseText.Text);
      Assert.AreEqual("A223", status.Tag);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_Bad()
    {
      var response = ReceiveResponse("* BAD Disk crash, attempting salvage to a new disk!\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Bad, status.Condition);
      Assert.AreEqual("Disk crash, attempting salvage to a new disk!", status.ResponseText.Text);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_PreAuth()
    {
      var response = ReceiveResponse("* PREAUTH IMAP4rev1 server logged in as Smith\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.PreAuth, status.Condition);
      Assert.AreEqual("IMAP4rev1 server logged in as Smith", status.ResponseText.Text);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_Bye()
    {
      var response = ReceiveResponse("* BYE Autologout; idle for too long\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Bye, status.Condition);
      Assert.AreEqual("Autologout; idle for too long", status.ResponseText.Text);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithText()
    {
      var response = ReceiveResponse("* OK text\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("text", status.ResponseText.Text);

      // response status code
      Assert.IsNull(status.ResponseText.Code);
      Assert.AreEqual(0, status.ResponseText.Arguments.Count);
    }

    [TestCase("* OK [ALERT] text\r\n", "text")]
    [TestCase("* OK [ALERT] text \r\n", "text ")]
    [TestCase("* OK [ALERT]text\r\n", "text")] // lack of SP after ']' is allowed by current implementation
    public void TestReceiveResponse_StatusResponse_WithRespText(string resp, string expectedText)
    {
      var response = ReceiveResponse(resp);

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual(expectedText, status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.Alert);
      Assert.AreEqual(0, status.ResponseText.Arguments.Count);
    }

    [TestCase("* OK [ALERT]\r\n")] // lack of SP after ']' is allowed by current implementation
    [TestCase("* OK [ALERT] \r\n")]
    public void TestReceiveResponse_StatusResponse_WithRespText_EmptyText(string resp)
    {
      var response = ReceiveResponse(resp);

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.IsNotNull(status.ResponseText.Text);
      Assert.IsEmpty(status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.Alert);
      Assert.AreEqual(0, status.ResponseText.Arguments.Count);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_RespCodeOnly()
    {
      var response = ReceiveResponse("* OK [ALERT]\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.IsEmpty(status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.Alert);
      Assert.AreEqual(0, status.ResponseText.Arguments.Count);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_HasArguments_String()
    {
      var response = ReceiveResponse("* OK [UNSEEN 17] Message 17 is the first unseen message\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("Message 17 is the first unseen message", status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.Unseen);
      Assert.AreEqual(1, status.ResponseText.Arguments.Count);
      Assert.AreEqual(ImapDataFormat.Text, status.ResponseText.Arguments[0].Format);
      Assert.AreEqual("17", status.ResponseText.Arguments[0].GetTextAsString());
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_HasArguments_Parenthesized()
    {
      var response = ReceiveResponse("* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.PermanentFlags);
      Assert.AreEqual(1, status.ResponseText.Arguments.Count);
      Assert.AreEqual(ImapDataFormat.List, status.ResponseText.Arguments[0].Format);
      Assert.AreEqual(3, status.ResponseText.Arguments[0].List.Count);
      Assert.AreEqual("\\Deleted",  status.ResponseText.Arguments[0].List[0].GetTextAsString());
      Assert.AreEqual("\\Seen",     status.ResponseText.Arguments[0].List[1].GetTextAsString());
      Assert.AreEqual("\\*",        status.ResponseText.Arguments[0].List[2].GetTextAsString());
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_CodeWithSubCode()
    {
      var response = ReceiveResponse("* OK [METADATA TOOMANY] text\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("text", status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.MetadataTooMany);
      Assert.AreEqual(0, status.ResponseText.Arguments.Count);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_CodeWithSubCode_HasArguments()
    {
      var response = ReceiveResponse("* OK [METADATA MAXSIZE 1024] text\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("text", status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.IsTrue(status.ResponseText.Code == ImapResponseCode.MetadataMaxSize);
      Assert.AreEqual(1, status.ResponseText.Arguments.Count);
      Assert.AreEqual(ImapDataFormat.Text, status.ResponseText.Arguments[0].Format);
      Assert.AreEqual("1024", status.ResponseText.Arguments[0].GetTextAsString());
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_RespCodeUnknown()
    {
      var response = ReceiveResponse("* OK [X-UNKNOWN] text\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("text", status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.AreEqual("X-UNKNOWN", status.ResponseText.Code);
      Assert.AreEqual(0, status.ResponseText.Arguments.Count);
    }

    [Test]
    public void TestReceiveResponse_StatusResponse_WithRespText_RespCodeUnknown_HasArguments()
    {
      var response = ReceiveResponse("* OK [X-UNKNOWN ARG1 ARG2] text\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapUntaggedStatusResponse>(response);

      var status = response as ImapUntaggedStatusResponse;

      Assert.IsFalse(status.IsTaggedResponse);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.AreEqual("text", status.ResponseText.Text);

      // response status code
      Assert.IsNotNull(status.ResponseText.Code);
      Assert.AreEqual("X-UNKNOWN", status.ResponseText.Code);
      Assert.AreEqual(2, status.ResponseText.Arguments.Count);
      Assert.AreEqual(ImapDataFormat.Text, status.ResponseText.Arguments[0].Format);
      Assert.AreEqual("ARG1", status.ResponseText.Arguments[0].GetTextAsString());
      Assert.AreEqual(ImapDataFormat.Text, status.ResponseText.Arguments[1].Format);
      Assert.AreEqual("ARG2", status.ResponseText.Arguments[1].GetTextAsString());
    }

    [Test]
    public void TestReceiveResponse_TaggedStatusResponse()
    {
      var response = ReceiveResponse("A OK done\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapTaggedStatusResponse>(response);

      var status = response as ImapTaggedStatusResponse;

      Assert.IsTrue(status.IsTaggedResponse);
      Assert.AreEqual("A", status.Tag);
      Assert.AreEqual(ImapResponseCondition.Ok, status.Condition);
      Assert.IsNotNull(status.ResponseText);
      Assert.AreEqual("done", status.ResponseText.Text);
      Assert.IsNull(status.ResponseText.Code);
    }

    [Test]
    public void TestReceiveResponse_CommandContinuationRequest_TextOrBase64()
    {
      var response = ReceiveResponse("+ Ready for additional command text\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapCommandContinuationRequest>(response);

      var continuation = response as ImapCommandContinuationRequest;

      Assert.IsNotNull(continuation.Text);
      Assert.AreEqual("Ready for additional command text", continuation.Text);

      Assert.IsNull(continuation.ResponseText);
    }

    [Test]
    public void TestReceiveResponse_CommandContinuationRequest_TextOrBase64_Empty()
    {
      var response = ReceiveResponse("+ \r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapCommandContinuationRequest>(response);

      var continuation = response as ImapCommandContinuationRequest;

      Assert.IsNotNull(continuation.Text);
      Assert.AreEqual(0, continuation.Text.Length);

      Assert.IsNull(continuation.ResponseText);
    }

    [Test]
    public void TestReceiveResponse_CommandContinuationRequest_WithRespText()
    {
      var response = ReceiveResponse("+ [SERVERBUG] This should not happen\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapCommandContinuationRequest>(response);

      var continuation = response as ImapCommandContinuationRequest;

      Assert.IsNull(continuation.Text);

      Assert.IsNotNull(continuation.ResponseText);
      Assert.AreEqual(continuation.ResponseText.Code, ImapResponseCode.ServerBug);
      Assert.AreEqual("This should not happen", continuation.ResponseText.Text);
      Assert.IsEmpty(continuation.ResponseText.Arguments);
    }

    [Test]
    public void TestReceiveResponse_CommandContinuationRequest_WithRespText_RespCodeOnly()
    {
      var response = ReceiveResponse("+ [CONTACTADMIN]\r\n");

      Assert.IsNotNull(response);
      Assert.IsInstanceOf<ImapCommandContinuationRequest>(response);

      var continuation = response as ImapCommandContinuationRequest;

      Assert.IsNull(continuation.Text);

      Assert.IsNotNull(continuation.ResponseText);
      Assert.AreEqual(continuation.ResponseText.Code, ImapResponseCode.ContactAdmin);
      Assert.IsEmpty(continuation.ResponseText.Text);
      Assert.IsEmpty(continuation.ResponseText.Arguments);
    }

    [Test]
    public void TestReceiveResponseAsUTF8()
    {
      using (var receiver = new PseudoResponseReceiver("A001 OK LANGUAGE-Befehl ausgeführt\r\n")) {
        receiver.ReceiveResponseAsUTF8 = true;

        var response = receiver.ReceiveResponse();

        Assert.IsNotNull(response);
        Assert.AreEqual("LANGUAGE-Befehl ausgeführt", (response as ImapTaggedStatusResponse).ResponseText.Text);
      }
    }
  }
}
