using System;
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.IO.Streams.LineOriented;
using Smdn.Text.Encodings;
using Smdn.Net.Imap4.Protocol.Strings;

namespace Smdn.Net.Imap4.Protocol.Transmission {
  [TestFixture]
  public class ImapSenderTests {
    private class Sender : ImapSender {
      public int UnsentCount => UnsentFragmentCount;

      public Sender(LineOrientedBufferedStream stream)
        : base(stream)
      {
      }

      public void Enqueue(params ImapString[] strings) => Enqueue(false, strings);
      public void SendRemainder() => Send();
    }

    private Sender sender;
    private MemoryStream stream;

    [SetUp]
    public void SetUp()
    {
      stream = new MemoryStream();
      sender = new Sender(new LineOrientedBufferedStream(stream));
    }

    private string GetSent()
    {
      var pos = stream.Position;
      var buffer = new byte[stream.Length];

      try {
        stream.Position = 0L;

        var len = stream.Read(buffer, 0, buffer.Length);

        return OctetEncoding.EightBits.GetString(buffer, 0, len);
      }
      finally {
        stream.Position = pos;
      }
    }

    [Test]
    public void TestSend_String()
    {
      sender.Enqueue((ImapString)"string", "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("string NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_InsertSpace()
    {
      sender.Enqueue("NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("NIL\r\n",
                      GetSent());

      stream.SetLength(0L);

      sender.Enqueue("NIL", "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("NIL NIL\r\n",
                      GetSent());

      stream.SetLength(0L);

      sender.Enqueue(ImapStringList.CreateParenthesized(new[] { "NIL", "NIL" }), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("(NIL NIL) NIL\r\n",
                      GetSent());

      stream.SetLength(0L);

      sender.Enqueue("NIL", ImapStringList.CreateParenthesized(new[] { "NIL", "NIL" }), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("NIL (NIL NIL) NIL\r\n",
                      GetSent());

      stream.SetLength(0L);

      sender.Enqueue(ImapStringList.CreateParenthesized(new[] {
                                                          ImapStringList.CreateParenthesized(new[] { "NIL", "NIL" }),
                                                          ImapStringList.CreateParenthesized(new ImapString[0]),
                                                          ImapStringList.CreateParenthesized("NIL")
                                                        }),
                     "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("((NIL NIL) () (NIL)) NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_QuotedString()
    {
      sender.Enqueue(ImapQuotedString.Create("quoted"), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("\"quoted\" NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_QuotedString_ContainsDQuote()
    {
      sender.Enqueue(ImapQuotedString.Create("\"quoted\""), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("\"\\\"quoted\\\"\" NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_QuotedString_ContainsBackSlash()
    {
      sender.Enqueue(ImapQuotedString.Create("\\quoted\\"), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("\"\\\\quoted\\\\\" NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_QuotedString_ContainsDQuoteAndBackSlash()
    {
      sender.Enqueue(ImapQuotedString.Create("\"\\Sent\""), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("\"\\\"\\\\Sent\\\"\" NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_StringList()
    {
      sender.Enqueue(ImapStringList.Create(new[] { "0000", "NOOP" }), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("0000 NOOP NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_ParenthesizedString()
    {
      sender.Enqueue(ImapStringList.CreateParenthesized(new[] { "\\Seen", "\\Flagged" }), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("(\\Seen \\Flagged) NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_ParenthesizedString_Empty()
    {
      sender.Enqueue(ImapStringList.CreateParenthesized(new ImapString[0]), "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("() NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_ParenthesizedString_Nested()
    {
      sender.Enqueue(ImapStringList.CreateParenthesized(new[] {
                                                          (ImapString)"OR",
                                                          ImapStringList.CreateParenthesized(new[] { "UNFLAGGED", "UNSEEN" }),
                                                          ImapStringList.CreateParenthesized(new[] { "UNFLAGGED", "DRAFT" }),
                                                        }),
                     "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("(OR (UNFLAGGED UNSEEN) (UNFLAGGED DRAFT)) NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_LiteralString_Synchronizing()
    {
      sender.Enqueue(ImapLiteralString.Create("literal", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
                     "NIL\r\n");
      sender.SendRemainder();

      Assert.Greater(sender.UnsentCount, 0);
      Assert.AreEqual("{7}\r\n",
                      GetSent());

      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("{7}\r\nliteral NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_LiteralString_NonSynchronizing()
    {
      sender.Enqueue(ImapLiteralString.Create("literal", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
                     "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("{7+}\r\nliteral NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_LiteralString_Literal8_Synchronizing()
    {
      sender.Enqueue(ImapLiteralString.Create("literal\0", Encoding.ASCII, ImapLiteralOptions.Literal8 | ImapLiteralOptions.Synchronizing),
                     "NIL\r\n");
      sender.SendRemainder();

      Assert.Greater(sender.UnsentCount, 0);
      Assert.AreEqual("~{8}\r\n",
                      GetSent());

      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("~{8}\r\nliteral\0 NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_LiteralString_Literal8_NonSynchronizing()
    {
      sender.Enqueue(ImapLiteralString.Create("literal\0", Encoding.ASCII, ImapLiteralOptions.Literal8 | ImapLiteralOptions.NonSynchronizing),
                     "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("~{8+}\r\nliteral\0 NIL\r\n",
                      GetSent());
    }

    [Test]
    public void TestSend_LiteralString_ReretrieveStreamAgain()
    {
      var literal = ImapLiteralString.Create("差出人", Encoding.UTF8, ImapLiteralOptions.Literal8 | ImapLiteralOptions.NonSynchronizing);

      sender.Enqueue(literal, "NIL\r\n");
      sender.SendRemainder();

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("~{9+}\r\n\xE5\xB7\xAE\xE5\x87\xBA\xE4\xBA\xBA NIL\r\n",
                      GetSent());

      stream.SetLength(0L);

      sender.Enqueue(literal, "NIL\r\n");
      sender.SendRemainder(); // send again (stream *will* seeked to start)

      Assert.AreEqual(0, sender.UnsentCount);
      Assert.AreEqual("~{9+}\r\n\xE5\xB7\xAE\xE5\x87\xBA\xE4\xBA\xBA NIL\r\n",
                      GetSent());
    }
  }
}
