using System;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.Strings;
using Smdn.Net.MessageAccessProtocols.Strings;

namespace Smdn.Net.Imap4.Protocol {
  [TestFixture]
  public class ImapCapabilityTests {
    [Test]
    public void TestEquals()
    {
      var x = ImapCapability.Imap4Rev1;

      Assert.IsTrue(x.Equals(ImapCapability.Imap4Rev1));
      Assert.IsTrue(x.Equals(new ImapCapability("imap4REV1")));
      Assert.IsTrue(x.Equals((object)"IMAP4rev1"));
      Assert.IsTrue(x.Equals((object)"imap4rev1"));
      Assert.IsTrue(x.Equals((object)ImapCapability.Imap4Rev1));
      Assert.IsTrue(x.Equals("IMAP4rev1"));
      Assert.IsTrue(x.Equals("imap4rev1"));
      Assert.IsTrue(x.Equals(ImapCapability.Imap4Rev1.ToString()));

      Assert.IsFalse(x.Equals(ImapCapability.Imap4));
      Assert.IsFalse(x.Equals((ImapString)ImapCapability.Imap4Rev1.Value));
      Assert.IsFalse(x.Equals(new ImapMessageFlag("IMAP4rev1")));
      Assert.IsFalse(x.Equals((object)new ImapMessageFlag("IMAP4rev1")));
    }

    [Test]
    public void TestOpEquality()
    {
      var x = ImapCapability.Imap4Rev1;

      Assert.IsTrue(x == ImapCapability.Imap4Rev1);
      Assert.IsTrue(x == new ImapCapability("imap4REV1"));
      Assert.IsFalse(x != ImapCapability.Imap4Rev1);
      Assert.IsFalse(x != new ImapCapability("imap4REV1"));

      Assert.IsFalse(x == ImapCapability.Imap4);
      Assert.IsFalse(x == new ImapCapability("imap4"));
      Assert.IsTrue(x != ImapCapability.Imap4);
      Assert.IsTrue(x != new ImapCapability("imap4"));
    }

    [Test]
    public void TestSet_TryGetValue()
    {
      Assert.IsTrue(ImapCapability.AllCapabilities.TryGetValue("LITERAL+", out var capa1));
      Assert.AreSame(ImapCapability.LiteralNonSync, capa1);

      Assert.IsTrue(ImapCapability.AllCapabilities.TryGetValue("literal+", out var capa2));
      Assert.AreSame(ImapCapability.LiteralNonSync, capa2);

      Assert.AreSame(capa1, capa2);

      Assert.IsTrue(ImapCapability.AllCapabilities.TryGetValue("AUTH=PLAIN", out var capaAuthPlain1));
      Assert.IsNotNull(capaAuthPlain1);

      Assert.IsTrue(ImapCapability.AllCapabilities.TryGetValue("auth=plain", out var capaAuthPlain2));
      Assert.AreSame(capaAuthPlain1, capaAuthPlain2);
    }

    [Test]
    public void TestSet_Contains1()
    {
      TestSetContains(ImapCapability.AllCapabilities);
    }

    private void TestSetContains(IReadOnlyStringSet<ImapCapability> capabilities)
    {
      Assert.IsTrue(capabilities.Contains(ImapCapability.LiteralNonSync));
      Assert.IsTrue(capabilities.Contains("LITERAL+"));
      Assert.IsTrue(capabilities.Contains("literal+"));

      Assert.IsTrue(capabilities.Contains("AUTH=CRAM-MD5"));
      Assert.IsTrue(capabilities.Contains(new ImapCapability("auth=cram-md5")));

      Assert.IsFalse(capabilities.Contains("***non-existent-capability***"));
      Assert.IsFalse(capabilities.Contains(new ImapCapability("***non-existent-capability***")));
    }

    [Test]
    public void TestSet_Contains2()
    {
      var capabilities = new ImapPseudoCapabilitySet(
        ImapCapability.LiteralNonSync,
        ImapCapability.Binary,
        new ImapCapability("AUTH=CRAM-MD5"),
        new ImapCapability("AUTH=PLAIN")
      );

      TestSetContains(capabilities);

      Assert.IsFalse(capabilities.Contains(ImapCapability.Imap4Rev1));
      Assert.IsFalse(capabilities.Contains("imap4rev1"));

      Assert.IsFalse(capabilities.Contains(new ImapCapability("auth=login")));
      Assert.IsFalse(capabilities.Contains("auth=login"));
    }

    [Test]
    public void TestToString()
    {
      Assert.AreEqual("IMAP4rev1", ImapCapability.Imap4Rev1.ToString());
      Assert.AreEqual("AUTH=LOGIN", new ImapCapability("AUTH=LOGIN").ToString());
    }

    [Test]
    public void TestSerializeBinary()
    {
      TestUtils.Assert.IsSerializableBinaryFormat(ImapCapability.Imap4Rev1, delegate(ImapCapability deserialized) {
        Assert.AreNotSame(ImapCapability.Imap4Rev1, deserialized);
        Assert.AreEqual(ImapCapability.Imap4Rev1, deserialized);
      });
    }
  }
}
