using System;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.DataStructures;

namespace Smdn.Net.Imap4.Protocol {
  [TestFixture]
  public class ImapFetchDataItemsTests : ImapCombinableDataItemsTestsBase {
    [Test]
    public void TestMacro()
    {
      Assert.AreEqual("ALL", Send(ImapFetchDataItems.All));
      Assert.AreEqual("FAST", Send(ImapFetchDataItems.Fast));
      Assert.AreEqual("FULL", Send(ImapFetchDataItems.Full));
    }

    [Test]
    public void TestNonMacro()
    {
      Assert.AreEqual("(UID)", Send(ImapFetchDataItems.Uid));
    }

    [Test]
    public void TestCombine()
    {
      var fetch1 = ImapFetchDataItems.Envelope;
      var fetch2 = ImapFetchDataItems.Uid;

      Assert.AreEqual("(ENVELOPE UID)", Send(ImapFetchDataItems.Combine(fetch1, fetch2)));
      Assert.AreEqual("(ENVELOPE UID)", Send(fetch1 + fetch2));
      Assert.AreEqual("(ENVELOPE UID)", Send(fetch1.CombineWith(fetch2)));

      Assert.AreEqual("(UID ENVELOPE)", Send(ImapFetchDataItems.Combine(fetch2, fetch1)));
      Assert.AreEqual("(UID ENVELOPE)", Send(fetch2 + fetch1));
      Assert.AreEqual("(UID ENVELOPE)", Send(fetch2.CombineWith(fetch1)));
    }

    [Test]
    public void TestCombine_WithMacro()
    {
      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE UID)", Send(ImapFetchDataItems.Fast + ImapFetchDataItems.Uid));
      Assert.AreEqual("(UID FLAGS INTERNALDATE RFC822.SIZE)", Send(ImapFetchDataItems.Uid + ImapFetchDataItems.Fast));

      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE ENVELOPE UID)", Send(ImapFetchDataItems.All + ImapFetchDataItems.Uid));
      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE ENVELOPE BODY UID)", Send(ImapFetchDataItems.Full + ImapFetchDataItems.Uid));
    }

    [Test]
    public void TestBody()
    {
      Assert.AreEqual("(BODY[])", Send(ImapFetchDataItems.BodyText()));
      Assert.AreEqual("(BODY.PEEK[])", Send(ImapFetchDataItems.BodyPeek()));
      Assert.AreEqual("(BODY[])", Send(ImapFetchDataItems.BodyText(false)));
      Assert.AreEqual("(BODY.PEEK[])", Send(ImapFetchDataItems.BodyText(true)));
      Assert.AreEqual("(BODY[heADEr])", Send(ImapFetchDataItems.BodyText("heADEr")));
      Assert.AreEqual("(BODY.PEEK[]<1000.2000>)", Send(ImapFetchDataItems.BodyPeek(1000L, 2000L)));
      Assert.AreEqual("(BODY.PEEK[]<1000.2000>)", Send(ImapFetchDataItems.BodyPeek(new ImapFetchRange(1000L, 2000L))));
      Assert.AreEqual("(BODY[HEADER (subject date)]<1000.2000>)", Send(ImapFetchDataItems.BodyText("HEADER (subject date)", 1000, 2000)));
      Assert.AreEqual("(BODY.PEEK[HEADER (from to)]<0.1000>)", Send(ImapFetchDataItems.BodyPeek("HEADER (from to)", 0, 1000)));
    }

    [Test]
    public void TestBody_InvalidArgument()
    {
      Assert.Throws<ArgumentException>(() => ImapFetchDataItems.BodyPeek(new ImapFetchRange(1000L)), "#1");
      Assert.Throws<ArgumentException>(() => ImapFetchDataItems.BodyText(new ImapFetchRange(1000L)), "#2");
    }

    [Test]
    public void TestBody_ArgumentOutOfRange()
    {
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.BodyPeek(1000L, 0L), "#1");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.BodyText(1000L, 0L), "#2");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.BodyPeek(-1L, 1L), "#3");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.BodyText(-1L, 1L), "#4");
    }

    [Test]
    public void TestBinary()
    {
      Assert.AreEqual("(BINARY[1.1])", Send(ImapFetchDataItems.Binary("1.1")));
      Assert.AreEqual("(BINARY.PEEK[1.1])", Send(ImapFetchDataItems.BinaryPeek("1.1")));
      Assert.AreEqual("(BINARY[1.1])", Send(ImapFetchDataItems.Binary(false, "1.1")));
      Assert.AreEqual("(BINARY.PEEK[1.1])", Send(ImapFetchDataItems.Binary(true, "1.1")));
      Assert.AreEqual("(BINARY[heADEr])", Send(ImapFetchDataItems.Binary("heADEr")));
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<0.1024>)", Send(ImapFetchDataItems.BinaryPeek("1.1.TEXT", 0L, 1024L)));
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<0.1024>)", Send(ImapFetchDataItems.BinaryPeek("1.1.TEXT", new ImapFetchRange(0L, 1024L))));
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<1000.2000>)", Send(ImapFetchDataItems.BinaryPeek("1.1.TEXT", 1000L, 2000L)));
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<1000.2000>)", Send(ImapFetchDataItems.BinaryPeek("1.1.TEXT", new ImapFetchRange(1000L, 2000L))));
    }

    [Test]
    public void TestBinary_InvalidArgument()
    {
      Assert.Throws<ArgumentException>(() => ImapFetchDataItems.BinaryPeek("1.1", new ImapFetchRange(1000L)), "#1");
      Assert.Throws<ArgumentException>(() => ImapFetchDataItems.Binary    ("1.1", new ImapFetchRange(1000L)), "#2");
    }

    [Test]
    public void TestBinary_ArgumentOutOfRange()
    {
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.BinaryPeek("1.1", 1000L, 0L), "#1");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.Binary    ("1.1", 1000L, 0L), "#2");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.BinaryPeek("1.1", -1L, 1L), "#3");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItems.Binary    ("1.1", -1L, 1L), "#4");
    }

    [Test]
    public void TestBinary_SectionNull()
    {
      Assert.Throws<ArgumentNullException>(() => ImapFetchDataItems.Binary(null));
    }

    [Test]
    public void TestBinary_SectionEmpty()
    {
      Assert.Throws<ArgumentException>(() => ImapFetchDataItems.Binary(string.Empty));
    }

    [Test]
    public void TestCombine_ContainsRequiredCapability_1()
    {
      var dataItem1 = ImapFetchDataItems.ModSeq;
      var dataItem2 = ImapFetchDataItems.InternalDate;
      var capabilitiesCondStore = new ImapPseudoCapabilitySet(ImapCapability.CondStore);

      var combined = dataItem1 + dataItem2;

      Assert.AreEqual("(MODSEQ INTERNALDATE)", Send(combined));
      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(combined));
      Assert.IsTrue(capabilitiesCondStore.IsCapable(combined));

      combined = dataItem2 + dataItem1;

      Assert.AreEqual("(INTERNALDATE MODSEQ)", Send(combined));
      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(combined));
      Assert.IsTrue(capabilitiesCondStore.IsCapable(combined));
    }

    [Test]
    public void TestCombine_ContainsRequiredCapability_2()
    {
      var dataItem1 = ImapFetchDataItems.ModSeq;
      var dataItem2 = ImapFetchDataItems.Binary("1");
      var capabilitiesCondStore = new ImapPseudoCapabilitySet(ImapCapability.CondStore);
      var capabilitiesCondStoreBinary = new ImapPseudoCapabilitySet(ImapCapability.CondStore, ImapCapability.Binary);

      var combined = dataItem1 + dataItem2;

      Assert.AreEqual("(MODSEQ BINARY[1])", Send(combined));
      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(combined));
      Assert.IsFalse(capabilitiesCondStore.IsCapable(combined));
      Assert.IsTrue(capabilitiesCondStoreBinary.IsCapable(combined));

      combined = dataItem2 + dataItem1;

      Assert.AreEqual("(BINARY[1] MODSEQ)", Send(combined));
      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(combined));
      Assert.IsFalse(capabilitiesCondStore.IsCapable(combined));
      Assert.IsTrue(capabilitiesCondStoreBinary.IsCapable(combined));
    }

    [Test]
    public void TestCombine_WithMacro_ContainsRequiredCapability()
    {
      var dataItem1 = ImapFetchDataItems.ModSeq;
      var dataItem2 = ImapFetchDataItems.Fast;
      var capabilitiesCondStore = new ImapPseudoCapabilitySet(ImapCapability.CondStore);

      var combined = dataItem1 + dataItem2;

      Assert.AreEqual("(MODSEQ FLAGS INTERNALDATE RFC822.SIZE)", Send(combined));
      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(combined));
      Assert.IsTrue(capabilitiesCondStore.IsCapable(combined));

      combined = dataItem2 + dataItem1;

      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE MODSEQ)", Send(combined));
      Assert.IsFalse(ImapPseudoCapabilitySet.Empty.IsCapable(combined));
      Assert.IsTrue(capabilitiesCondStore.IsCapable(combined));
    }
  }
}
